package be.business.connector.core.utils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.xml.bind.JAXBException;

import org.apache.commons.lang3.Validate;
import org.apache.log4j.Logger;

import be.business.connector.core.exceptions.IntegrationModuleException;
import be.recipe.services.executor.GetPrescriptionForExecutorResult;

/**
 * The Class PersistentCache.
 */
public class PersistentCache<K, V> {

	private static final Logger LOG = Logger.getLogger(PersistentCache.class);
	
	private Map<K, V> cache = new LinkedHashMap<K, V>();

	/** The Constant MAX_SIZE. */
	private final int maxSize;

	/** The archiving path. */
	private final String archivingPath;

	/** The property handler. */
	private final PropertyHandler propertyHandler;

	/**
	 * Instantiates a new persistent key cache.
	 *
	 * @param propertyHandler
	 *            the property handler
	 * @param cacheType
	 *            the cache type
	 * @throws Exception
	 *             the exception
	 */
	public PersistentCache(final PropertyHandler propertyHandler, PersistentCacheType cacheType) throws IntegrationModuleException {
		Validate.notNull(propertyHandler, "propertyHandler can't be Null");
		Validate.notNull(cacheType, "cacheType can't be Null");
		this.propertyHandler = propertyHandler;
		this.maxSize = getPropertyHandler().getIntegerProperty(cacheType + ".cache.size");
		this.archivingPath = getPropertyHandler().getProperty(cacheType + ".cache.directory");
		loadEntries();
	}

	/**
	 * Load entries.
	 *
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 * @throws IntegrationModuleException
	 */
	private void loadEntries() throws IntegrationModuleException {
		try {
			LOG.info("archivingPath: " + archivingPath);
			final File rootFolder = new File(archivingPath);
			if (!rootFolder.exists() || !rootFolder.isDirectory()) {
				rootFolder.mkdir();
			}

			final File[] files = rootFolder.listFiles();

			Arrays.sort(files, new Comparator<File>() {
				public int compare(File f1, File f2) {
					return Long.valueOf(f2.lastModified()).compareTo(f1.lastModified());
				}
			});
			for (final File file : files) {
				final byte[] content = Files.readAllBytes(file.toPath());
				put((K) file.getName(), (V) content);
			}
		} catch (Exception e) {
			throw new IntegrationModuleException(e);
		}
	}

	/**
	 * Gets the single instance of PersistentKeyCache.
	 *
	 * @return single instance of PersistentKeyCache
	 */
	public Map<K, V> getInstance() {
		return cache;
	}

	/**
	 * Gets the key.
	 *
	 * @param keyId
	 *            the key id
	 * @return the key
	 */
	public V get(final K keyId) {
		Validate.notNull(keyId, "keyId can't be Null");
		if (cache.containsKey(keyId)) {
			return cache.get(keyId);
		}
		return null;
	}

	public boolean containsKey(final String keyId) {
		return cache.containsKey(keyId);
	}

	/**
	 * Put.
	 *
	 * @param key
	 *            the key id
	 * @param value
	 *            the key
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 */
	public void put(final K key, final V value) {
		Validate.notNull(key, "key can't be Null");
		Validate.notNull(value, "value can't be Null");
		if (!cache.containsKey(key) && cache.keySet().size() == maxSize) {
			final K lastKey = (K) cache.keySet().toArray()[0];
			cache.remove(lastKey);
			remove(lastKey);

		}
		if (!cache.containsKey(key)) {
			writeKey(key, value);
			cache.put(key, value);
		}
	}

	public static String rewriteKey(String key) {
		Validate.notNull(key, "key can't be Null");
		return key.replaceAll("\\/", "_");
	}

	/**
	 * Removes the.
	 *
	 * @param keyId
	 *            the key id
	 */
	private void remove(final K keyId) {
		Validate.notNull(keyId, "keyId can't be Null");
		final File file = new File(cache + File.separator + keyId);
		file.delete();
	}

	/**
	 * Write key.
	 *
	 * @param keyId
	 *            the key id
	 * @param value
	 *            the value
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 * @throws JAXBException
	 */
	private void writeKey(final K keyId, final V value) {
		Validate.notNull(keyId, "keyId can't be Null");
		Validate.notNull(value, "value can't be Null");
		try {
			final File file = new File(archivingPath + File.separator + keyId);
			if (!file.exists()) {
				byte[] bytes = null;
				if (value instanceof byte[]) {
					bytes = (byte[]) value;
				}
				if (value instanceof GetPrescriptionForExecutorResult) {
					final MarshallerHelper<GetPrescriptionForExecutorResult, GetPrescriptionForExecutorResult> marshallerHelpe = new MarshallerHelper<GetPrescriptionForExecutorResult, GetPrescriptionForExecutorResult>(
							GetPrescriptionForExecutorResult.class, GetPrescriptionForExecutorResult.class);

					bytes = marshallerHelpe.marsh((GetPrescriptionForExecutorResult) value).getBytes();
				}
				Files.write(file.toPath(), bytes, StandardOpenOption.CREATE_NEW);
			}
		} catch (JAXBException | IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Prints the.
	 */
	public void print() {
		final Iterator<K> iterator = cache.keySet().iterator();
		while (iterator.hasNext()) {
			final K next = iterator.next();
			LOG.info(next + " " + cache.get(next));
		}
	}

	/**
	 * Gets the property handler.
	 *
	 * @return the property handler
	 */
	public PropertyHandler getPropertyHandler() {
		return propertyHandler;
	}

	public void clear() {
		cache.clear();
	}
}