package be.business.connector.gfddpp.pcdh;

import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.Semaphore;

import be.business.connector.common.*;
import be.business.connector.core.utils.*;
import net.sf.ehcache.Element;

import org.apache.log4j.Logger;
import org.perf4j.aop.Profiled;

import be.apb.gfddpp.common.log.LogsSMC;
import be.apb.gfddpp.common.log.LogsSMC.IntermediateMessage;
import be.apb.gfddpp.common.medicationscheme.status.MSStatusResolver;
import be.apb.gfddpp.common.status.StatusCode;
import be.apb.gfddpp.common.utils.SingleMessageWrapper;
import be.apb.gfddpp.domain.PharmacyIdType;
import be.apb.gfddpp.helper.SingleMessageValidationHelper;
import be.apb.standards.gfddpp.request.PatientType;
import be.apb.standards.smoa.schema.model.v1.HistoryProductType;
import be.apb.standards.smoa.schema.model.v1.MedicationHistoryType;
import be.apb.standards.smoa.schema.model.v1.ReferencePharmacyType;
import be.apb.standards.smoa.schema.v1.*;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.common.module.AbstractIntegrationModule;
import be.business.connector.core.technical.connector.utils.Crypto;
import be.business.connector.gfddpp.domain.ThreeStateBoolean;
import be.business.connector.gfddpp.domain.medicationscheme.Error;
import be.business.connector.gfddpp.domain.medicationscheme.Status;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.FetchDataEntriesRequest;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.FetchDataEntriesResponse;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.RetrieveTimestampsRequest;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.RetrieveTimestampsResponse;
import be.business.connector.gfddpp.tipsystem.threading.KGSSThread;
import be.business.connector.gfddpp.tipsystem.threading.ProductDecryptorThread;
import be.business.connector.gfddpp.utils.ConsolidateUtils;
import be.business.connector.gfddpp.utils.GFDDPPValidationUtils;
import be.business.connector.gfddpp.utils.MappingUtils;
import be.business.connector.gfddpp.utils.RequestCreatorUtils;
import be.business.connector.projects.common.services.pcdh.PcdhServiceImpl;
import be.business.connector.projects.common.utils.ValidationUtils;
import be.ehealth.apb.gfddpp.services.pcdh.SealedRequestType;
import be.ehealth.apb.gfddpp.services.pcdh.SealedResponseType;
import be.ehealth.apb.gfddpp.services.pcdh.StatusType;
import be.fgov.ehealth.etee.crypto.encrypt.EncryptionToken;
import org.apache.commons.collections.CollectionUtils;

import java.util.*;

import static be.apb.gfddpp.common.medicationscheme.status.MSStatusCode.MEDICATION_SCHEME_VALIDATION_ERRORS;
import static org.apache.commons.lang.StringUtils.defaultIfBlank;

public class PCDHIntegrationModuleImpl extends AbstractIntegrationModule implements PCDHIntegrationModule {
    private final static Logger LOG = Logger.getLogger(PCDHIntegrationModuleImpl.class);
    private SingleMessageValidationHelper xmlValidationhelper;
    private final int MAX_THREAD = 30;
    private int threadLimit = 10;
    private final String THREAD_PROPERTY = "thread.number.limit";


    public PCDHIntegrationModuleImpl() throws IntegrationModuleException {
        super();
        this.xmlValidationhelper = new SingleMessageValidationHelper();
        initThread();

        getJaxContextCentralizer().addContext(HistoryProductType.class);

        LOG.debug("******************** PCDH module init **********************");
    }


    @Profiled(logFailuresSeparately = true, tag = "0.PCDHIntegrationModuleImpl#getMedicationSchemeTimestamps", logger = "org.perf4j.TimingLogger_Executor")
    public RetrieveTimestampsResponse getMedicationSchemeTimestamps(final RetrieveTimestampsRequest retrieveTimestampsRequest) throws IntegrationModuleException {
        ApplicationConfig.getInstance().assertValidSession();

        LOG.debug("******************** GET_MEDICATIONSCHEME_TIMESTAMPS - Validate incoming fields **********************");
        final List<Error> errors = GFDDPPValidationUtils.validate(retrieveTimestampsRequest);
        if (CollectionUtils.isNotEmpty(errors)) {
            final RetrieveTimestampsResponse response = new RetrieveTimestampsResponse();
            final Status status = new Status(MEDICATION_SCHEME_VALIDATION_ERRORS.getCode(), MSStatusResolver.getStatusMessage(MEDICATION_SCHEME_VALIDATION_ERRORS));
            status.getErrors().addAll(errors);
            response.setStatus(status);
            response.setClientMessageID(retrieveTimestampsRequest == null ? null : retrieveTimestampsRequest.getClientMessageID());
            response.setServerMessageID(UUID.randomUUID().toString());
            response.setCurrentDateTime(Calendar.getInstance());
            return response;
        }

        final String clientMessageId = retrieveTimestampsRequest.getClientMessageID();
        try {
            LOG.debug("******************** GET_MEDICATIONSCHEME_TIMESTAMPS -  Create request  **********************");
            SealedRequestType sealedRequestType = new SealedRequestType();
            String requestorId = StandaloneRequestorProvider.getRequestorIdInformation();
            String requestorType = StandaloneRequestorProvider.getRequestorTypeInformation();

            String xml = RequestCreatorUtils.createRequestMedicationSchemeTimestamps(requestorId, requestorType, retrieveTimestampsRequest.getSubjectIDs(), clientMessageId);

            LOG.debug("******************** GET_MEDICATIONSCHEME_TIMESTAMPS - UNSEALED request: " + xml + " ************************");

            LOG.debug("******************** GET_MEDICATIONSCHEME_TIMESTAMPS -  Seal request xml  **********************");

            List<EncryptionToken> etks = getEtkHelper().getPCDH_ETK();
            byte[] Request = SealedProcessor.createSealedSync(etks, xml, "GetTimestamps");
            sealedRequestType.setRequestParametersSealed(Request);

            LOG.info("******************** GET_MEDICATIONSCHEME_TIMESTAMPS - CALL PCDH **********************");
            SealedResponseType sealedResponseType = PcdhServiceImpl.getInstance().getData(sealedRequestType);

            LOG.info("******************** GET_MEDICATIONSCHEME_TIMESTAMPS - VALIDATE RESPONSE PCDH **********************");
            validateResponse(sealedResponseType);

            LOG.info("******************** GET_MEDICATIONSCHEME_TIMESTAMPS - UNSEAL SingleMessage response **********************");
            byte[] singleMessageSealed = sealedResponseType.getSingleMessageSealed();

            byte[] singleMessage = Crypto.unseal(singleMessageSealed);

            xmlValidationhelper.assertValidSingleMessage(singleMessage);
            final SingleMessage smg = getJaxContextCentralizer().toObject(SingleMessage.class, singleMessage);
            final SingleMessageWrapper wrapper = new SingleMessageWrapper(smg);

            final List<AbstractEventType> events = wrapper.getEvents();
            if (CollectionUtils.isEmpty(events) || events.size() > 1 || !(events.get(0) instanceof MedicationSchemeTimestampsResponse)) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.number.of.timestamp.responses"));
            }

            final MedicationSchemeTimestampsResponse pcdhResponse = (MedicationSchemeTimestampsResponse) events.get(0);
            if (!defaultIfBlank(clientMessageId, "").equals(defaultIfBlank(pcdhResponse.getClientMessageId(), ""))) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.clientmessageids.not.equal"));
            }

            LOG.info("******************** GETTIMESTAMPS - Create response **********************");
            return MappingUtils.mapMedicationSchemeTimestampsResponse(pcdhResponse);
        } catch (Throwable t) {
            Exceptionutils.errorHandler(t);
            return null;
        }
    }

    @Profiled(logFailuresSeparately = true, tag = "0.PCDHIntegrationModuleImpl#getMedicationSchemeEntries", logger = "org.perf4j.TimingLogger_Executor")
    public FetchDataEntriesResponse getMedicationSchemeEntries(final FetchDataEntriesRequest getMedicationEntriesRequest) throws IntegrationModuleException {
        ApplicationConfig.getInstance().assertValidSession();

        LOG.debug("******************** GET_MEDICATIONSCHEME_ENTRIES - Validate incoming fields **********************");
        final List<Error> errors = GFDDPPValidationUtils.validate(getMedicationEntriesRequest);
        if (CollectionUtils.isNotEmpty(errors)) {
            final FetchDataEntriesResponse response = new FetchDataEntriesResponse();
            final Status status = new Status(MEDICATION_SCHEME_VALIDATION_ERRORS.getCode(), MSStatusResolver.getStatusMessage(MEDICATION_SCHEME_VALIDATION_ERRORS));
            status.getErrors().addAll(errors);
            response.setStatus(status);
            response.setServerMessageID(UUID.randomUUID().toString());
            response.setClientMessageID(getMedicationEntriesRequest == null ? null : getMedicationEntriesRequest.getClientMessageID());
            return response;
        }

        final String subjectId = getMedicationEntriesRequest.getSubjectID();
        String clientMessageId = getMedicationEntriesRequest.getClientMessageID();
        try {
            LOG.debug("******************** GET_MEDICATIONSCHEME_ENTRIES -  Create request  **********************");
            SealedRequestType sealedRequestType = new SealedRequestType();
            String requestorId = StandaloneRequestorProvider.getRequestorIdInformation();
            String requestorType = StandaloneRequestorProvider.getRequestorTypeInformation();

            Integer paginationIndex = 1;
            if (getMedicationEntriesRequest.getPagination() != null) {
                paginationIndex = getMedicationEntriesRequest.getPagination().getPaginationIndex();
            }
            String xml = RequestCreatorUtils.createRequestMedicationSchemeEntries(requestorId, requestorType, subjectId, paginationIndex, clientMessageId);

            LOG.debug("******************** GET_MEDICATIONSCHEME_ENTRIES - UNSEALED request: " + xml + " ************************");

            LOG.debug("******************** GET_MEDICATIONSCHEME_ENTRIES -  Seal request xml  **********************");

            List<EncryptionToken> etks = getEtkHelper().getPCDH_ETK();
            byte[] Request = SealedProcessor.createSealedSync(etks, xml, "GetMedicationScheme");
            sealedRequestType.setRequestParametersSealed(Request);

            LOG.info("******************** GET_MEDICATIONSCHEME_ENTRIES - CALL PCDH **********************");
            SealedResponseType sealedResponseType = PcdhServiceImpl.getInstance().getData(sealedRequestType);

            LOG.info("******************** GET_MEDICATIONSCHEME_ENTRIES - VALIDATE RESPONSE PCDH **********************");
            validateResponse(sealedResponseType);

            LOG.info("******************** GET_MEDICATIONSCHEME_ENTRIES - UNSEAL SingleMessage response **********************");
            byte[] singleMessageSealed = sealedResponseType.getSingleMessageSealed();

            byte[] singleMessage = Crypto.unseal(singleMessageSealed);

            xmlValidationhelper.assertValidSingleMessage(singleMessage);
            final SingleMessage smg = getJaxContextCentralizer().toObject(SingleMessage.class, singleMessage);
            final SingleMessageWrapper wrapper = new SingleMessageWrapper(smg);

            final List<AbstractEventType> events = wrapper.getEvents();
            if (CollectionUtils.isEmpty(events) || events.size() > 1 || !(events.get(0) instanceof MedicationSchemeEntriesResponse)) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.number.of.fetchmedicationentries.responses"));
            }

            final MedicationSchemeEntriesResponse pcdhResponse = (MedicationSchemeEntriesResponse) events.get(0);
            if (!defaultIfBlank(clientMessageId, "").equals(defaultIfBlank(pcdhResponse.getClientMessageId(), ""))) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.clientmessageids.not.equal"));
            }
            if (!defaultIfBlank(subjectId, "").equals(defaultIfBlank(pcdhResponse.getSubjectId(), ""))) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.subjectids.not.equal"));
            }

            LOG.info("******************** GET_MEDICATIONSCHEME_ENTRIES - load encryption keys **********************");
            final Set<String> encryptionKeyIds = new HashSet<String>();
            if (pcdhResponse.getMedicationScheme() != null) {
                for (DataEntryResponse pcdhDataEntry : pcdhResponse.getMedicationScheme().getDataEntry()) {
                    if (pcdhDataEntry.getEncryptedData() != null) {
                        encryptionKeyIds.add(pcdhDataEntry.getEncryptedData().getEncryptionKeyId());
                    }
                }
            }
            loadEncryptionKeysToCache(new ArrayList<String>(encryptionKeyIds), getSystemETKFromCache());

            LOG.info("******************** GET_MEDICATIONSCHEME_ENTRIES - create Response **********************");
            final FetchDataEntriesResponse response = MappingUtils.mapMedicationSchemeEntriesResponse(pcdhResponse, getKgssCache(), threadLimit);
            return response;
        } catch (Throwable t) {
            Exceptionutils.errorHandler(t);
            return null;
        }
    }

    @Override
    @Profiled(logFailuresSeparately = true, tag = "0.PCDHIntegrationModuleImpl#getPharmacyDetails", logger = "org.perf4j.TimingLogger_Executor")
    public String getPharmacyDetails(String patientIdType, String patientId, String dGuid, String motivationType, String motivationText, boolean requestPatientSignature, ThreeStateBoolean therapeuticRelationShip) throws IntegrationModuleException {
        ApplicationConfig.getInstance().assertValidSession();

        LOG.debug("******************** GETPHARMACYDETAILS - Validate incoming fields **********************");
        ValidationUtils.validateIncomingFieldsGetPharmacyDetails(patientIdType, patientId, dGuid, motivationText, motivationType);
        ValidationUtils.validatePatientId(patientId);

        String pharmacyId = null;

        try {
            LOG.debug("******************** GETPHARMACYDETAILS -  Create request  **********************");
            SealedRequestType sealedRequestType = new SealedRequestType();
            PatientType patient = RequestCreatorUtils.createPatientType(patientId, patientIdType);

            LOG.debug("******************** GETPHARMACYDETAILS - Get requestor information from SAML  **********************");
            String requestorId = StandaloneRequestorProvider.getRequestorIdInformation();
            String requestorType = StandaloneRequestorProvider.getRequestorTypeInformation();

            String xml = RequestCreatorUtils.createRequestPharmacyDetails(requestorId, requestorType, patient, dGuid, motivationType, motivationText);

            LOG.debug("******************** GETPHARMACYDETAILS - UNSEALED request: " + xml + " ************************");

            LOG.debug("******************** GETPHARMACYDETAILS -  Seal request xml  **********************");

            List<EncryptionToken> etks = getEtkHelper().getPCDH_ETK();
            byte[] Request = SealedProcessor.createSealedSync(etks, xml, "RequestPharmacyDetails");
            sealedRequestType.setRequestParametersSealed(Request);

            LOG.info("Request sealed");
            sealedRequestType.setAuthorizationParameters(RequestCreatorUtils.createPCDHAuthorizationParameters(requestPatientSignature, therapeuticRelationShip, dGuid, patient.getPatientId(), getEncryptionUtils()));

            if (requestPatientSignature) {
                LOG.info("Signature required");
                sealedRequestType.getAuthorizationParameters().getPatientConsent().getPatientSignature().setSignature(Crypto.seal(etks, sealedRequestType.getAuthorizationParameters().getPatientConsent().getPatientSignature().getSignature()));
                LOG.info("Message signed");
            }

            LOG.info("******************** GETPHARMACYDETAILS - CALL PCDH **********************");
            SealedResponseType sealedResponseType = PcdhServiceImpl.getInstance().getPharmacyDetails(sealedRequestType);

            LOG.info("******************** GETPHARMACYDETAILS - VALIDATE RESPONSE PCDH **********************");
            validateResponse(sealedResponseType);

            LOG.info("******************** GETPHARMACYDETAILS - UNSEAL SingleMessage response **********************");
            byte[] singleMessageSealed = sealedResponseType.getSingleMessageSealed();

            byte[] singleMessage = Crypto.unseal(singleMessageSealed);

            xmlValidationhelper.assertValidSingleMessage(singleMessage);
            SingleMessage sng = getJaxContextCentralizer().toObject(SingleMessage.class, singleMessage);

            SingleMessageWrapper wrapper = new SingleMessageWrapper(sng);
            if (wrapper.getEntities() != null && wrapper.getEntities().size() > 0) {
                ReferencePharmacyType referencePharmacyType = ((ReferencePharmacyType) wrapper.getEntities().get(0));
                pharmacyId = PharmacyIdType.valueOf(referencePharmacyType.getPharmacyId().getClass()).getIdFrom(referencePharmacyType.getPharmacyId());
            }

            LOG.debug("returned pharmacyId: " + pharmacyId);
        } catch (Throwable t) {
            Exceptionutils.errorHandler(t);
        }

        return pharmacyId;
    }

    @Override
    @Profiled(logFailuresSeparately = true, tag = "0.PCDHIntegrationModuleImpl#getData", logger = "org.perf4j.TimingLogger_Executor")
    public String getData(String patientIdType, String patientId, String dataType, String dateRange, boolean ExcludeOwnData, byte[] localData, boolean requestPatientSignature, ThreeStateBoolean therapeuticRelationShip) throws IntegrationModuleException {
        ApplicationConfig.getInstance().assertValidSession();

        LOG.debug("******************** GETDATA - Validate incoming fields **********************");
        ValidationUtils.validateIncomingFieldsGetData(patientIdType, patientId, dataType, dateRange);
        ValidationUtils.validatePatientId(patientId);

        String response = null;
        try {
            LOG.debug("******************** GETDATA - Create request  **********************");
            SealedRequestType sealedRequestType = new SealedRequestType();

            LOG.debug("******************** GETDATA - Create Patient object  **********************");
            PatientType patient = RequestCreatorUtils.createPatientType(patientId, patientIdType);

            LOG.debug("******************** GETDATA - Get requestor information from SAML  **********************");
            String requestorId = StandaloneRequestorProvider.getRequestorIdInformation();
            String requestorType = StandaloneRequestorProvider.getRequestorTypeInformation();

            LOG.debug("******************** GETDATA - Create get data request **********************");
            String xml = RequestCreatorUtils.createGetDataParams(patient, requestorId, requestorType, ExcludeOwnData, dateRange, dataType);
            LOG.debug("******************** GETDATA - Request parameter sealed \n" + xml);
            LOG.info("******************** GETDATA - Seal request xml  **********************");
            List<EncryptionToken> etks = getEtkHelper().getPCDH_ETK();
            byte[] Request = SealedProcessor.createSealedSync(etks, xml, "GetDataParams");
            sealedRequestType.setRequestParametersSealed(Request);
            LOG.debug("******************** GETDATA - add authorization parameters  **********************");
            sealedRequestType.setAuthorizationParameters(RequestCreatorUtils.createPCDHAuthorizationParameters(requestPatientSignature, therapeuticRelationShip, null, patient.getPatientId(), getEncryptionUtils()));

            if (requestPatientSignature) {
                LOG.info("Signature required");
                sealedRequestType.getAuthorizationParameters().getPatientConsent().getPatientSignature().setSignature(Crypto.seal(etks, sealedRequestType.getAuthorizationParameters().getPatientConsent().getPatientSignature().getSignature()));
                LOG.info("Message signed");
            }

            LOG.info("******************** GETDATA - CALL PCDH **********************");
            SealedResponseType sealedResponseType = PcdhServiceImpl.getInstance().getData(sealedRequestType);

            LOG.info("******************** GETDATA - VALIDATE RESPONSE PCDH  **********************");
            validateResponse(sealedResponseType);
            LOG.info("******************** GETDATA - UNSEAL SingleMessage response **********************");
            byte[] singleMessageSealed = sealedResponseType.getSingleMessageSealed();
            byte[] singleMessage = Crypto.unseal(singleMessageSealed);

            LOG.info("******************** GETDATA - VALIDATE SingleMessage response **********************");
            xmlValidationhelper.assertValidSingleMessage(singleMessage);
            LOG.debug("******************** GETDATA - LENGTH singlemessagee SEALED client side: " + singleMessageSealed.length);
            LogsSMC.logDguidSguid(singleMessage, IntermediateMessage.UNSEAL, this.getClass());

            StatusType status = sealedResponseType.getStatus();
            SingleMessage sm = null;
            if (status != null) {
                if (localData != null && localData.length > 0) {
                    LOG.debug("****************** GETDATA - Consolidation **********************");
                    LOG.debug("****************** GETDATA - LOCAL DATA: " + Arrays.toString(localData) + " *****************");
                    sm = ConsolidateUtils.createConsolidatedGetDataResponse(singleMessage, localData, patientId);
                } else {
                    LOG.debug("****************** GETDATA - NO LOCAL DATA: *****************");
                    LOG.debug("****************** GETDATA - RETURN STRING REPRESENTATION OF RESPONSE BEFORE UNSEALING PRODUCTS: " + new String(singleMessage) + " *****************");
                    sm = getJaxContextCentralizer().toObject(SingleMessage.class, singleMessage);
                }
                SingleMessageWrapper wrapper = new SingleMessageWrapper(sm);
                LogsSMC.logDguidSguid(wrapper, IntermediateMessage.CONSOLIDATED, this.getClass());
                convertEncryptedContentToProduct(wrapper);
                String updatedSm = getJaxContextCentralizer().toXml(SingleMessage.class, wrapper.getWrappedMessage());

                return updatedSm;
            }
        } catch (Throwable t) {
            Exceptionutils.errorHandler(t);
        }
        return response;
    }

    public void convertEncryptedContentToProduct(SingleMessageWrapper wrapper) throws IntegrationModuleException {
        Semaphore availableThread = new Semaphore(threadLimit, true);

        EncryptionToken systemETK = getSystemETKFromCache();
        List<String> uniqueEncryptionKeys = wrapper.getAllUniqueMedicationHistoryEntriesEncryptionIds();

        LOG.info("uniqueEncryptionKeys.size = " + uniqueEncryptionKeys.size());
        loadEncryptionKeysToCache(uniqueEncryptionKeys, systemETK);

        // PerfLog.logExternalSguid(LOG, "PCDHIntegrationModuleImpl",
        // "CALLKGSS", start, end, wrapper);

        LOG.debug("******************* convertEncryptedContentToProduct ***********************");
        if (uniqueEncryptionKeys.size() > 0) {

            for (MedicationHistoryType medicationHistoryType : wrapper.getAllMedicationHistoryEntries()) {
                ProductDecryptorThread productDecryptorThread = new ProductDecryptorThread(availableThread, medicationHistoryType, systemETK.getEncoded(), this, getKgssCache());
                availableThread.acquireUninterruptibly();
                productDecryptorThread.start();
            }
            try {
                availableThread.acquireUninterruptibly(threadLimit);
            } catch (IllegalArgumentException e) {
                LOG.debug("Incorrect Thread configuration : " + e);
            }
        }
        LOG.debug("******************* convertEncryptedContentToProduct DONE ***********************");
    }


    private EncryptionToken getSystemETKFromCache() throws IntegrationModuleException {
        EncryptionToken systemETK = null;
        if (getEtkCache().get("SystemETK") == null) {
            systemETK = getEtkHelper().getSystemETK().get(0);
            Element element = new Element("SystemETK", systemETK);
            getEtkCache().put(element);
        } else {
            systemETK = (EncryptionToken) getEtkCache().get("SystemETK").getObjectValue();
        }
        return systemETK;
    }

    /**
     * Loads the encryptionKeys with the given Ids into the KGSS cache
     *
     * @param encryptionKeyIds
     * @param systemETK
     * @throws IntegrationModuleException
     */
    private void loadEncryptionKeysToCache(final List<String> encryptionKeyIds, final EncryptionToken systemETK) throws IntegrationModuleException {
        LOG.info(String.format("loading [{0}] encryption keys to cache.", encryptionKeyIds.size()));
        final int maxNumberOfThreads = Math.min(threadLimit, encryptionKeyIds.size());
        Semaphore availableKGSSThread = new Semaphore(maxNumberOfThreads, true);

        boolean errorsOccured = false;
        for (String encryptionId : encryptionKeyIds) {
            if (getKgssCache().get(encryptionId) == null) {
                KGSSThread kgssThread = new KGSSThread(availableKGSSThread, getKgssCache(), encryptionId, systemETK.getEncoded(), this);
                availableKGSSThread.acquireUninterruptibly();
                kgssThread.start();
                if (kgssThread.isErrorOccured()) {
                    errorsOccured = true;
                }
            }
        }

        try {
            availableKGSSThread.acquireUninterruptibly(maxNumberOfThreads);
            if (errorsOccured) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.validation.kgss.encryption.keys.not.equal"));
            }
        } catch (IllegalArgumentException e) {
            LOG.debug("Incorrect Thread configuration : " + e);
        }

    }


    @Override
    @Profiled(logFailuresSeparately = true, tag = "0.PCDHIntegrationModuleImpl#getDataTypes", logger = "org.perf4j.TimingLogger_Executor")
    public String getDataTypes(String patientIdType, String patientId, boolean requestPatientSignature, ThreeStateBoolean therapeuticRelationShip) throws IntegrationModuleException {
        ApplicationConfig.getInstance().assertValidSession();

        LOG.debug("******************** GETDATATYPES - Validate incoming fields **********************");
        ValidationUtils.validateIncomingFieldsGetDataTypes(patientId, patientIdType);
        ValidationUtils.validatePatientId(patientId);
        byte[] singleMessage = null;

        try {
            LOG.debug("******************** GETDATATYPES -  Create GetDataTypes request  **********************");
            SealedRequestType sealedRequestType = new SealedRequestType();
            PatientType patient = RequestCreatorUtils.createPatientType(patientId, patientIdType);

            LOG.debug("******************** GETDATATYPES - Get requestor information from SAML  **********************");
            String requestorId = StandaloneRequestorProvider.getRequestorIdInformation();
            String requestorType = StandaloneRequestorProvider.getRequestorTypeInformation();

            String xml = RequestCreatorUtils.createGetDataTypesParams(patient, requestorId, requestorType);

            LOG.info("******************** GETDATATYPES -  Seal GetData request xml  **********************");
            List<EncryptionToken> etks = getEtkHelper().getPCDH_ETK();
            byte[] Request = SealedProcessor.createSealedSync(etks, xml, "GetDataTypesParams");
            sealedRequestType.setRequestParametersSealed(Request);

            sealedRequestType.setAuthorizationParameters(RequestCreatorUtils.createPCDHAuthorizationParameters(requestPatientSignature, therapeuticRelationShip, null, patient.getPatientId(), getEncryptionUtils()));

            if (requestPatientSignature) {
                sealedRequestType.getAuthorizationParameters().getPatientConsent().getPatientSignature().setSignature(Crypto.seal(etks, sealedRequestType.getAuthorizationParameters().getPatientConsent().getPatientSignature().getSignature()));
            }

            LOG.info("******************** GETDATATYPES - CALL PCDH  **********************");
            SealedResponseType sealedResponseType = null;
            sealedResponseType = PcdhServiceImpl.getInstance().getDataTypes(sealedRequestType);

            LOG.info("******************** GETDATATYPES - VALIDATE RESPONSE PCDH  **********************");
            validateResponse(sealedResponseType);
            LOG.info("******************** GETDATATYPES - UNSEAL SingleMessage response **********************");
            byte[] singleMessageSealed = sealedResponseType.getSingleMessageSealed();

            singleMessage = Crypto.unseal(singleMessageSealed);
            xmlValidationhelper.assertValidSingleMessage(singleMessage);

            SingleMessage smo = getJaxContextCentralizer().toObject(SingleMessage.class, singleMessage);
            ValidationUtils.validateDataTypesResult(smo);

        } catch (Throwable t) {
            Exceptionutils.errorHandler(t);
        }

        return new String(singleMessage);

    }

    private void validateResponse(SealedResponseType sealedResponseType) throws IntegrationModuleException {
        if (sealedResponseType != null) {
            StatusType status = sealedResponseType.getStatus();
            if (status == null) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.technical.status.null"));
            }
            String message = null;

            if (!status.getCode().equals("200") && !status.getCode().equals(StatusCode.PCDH_SUCCESS.toString())) {
                for (be.ehealth.apb.gfddpp.services.pcdh.LocalisedString localisedString : status.getMessage()) {
                    // message += "\n" + localisedString.getValue();
                    LOG.debug("******************** STATUS MESSAGE: " + message + " **********************");

                    String lang = localisedString.getLang().toString();
                    // if (lang.equals("NL")) {
                    // lang = "nl_BE";
                    // }
                    System.out.println("Language default :" + Locale.getDefault().toString());
                    if (Locale.getDefault().toString().contains(lang.toLowerCase())) {
                        throw new IntegrationModuleException("[" + status.getCode() + "] " + localisedString.getValue());
                    }
                }
            }

            LOG.debug("******************** STATUS CODE " + status.getCode() + " **********************");
            // if (status.getCode().equals("301")) {
            // throw new
            // IntegrationModuleException(I18nHelper.getLabel("error.invalid.dguid.response"));
            // } else if (status.getCode().equals("100")) {
            // throw new
            // IntegrationModuleException(I18nHelper.getLabel("error.technical.status.100"));
            // } else if (!status.getCode().equals("200") &&
            // !status.getCode().equals(StatusCode.PCDH_SUCCESS.toString())) {
            // throw new
            // IntegrationModuleException(I18nHelper.getLabel("error.technical.status.invalid"));
            // }

        } else {
            throw new IntegrationModuleException(I18nHelper.getLabel("error.technical.response.null"));
        }
    }

    private void initThread() {
        final PropertyHandler propertyHandler = getPropertyHandler();
        if (propertyHandler != null && propertyHandler.hasProperty(THREAD_PROPERTY)) {
            int threadLimitTmp = propertyHandler.getIntegerProperty(THREAD_PROPERTY);
            if (threadLimitTmp > MAX_THREAD) {
                threadLimit = MAX_THREAD;
                LOG.debug("The " + THREAD_PROPERTY + " is higher than what is tolerated, the maximum value will be used instead. The maximum value is :" + MAX_THREAD);
            } else {
                threadLimit = threadLimitTmp;
            }
        }
        LOG.debug("A maximum of " + threadLimit + "has been defined ");
    }


}
