package be.business.connector.gfddpp.tipsystem.offline.queue;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.security.PublicKey;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Queue;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.bouncycastle.util.encoders.Base64;

import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.utils.EncryptionUtils;
import be.business.connector.core.utils.PropertyHandler;

/**
 * Scheduler & Queue Pattern: A scheduled daemon will send the messages from the
 * queue to the TIP System. The first in, first out acronym describes the
 * principle of a queue processing technique or servicing conflicting demands by
 * ordering process by first-come, first-served (FCFS or FIFO) behavior: where
 * the messages leave the queue in the order they are put on the queue.
 */
public class MessageQueue<T extends Serializable> implements Queue<T> {
	
	private static final String MESSAGE_QUEUE_FOLDER_PROPERTY_NAME = "MESSAGE_QUEUE_FOLDER";
	private static final SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddHHmmssSSS");
	private static final String FILE_NAME_SEPARATOR = "_";
	public static final String FILE_LOCK_SUFFIX = "_LOCK";

	private final static Logger LOG = Logger.getLogger(MessageQueue.class);

	private String folder;
	private File queueFolder;
	private int fileCounter = 0;
	private PropertyHandler property;
	private EncryptionUtils encryptionUtils;
	private PublicKey publicKey;

	public MessageQueue(PropertyHandler propertyHandler, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		property = propertyHandler;
		this.encryptionUtils = encryptionUtils;
		initialize();
	}

	private void initialize() throws IntegrationModuleException {
		if (property.hasProperty(MESSAGE_QUEUE_FOLDER_PROPERTY_NAME)) {
			setMessageQueuePath(property.getProperty(MESSAGE_QUEUE_FOLDER_PROPERTY_NAME));
		} else {
			throw new RuntimeException("MESSAGE_QUEUE_FOLDER not found in properties file !");
		}
		queueFolder = new File(getMessageQueuePath());
		publicKey = encryptionUtils.getPublicKey();
		// create folder if it doesn't exist
		boolean exists = (queueFolder).exists();
		if (!exists) {
			boolean created = queueFolder.mkdirs();
			if (!created)
				throw new RuntimeException("Not able to create folder " + getMessageQueuePath());
		}
	}

	private void setMessageQueuePath(String folder) {
		this.folder = folder;
	}

	@Override
	public boolean isEmpty() {
		return ArrayUtils.isEmpty(queueFolder.list());
	}

	@Override
	public boolean add(T e) {
		return offer(e);
	}

	@Override
	public boolean offer(T e) {
		return save(e);
	}

	@Override
	public T remove() {
		if (!isEmpty()) {
			File first = getFirstFile();
			T result = load(first);
			boolean deleted = first.delete();
			assert deleted;
			return result;
		} else
			throw new NoSuchElementException();
	}

	@Override
	public T poll() {
		if (!isEmpty())
			return remove();
		else
			return null;
	}

	@Override
	public T element() {
		if (!isEmpty())
			return load(getFirstFile());
		else
			throw new NoSuchElementException();
	}
	
	public T elementAndLockItem(){
		if (!isEmpty())
			return load(getFirstFileAndLock());
		else
			throw new NoSuchElementException();
	}

	@Override
	public T peek() {
		if (!isEmpty())
			return load(getFirstFile());
		else
			return null;
	}

	@Override
	public int size() {
		return queueFolder.list().length;
	}

	/**
	 * Not supported
	 */
	@Override
	public boolean contains(Object o) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public boolean removeAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public boolean retainAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public void clear() {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public boolean containsAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public Iterator<T> iterator() {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public Object[] toArray() {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@Override
	public boolean remove(Object o) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Not supported
	 */
	@SuppressWarnings("hiding")
	@Override
	public <T> T[] toArray(T[] a) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean addAll(Collection<? extends T> c) {
		boolean result = true;
		for (T element : c)
			result &= offer(element);
		return result;
	}

	/**
	 * <ol>
	 * <li>Serialize object to a <code>byte[]</code></li>
	 * <li>Seal the <code>byte[]</code> into a sealed <code>byte[]</code></li>
	 * <li>Save the sealed <code>byte[]</code> on disk
	 * </ol>
	 * 
	 * @param clearObject
	 * @return
	 */
	private boolean save(T clearObject) {
		byte[] serialiazed = serialize(clearObject);
		byte[] sealedObject = null;
		FileOutputStream saveFile = null;
		ObjectOutputStream save = null;
		try {
			sealedObject = seal(serialiazed);
		} catch (IntegrationModuleException e) {
			LOG.info(e.getLocalizedMessage());
			return false;
		}

		try {
			 saveFile = new FileOutputStream(getMessageQueuePath() + File.separator + generateFileName());
			 save = new ObjectOutputStream(saveFile);
			save.writeObject(sealedObject);
			save.close();
			saveFile.close();
			return true;
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
		finally{
			
			try{
				if (null !=save ) save.close();
				if (null !=saveFile )saveFile.close();
			} catch (IOException e){
				LOG.error("Can not save object " + clearObject.toString() , e );;
			}
	}
	}

	/**
	 * <ol>
	 * <li>Load a sealed <code>byte[]</code> from disk</li>
	 * <li>Unseal the <code>byte[]</code></li>
	 * <li>Unserialized the unsealed <code>byte[]</code> to a
	 * <code>Serializable</code> object
	 * </ol>
	 * 
	 * @param file
	 * @return
	 */
	private T load(File file) {
		FileInputStream saveFile =null;
		ObjectInputStream save = null;
		
		try {
			 saveFile = new FileInputStream(file);
			 save = new ObjectInputStream(saveFile);
			
			byte[] sealedObject = (byte[]) save.readObject();
			byte[] serializedObject = unseal(sealedObject);
			
			T result = unserialize(serializedObject);
			return result;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return null;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			return null;
		} catch (IntegrationModuleException e) {
			LOG.info(e.getLocalizedMessage());
			e.printStackTrace();
			return null;
		} 
		finally{
			
				try{
					if (null !=save ) save.close();
					if (null !=saveFile )saveFile.close();
				} catch (IOException e){
					LOG.error("Can not load file" + file.getAbsolutePath(), e );;
				}
		}
	}
	
	
	private File getFirstFileAndLock(){
		String fileName = getFirstFileName();
		if(StringUtils.isEmpty(fileName)){
			//TODO: DENNIS: throw error?
		}
			
		return lockFile(new File(getMessageQueuePath() + File.separator + fileName));
	}
	
	private T lockAndLoad(File file) {
		FileInputStream saveFile =null;
		ObjectInputStream save = null;
		File newFile = null;

		try {
			newFile = new File( file.getAbsolutePath() + FILE_LOCK_SUFFIX);
			boolean successfullyRenamed = file.renameTo(newFile);
			
			if(!successfullyRenamed){
				throw new IOException("Rename failed");
			}
			
			 saveFile = new FileInputStream(newFile);
			 save = new ObjectInputStream(saveFile);

			byte[] sealedObject = (byte[]) save.readObject();

			save.close();
			saveFile.close();

			byte[] serializedObject = unseal(sealedObject);
			T result = unserialize(serializedObject);
			return result;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return null;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			return null;
		} catch (IntegrationModuleException e) {
			LOG.info(e.getLocalizedMessage());
			e.printStackTrace();
			return null;
		}
		finally{
			
			try{
				if (null !=save ) save.close();
				if (null !=saveFile )saveFile.close();
				
			} catch (IOException e){
				LOG.error("Can not load or lock file " + file.getAbsolutePath(), e );
			}
	}
	}

	/**
	 * Transform a clear object into a <code>byte[]</code>
	 * 
	 * @param clearObject
	 * @return
	 */
	public byte[] serialize(T clearObject) {
		ByteArrayOutputStream bos = null;
		ObjectOutput out = null;
		try {
			bos = new ByteArrayOutputStream();
			out = new ObjectOutputStream(bos);
			out.writeObject(clearObject);
			byte[] yourBytes = bos.toByteArray();
			out.close();
			bos.close();
			//System.out.println(new String(yourBytes));TODO prevention to appearing raw data in logs
			byte[] encodedBytes = Base64.encode(yourBytes);
			return encodedBytes;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		} finally {
			try{
			if ( null !=bos) bos.close();
			if (null!= out) out.close();
		} catch (IOException e){
			LOG.error("error serializing " + clearObject.toString() , e );
		}
	}//finally
		
	}

	/**
	 * Transform a unsealed <code>byte[]</code> into an
	 * <code>Serializable</code> object
	 * 
	 * @param serializedObject
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public T unserialize(byte[] serializedObject) {
		try {
			byte[] decodedBytes = Base64.decode(serializedObject);

			ByteArrayInputStream bis = new ByteArrayInputStream(decodedBytes);
			ObjectInput in = new ObjectInputStream(bis);
			Object o = in.readObject();
			bis.close();
			in.close();
			return (T) o;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			return null;
		}
	}

	private byte[] seal(byte[] clearObject) throws IntegrationModuleException {
		byte[] sealedObject = null;
		sealedObject = encryptionUtils.queueEncrypt(clearObject, publicKey);
		return sealedObject;
	}

	private byte[] unseal(byte[] sealedObject) throws IntegrationModuleException {
		String rslt = encryptionUtils.queueDecrypt(sealedObject, publicKey);
		try {
			return rslt.getBytes("UTF-8");
		} catch (UnsupportedEncodingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	
	private File getFirstFile() {
		String fileName = getFirstFileName();
		if(StringUtils.isEmpty(fileName)){
			//TODO: DENNIS: throw error?
		}
		return new File(getMessageQueuePath() + File.separator + fileName);
	}
	
	private File lockFile(File fileToLock){
		if(StringUtils.endsWith(fileToLock.getAbsolutePath(), FILE_LOCK_SUFFIX)){
			//TODO: DENNIS: throw error that file is already locked?
			return fileToLock;
		}
		File lockedFile = new File(fileToLock.getAbsolutePath() + FILE_LOCK_SUFFIX);
		fileToLock.setLastModified(new Date().getTime());
		boolean successfullyLocked = fileToLock.renameTo(lockedFile);
		
		if(!successfullyLocked){
			LOG.error("Problem with locking file.");
		}else{
			LOG.info("File successfully locked.");
		}
		return lockedFile;
	}
	
	public void unlockFirstFile(){
		if(queueFolder.list().length > 0){
			unlockFile(getFirstFile());
		}else{
			LOG.info("MessageQueue is empty, no file is unlocked");
		}
	}
	
	private void unlockFile(File fileToUnlock){
		if(!StringUtils.endsWith(fileToUnlock.getAbsolutePath(), FILE_LOCK_SUFFIX)){
			LOG.info("File to unlock is already unlocked.");
		}
		
		File unlockedFile = new File(StringUtils.remove(fileToUnlock.getAbsolutePath(), FILE_LOCK_SUFFIX));
		
		boolean successfullyUnlocked = fileToUnlock.renameTo(unlockedFile);
		if(!successfullyUnlocked){
			LOG.error("Problem with unlocking file.");
		}else{
			LOG.info("File successfully unlocked.");
		}
	}
	
	/**
	 * According to API "there is no guarantee that the name strings in the
	 * resulting array will appear in any specific order; they are not, in
	 * particular, guaranteed to appear in alphabetical order." for method
	 * file.list(). So we have to implement
	 * 
	 * @return The filename from the queue
	 */
	public String getFirstFileName(){
		String[] files = queueFolder.list();
		switch (files.length) {
		case 0:
			return null;
		case 1:
			return files[0];
		default:
			String first = files[0];
			int i = 1;
			while (i < files.length) {
				if (files[i].compareTo(first) < 0)
					first = files[i];
				i++;
			}
			return first;
		}
	}
	
	public boolean isFirstItemLocked(){
		if(StringUtils.endsWith(getFirstFileName(), FILE_LOCK_SUFFIX)){
			LOG.debug("First item on queue is locked.");
			return true;
		}
		return false;
	}
	/**
	 * Use with care -- it only an assumption the file can not be decrypted
	 * removes first queue element which can not be decrypted
	 * @return true in case of success 
	 */
	public boolean moveAbnormal(String moveToFolderName ){
		File moveToDir = new File(moveToFolderName);
		if (false == moveToDir.exists() && false==moveToDir.isDirectory()){
			throw new RuntimeException("Move to folder does not exist " + moveToFolderName );
		}
		File first = getFirstFile() ; String firstFileName=first.getName();
		try {
			
			File destFile = new File(moveToDir.getAbsolutePath() + 
					File.separator + firstFileName.substring(0, firstFileName.lastIndexOf(FILE_LOCK_SUFFIX)));
			FileUtils.moveFile(first , destFile ) ;
			
			LOG.debug("copying file to not decryptable dir");
			
			LOG.debug(" removing first file in message queue");
		} catch (IOException e) {
			
			LOG.error("unable to copy unencrypted file" , e);
		} 
		
		return true;
	}
	

	/**
	 * The file name of each message is the data, time and counter value. Format
	 * filename: YYYYMMDDHHmmssSSS_<counter value>
	 * 
	 * @return file name <code>String</code> with required format
	 */
	private String generateFileName() {
		return dateFormat.format(new Date()) + FILE_NAME_SEPARATOR + (fileCounter++);
	}

	public String getMessageQueuePath() {
		return folder;
	}

	
}
