package be.business.connector.gfddpp.tipsystem.status.offline.queue;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.security.PublicKey;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Queue;

import org.apache.log4j.Logger;


import be.apb.gfddpp.common.exceptions.GFDDPPException;
import be.apb.gfddpp.common.utils.JaxContextCentralizer;
import be.apb.standards.smoa.schema.model.v1.StatusMessageType;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.utils.EncryptionUtils;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.core.utils.PropertyHandler;

public class StatusMessageQueue implements Queue<StatusMessageType> {
	private static final String STATUS_MESSAGE_QUEUE_FOLDER_PROPERTY_NAME = "STATUS_MESSAGE_QUEUE_FOLDER";
	private static final String FILE_NAME_SEPARATOR = "_";
	private static final SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddHHmmssSSS");

	private String folder;
	private File queueFolder;
	private int fileCounter = 0;
	private PropertyHandler property;
	private EncryptionUtils encryptionUtils;
	private PublicKey publicKey;
	private JaxContextCentralizer jaxContext = JaxContextCentralizer.getInstance();
	
	private final static Logger LOG = Logger.getLogger(StatusMessageQueue.class);

	public StatusMessageQueue() {

	}

	public StatusMessageQueue(PropertyHandler propertyHandler, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		property = propertyHandler;
		this.encryptionUtils = encryptionUtils;

		try {
			initialize();
		} catch (Exception e) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.creation.statusMessageQueue"), e);
		}
	}

	private void initialize() throws IntegrationModuleException {
		if (property.hasProperty(STATUS_MESSAGE_QUEUE_FOLDER_PROPERTY_NAME)) {
			setStatusMessageQueuePath(property.getProperty(STATUS_MESSAGE_QUEUE_FOLDER_PROPERTY_NAME));
		} else {
			throw new RuntimeException("MESSAGE_QUEUE_FOLDER not found in properties file !");
		}
		queueFolder = new File(getStatusMessageQueuePath());
		publicKey = encryptionUtils.getPublicKey();
		// create folder if it doesn't exist
		boolean exists = (queueFolder).exists();
		if (!exists) {
			boolean created = queueFolder.mkdirs();
			if (!created)
				throw new RuntimeException("Not able to create folder " + getStatusMessageQueuePath());
		}
	}

	private void setStatusMessageQueuePath(String folder) {
		this.folder = folder;
	}

	public String getStatusMessageQueuePath() {
		return folder;
	}

	@Override
	public int size() {
		return queueFolder.list().length;
	}

	@Override
	public boolean isEmpty() {
		FileFilter directoryFilter = new FileFilter() {
			public boolean accept(File file) {
				return !file.isDirectory();
			}
		};
		return new ArrayList<File>(Arrays.asList(queueFolder.listFiles(directoryFilter))).isEmpty();
	}

	@Override
	public boolean contains(Object o) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Iterator<StatusMessageType> iterator() {
		throw new UnsupportedOperationException();
	}

	@Override
	public Object[] toArray() {
		throw new UnsupportedOperationException();
	}

	@Override
	public <T> T[] toArray(T[] a) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean remove(Object o) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean addAll(Collection<? extends StatusMessageType> c) {
		boolean result = true;
		for (StatusMessageType element : c)
			result &= offer(element);
		return result;
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean retainAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void clear() {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean add(StatusMessageType statusMessage) {
		byte[] serialiazed;
		try {
			serialiazed = jaxContext.toXml(StatusMessageType.class,statusMessage).getBytes();
		} catch (GFDDPPException e) {
			LOG.info(e.getLocalizedMessage());
			return false;
		}
		byte[] sealedObject = null;
		try {
			sealedObject = seal(serialiazed);
		} catch (IntegrationModuleException e) {
			LOG.info(e.getLocalizedMessage());
			return false;
		}
		try {
			FileOutputStream saveFile = new FileOutputStream(getStatusMessageQueuePath() + File.separator + generateFileName());
			ObjectOutputStream save = new ObjectOutputStream(saveFile);
			save.writeObject(sealedObject);
			save.flush();
			save.close();
			saveFile.close();
			return true;
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
	}

	@Override
	public boolean offer(StatusMessageType e) {
		return add(e);
	}

	@Override
	public StatusMessageType remove() {
		if (!isEmpty()) {
			File first = getFirst();
			StatusMessageType result = load(first);
			boolean deleted = first.delete();
			assert deleted;
			return result;
		}
		throw new NoSuchElementException();
	}

	@Override
	public StatusMessageType poll() {
		if (!isEmpty()) {
			return remove();
		}
		return null;
	}

	@Override
	public StatusMessageType element() {
		//  Auto-generated method stub
		return null;
	}

	@Override
	public StatusMessageType peek() {
		if (!isEmpty()) {
			return load(getFirst());
		}
		return null;
	}

	public String getMessageQueuePath() {
		return folder;
	}

	private String generateFileName() {
		return dateFormat.format(new Date()) + FILE_NAME_SEPARATOR + (fileCounter++);
	}

	private File getFirst() {
		FileFilter directoryFilter = new FileFilter() {
			public boolean accept(File file) {
				return !file.isDirectory();
			}
		};
		
		File[] files = queueFolder.listFiles(directoryFilter);
		switch (files.length) {
		case 0:
			return null;
		case 1:
			return files[0];
		default:
			File first = files[0];
			int i = 1;
			while (i < files.length) {
				if (files[i].compareTo(first) < 0)
					first = files[i];
				i++;
			}
			return first;
		}
	}

	private StatusMessageType load(File file) {
		try {
			FileInputStream saveFile = new FileInputStream(file);
			ObjectInputStream save = new ObjectInputStream(saveFile);
			byte[] sealedObject = null;
			try {
				sealedObject = (byte[]) save.readObject();
			} catch (IOException e) {
				e.printStackTrace();
				LOG.error("IOException: the file:" + file.getName() + "will be deleted\n" + e.getMessage());
				save.close();
				saveFile.close();
				file.delete();
				return null;
			}
			save.close();
			saveFile.close();

			byte[] serializedObject = unseal(sealedObject);
			StatusMessageType result = jaxContext.toObject(StatusMessageType.class, serializedObject);
			return result;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return null;
		} catch (IOException e) {
			e.printStackTrace();
			LOG.error("IOException: the file:" + file.getName() + "will be deleted\n" + e.getMessage());
			return null;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			return null;
		} catch (IntegrationModuleException e) {
			e.printStackTrace();
			return null;
		} catch (GFDDPPException e) {
			e.printStackTrace();
			return null;
		}
	}

	private byte[] unseal(byte[] sealedObject) throws IntegrationModuleException, GFDDPPException {
		try {
			return encryptionUtils.queueDecrypt(sealedObject, publicKey).getBytes("UTF-8");
		} catch (UnsupportedEncodingException e) {
			LOG.error("UnsupportedEncodingException in unseal method", e);
			throw new GFDDPPException("UnsupportedEncodingException in unseal method");
		}
	}

	private byte[] getBytes(InputStream inputStream) {
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			int ch;
			while ((ch = inputStream.read()) >= 0)
				baos.write(ch);
			byte[] unsealedData = baos.toByteArray();
			return unsealedData;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Transform a unsealed <code>byte[]</code> into an
	 * <code>Serializable</code> object
	 * 
	 * @param serializedObject
	 * @return
	 */
//	@SuppressWarnings("unchecked")
//	private StatusMessageType unserialize(byte[] serializedObject) {
//		try {
//			ByteArrayInputStream bis = new ByteArrayInputStream(serializedObject);
//			ObjectInput in = new ObjectInputStream(bis);
//			Object o = in.readObject();
//			bis.close();
//			in.close();
//			return (StatusMessageType) o;
//		} catch (IOException e) {
//			e.printStackTrace();
//			return null;
//		} catch (ClassNotFoundException e) {
//			e.printStackTrace();
//			return null;
//		}
//	}

	private byte[] seal(byte[] clearObject) throws IntegrationModuleException {
		byte[] sealedObject = null;
		sealedObject = encryptionUtils.queueEncrypt(clearObject, publicKey);
		return sealedObject;
	}

//	private byte[] serialize(StatusMessageType clearObject) {
//		try {
//			ByteArrayOutputStream bos = new ByteArrayOutputStream();
//			ObjectOutput out = new ObjectOutputStream(bos);
//			out.writeObject(clearObject); 
//			System.out.println("BOS SIZE : "+ bos.size());
//			byte[] yourBytes = bos.toByteArray();
//			out.close();
//			bos.close();
//			return yourBytes;
//		} catch (IOException e) {
//			e.printStackTrace();
//			return null;
//		}
//	}

}
