package be.business.connector.gfddpp.utils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.datatype.XMLGregorianCalendar;

import org.apache.log4j.Logger;


import be.apb.gfddpp.common.exceptions.GFDDPPException;
import be.apb.gfddpp.common.utils.JaxContextCentralizer;
import be.apb.gfddpp.common.utils.SingleMessageWrapper;
import be.apb.gfddpp.helper.SingleMessageValidationHelper;
import be.apb.gfddpp.validation.exception.SingleMessageValidationException;
import be.apb.standards.smoa.schema.id.v1.EntityIdType;
import be.apb.standards.smoa.schema.model.v1.MedicationHistoryType;
import be.apb.standards.smoa.schema.v1.AbstractEventType;
import be.apb.standards.smoa.schema.v1.MedicationHistoryEvent;
import be.apb.standards.smoa.schema.v1.SingleMessage;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.projects.common.utils.ValidationUtils;

public class ConsolidateUtils {
	private final static Logger LOG = Logger.getLogger(ConsolidateUtils.class);

	private static JaxContextCentralizer jaxContextCentralizer = JaxContextCentralizer.getInstance();

	public static SingleMessage createConsolidatedGetDataResponse(byte[] response, byte[] local, String patientid) throws IntegrationModuleException, SingleMessageValidationException {
	
		SingleMessage singleMessageResponse = null;
		try {
			singleMessageResponse = jaxContextCentralizer.toObject(SingleMessage.class, response);
		} catch (GFDDPPException e) {
			throw new IntegrationModuleException(e.getLocalizedMessage(),e);
		}
		SingleMessageWrapper singleMessageWrapperResponse = new SingleMessageWrapper(singleMessageResponse);

		SingleMessageInfo smilocal = new SingleMessageInfo(local);
		smilocal.validateEmptyHeader();
		smilocal.validateSGuidFilledOk();
		smilocal.validateDGuidFilledOk();
		smilocal.validateMedicationHistorySessionDateFilledOk();
		smilocal.validateMedicationHistoryPatient(patientid);
		if (!smilocal.getHasMedicationHistoryEvent()) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.consolidation.local.message.minimal.dataset"));
		}
		
		SingleMessageWrapper singleMessageWrapperLocal = smilocal.getWrapper();
		if (singleMessageWrapperLocal.getAllEventsOfType(MedicationHistoryEvent.class).size() == 0) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.consolidation.local.message.minimal.dataset"));
		}
		
		SingleMessage singleMessageLocal;
		try {
			singleMessageLocal = jaxContextCentralizer.toObject(SingleMessage.class, local);
		} catch (GFDDPPException e) {
			throw new IntegrationModuleException(e.getLocalizedMessage(),e);
		}
		
		ValidationUtils.validateSingleMessagePatient(patientid, singleMessageLocal);
		if (singleMessageWrapperLocal.getAllEventsOfType(MedicationHistoryEvent.class).size() == 0) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.consolidation.local.message.minimal.dataset"));
		}

		LOG.debug("******************** Merging (dguid) **********************");
		Map<String, MedicationHistoryType> mergedMedicationHistoryType = merging(singleMessageWrapperResponse, singleMessageWrapperLocal);

		LOG.debug("******************** Remove all old getMedicationHistories & add new getMedicationHistories**********************");
		// if more then one medicationHistoryEvent is returned exception will be thrown because our code doesn't support it yet.
		List<MedicationHistoryEvent> medicationHistoryEvents = singleMessageWrapperResponse.getAllEventsOfType(MedicationHistoryEvent.class);
		if (medicationHistoryEvents.size() != 1) {
			if(medicationHistoryEvents.size() > 1 ){
				throw new IntegrationModuleException(I18nHelper.getLabel("to.much.medicalhistoryevents.found"));
			}else{ //medicationHistoryEvents == 0 at this point
				List<MedicationHistoryEvent> localMedicationHistoryEvents = singleMessageWrapperLocal.getAllEventsOfType(MedicationHistoryEvent.class);
				if(localMedicationHistoryEvents.size() != 1){
					if(localMedicationHistoryEvents.size() > 1 ){
						throw new IntegrationModuleException(I18nHelper.getLabel("too.much.localmedicalhistoryevents.found")); 
					}else{ //localMedicationHistoryEvents == 0
						throw new IntegrationModuleException(I18nHelper.getLabel("no.localmedicalhistoryevents.and.medicalhistoryevents.found")); 
					}
				}
			}
		}
		
		//sort the mergedMedicationHistoryTypes before adding to the response
		List<MedicationHistoryType> mergedMedicationHistoryTypes = new ArrayList<MedicationHistoryType>( mergedMedicationHistoryType.values());
		Collections.sort(mergedMedicationHistoryTypes, createComparatorMedicationHistoryType());

		
		List<AbstractEventType> abstractEventTypes = singleMessageWrapperResponse.getEvents();
		boolean abstractEventTypesContainsMedicationHistoryEvent = false;
		for(AbstractEventType abstractEventType : abstractEventTypes){
			if(abstractEventType instanceof MedicationHistoryEvent){
				abstractEventTypesContainsMedicationHistoryEvent = true;
				((MedicationHistoryEvent) abstractEventType).getMedicationHistoryEntity().clear();
				((MedicationHistoryEvent) abstractEventType).getMedicationHistoryEntity().addAll(mergedMedicationHistoryTypes);
			}
		}
		if(!abstractEventTypesContainsMedicationHistoryEvent){
			MedicationHistoryEvent medicationHistoryEvent = new MedicationHistoryEvent();
			medicationHistoryEvent.getMedicationHistoryEntity().addAll(mergedMedicationHistoryTypes);
			abstractEventTypes.add(medicationHistoryEvent);
		}
		
		SingleMessageValidationHelper helper = new SingleMessageValidationHelper();
		
		long startValidate = System.nanoTime();
		try {
			helper.assertValidSingleMessage(jaxContextCentralizer.toXml(SingleMessage.class, singleMessageResponse).getBytes());
		} catch (GFDDPPException e) {
			throw new IntegrationModuleException(e.getLocalizedMessage(),e);
		}
		long endValidate = System.nanoTime();
		LOG.info("//PERF EXTERNAL - CONSOLIDATEUTILS - VALIDATE MERGE - START: " + startValidate + " - END: " + endValidate + " - DURATION: "+ (endValidate-startValidate));
		
		return singleMessageResponse;
	}

	private static Comparator<MedicationHistoryType> createComparatorMedicationHistoryType() {
		return new Comparator<MedicationHistoryType>() {
			public int compare(MedicationHistoryType o1, MedicationHistoryType o2) {
				XMLGregorianCalendar date1 = extractDate(o1);
				XMLGregorianCalendar date2 = extractDate(o2);
				return date1.compare(date2);
			}

			private XMLGregorianCalendar extractDate(MedicationHistoryType medicationHistoryType) {
				return medicationHistoryType.getDeliveryDate();
			}
		};
	}

	/*
	 * ATTENTION - The localdata has to be always of the type minimal dataset!!!
	 */
	private static Map<String, MedicationHistoryType> merging(SingleMessageWrapper singleMessageWrapperResponse, SingleMessageWrapper singleMessageWrapperLocal) throws IntegrationModuleException {
		Map<String, MedicationHistoryType> result = new HashMap<String, MedicationHistoryType>();

		//		List<String> responseDguids = new ArrayList<String>();
		for (MedicationHistoryType medicationHistoryType : singleMessageWrapperResponse.getAllMedicationHistoryEntries()) {
			String dGuid = ((EntityIdType) medicationHistoryType.getEntityId()).getId();
			//responseDguids.add(dGuid);
			result.put(dGuid, medicationHistoryType);
		}
		if (singleMessageWrapperLocal.getWrappedMessage() != null) {
			for (MedicationHistoryType medicationHistoryType : singleMessageWrapperLocal.getAllMedicationHistoryEntries()) {
				String dGuid = ((EntityIdType) medicationHistoryType.getEntityId()).getId();
				result.put(dGuid, medicationHistoryType);
				//responseDguids.remove(dGuid);
			}
		}
		return result;
	}
}
