package be.business.connector.gfddpp.utils;

import be.business.connector.projects.common.utils.INSZUtils;
import org.apache.commons.lang3.StringUtils;

/**
 * @author Liesje Demuynck.
 */
public class DataEntryURI {


    private enum Purpose {CREATE, UPDATE, READ_ALL, READ;}

    private static final String SUBJECT_TAG = "subject";
    private static final String MEDICATION_SCHEME_TAG = "medication-scheme";
    private static final String NEW_TAG = "new";

    public static final String READ_URI_FORMAT = "/subject/%s/medication-scheme/%s/%s";
    public static final String READ_ALL_URI_FORMAT = "/subject/%s/medication-scheme";
    public static final String UPDATE_URI_FORMAT = "/subject/%s/medication-scheme/%s/new/%s";
    public static final String NEW_URI_FORMAT = "/subject/%s/medication-scheme/new";

    private boolean valid;
    private Purpose purpose;
    private String subjectId;
    private String dataEntryId;
    private int dataEntryVersion;
    private String fullURI;


    public DataEntryURI(String URI) {
        fullURI = URI;
        valid = false;

        if (!StringUtils.isBlank(URI)) {
            final String[] tokens = StringUtils.split(URI, "/");
            switch (tokens.length) {
                case 3: // /subject/<ssin>/medication-scheme
                    if (SUBJECT_TAG.equals(tokens[0])
                            && INSZUtils.isValidINSZ(tokens[1])
                            && MEDICATION_SCHEME_TAG.equals(tokens[2])) {
                        valid = true;
                        subjectId = tokens[1];
                        purpose = Purpose.READ_ALL;
                    }
                    break;
                case 4: // /subject/<ssin>/medication-scheme/new
                    if (SUBJECT_TAG.equals(tokens[0])
                            && INSZUtils.isValidINSZ(tokens[1])
                            && MEDICATION_SCHEME_TAG.equals(tokens[2])
                            && NEW_TAG.equals(tokens[3])) {
                        valid = true;
                        subjectId = tokens[1];
                        purpose = Purpose.CREATE;
                    }
                    break;
                case 5: // /subject/<ssin>/medication-scheme/<dataEntryId>/<dataEntryVersion>
                    if (SUBJECT_TAG.equals(tokens[0])
                            && INSZUtils.isValidINSZ(tokens[1])
                            && MEDICATION_SCHEME_TAG.equals(tokens[2])
                            && StringUtils.isNotBlank(tokens[3])) {
                        int version;
                        try {
                            version = Integer.valueOf(tokens[4]);
                        } catch (Exception e) {
                            return; // not valid
                        }
                        valid = true;
                        subjectId = tokens[1];
                        dataEntryId = tokens[3];
                        dataEntryVersion = version;
                        purpose = Purpose.READ;
                    }
                    break;

                case 6: // /subject/<ssin>/medication-scheme/<dataEntryId>/new/<dataEntryVersion>
                    if (SUBJECT_TAG.equals(tokens[0])
                            && INSZUtils.isValidINSZ(tokens[1])
                            && MEDICATION_SCHEME_TAG.equals(tokens[2])
                            && StringUtils.isNotBlank(tokens[3])
                            && NEW_TAG.equals(tokens[4])) {
                        int version;
                        try {
                            version = Integer.valueOf(tokens[5]);
                        } catch (Exception e) {
                            return; // not valid
                        }
                        valid = true;
                        subjectId = tokens[1];
                        dataEntryId = tokens[3];
                        dataEntryVersion = version;
                        purpose = Purpose.UPDATE;
                    }
                    break;
            }
        }
    }

    public static DataEntryURI updateURI(final String subjectId, final String dataEntryId, final int dataEntryVersion){
        return new DataEntryURI(String.format(UPDATE_URI_FORMAT, subjectId, dataEntryId, dataEntryVersion));
    }

    public static DataEntryURI newURI(final String subjectId){
        return new DataEntryURI(String.format(NEW_URI_FORMAT, subjectId));
    }

    public static DataEntryURI readURI(final String subjectId, final String dataEntryId, final int dataEntryVersion) {
        return new DataEntryURI(String.format(READ_URI_FORMAT, subjectId, dataEntryId, dataEntryVersion));
    }

    public static DataEntryURI readAllURI(final String subjectId) {
        return new DataEntryURI(String.format(READ_ALL_URI_FORMAT, subjectId));
    }

    public boolean isValid() {
        return valid;
    }

    public boolean isReadAllURI() {
        return isValid() && Purpose.READ_ALL.equals(purpose);
    }

    public boolean isReadURI() {
        return isValid() && Purpose.READ.equals(purpose);
    }

    public boolean isCreateURI() {
        return isValid() && Purpose.CREATE.equals(purpose);
    }

    public boolean isUpdateURI() {
        return isValid() && Purpose.UPDATE.equals(purpose);
    }

    public String getSubjectId() {
        return subjectId;
    }

    public String getDataEntryId() {
        return dataEntryId;
    }

    public int getDataEntryVersion() {
        return dataEntryVersion;
    }


    public String toString() {
        return fullURI;
    }
}
