package be.business.connector.gfddpp.utils;

import be.apb.gfddpp.common.medicationscheme.status.MSStatusCode;
import be.apb.gfddpp.common.medicationscheme.status.MSStatusResolver;
import be.apb.gfddpp.common.medicationscheme.validator.MedicationSchemeValidator;
import be.apb.gfddpp.common.medicationscheme.validator.ValidatorException;
import be.apb.standards.smoa.schema.v1.AvailabilityStatus;
import be.apb.standards.smoa.schema.v1.FormatCode;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.gfddpp.domain.medicationscheme.DataEntry;
import be.business.connector.gfddpp.domain.medicationscheme.Error;
import be.business.connector.gfddpp.domain.medicationscheme.ReferenceType;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.FetchDataEntriesRequest;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.RetrieveTimestampsRequest;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.StoreDataEntriesRequest;
import be.business.connector.projects.common.utils.INSZUtils;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.EnumUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static be.apb.gfddpp.common.medicationscheme.status.MSStatusCode.*;

/**
 * Validation Utils for GFDDPP specific data structures
 *
 * @author Liesje Demuynck.
 */
public class GFDDPPValidationUtils {

    private final static Logger LOG = Logger.getLogger(GFDDPPValidationUtils.class);

    private static final String AVAILABILITY_STATUS_KEY = "availabilityStatus";
    private static final String FORMAT_CODE_KEY = "formatCode";


    private static DocumentBuilder sourceParser;

    /**
     * lazy getter for the source parser
     *
     * @return
     */
    private static DocumentBuilder getSourceParser() throws IntegrationModuleException {
        if (sourceParser == null) {
            try {
                DocumentBuilderFactory pce = DocumentBuilderFactory.newInstance();
                pce.setNamespaceAware(true);
                sourceParser = pce.newDocumentBuilder();
            } catch (ParserConfigurationException e) {
                throw new IntegrationModuleException("XML/XSD parser could not be set up.", e);
            }
        }
        return sourceParser;
    }

    /**
     * Validation method for RetrieveTimestampsRequest objects
     *
     * @param retrieveTimestampsRequest
     * @return
     * @throws IntegrationModuleException
     */
    public static List<Error> validate(final RetrieveTimestampsRequest retrieveTimestampsRequest) {

        final List<Error> errors = new ArrayList<Error>();

        if (retrieveTimestampsRequest == null || retrieveTimestampsRequest.getSubjectIDs().size() <= 0) {
            errors.add(createError(MEDICATION_SCHEME_NO_DATA, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
        } else {

            if (retrieveTimestampsRequest.getSubjectIDs().size() > 10) {
                errors.add(createError(MEDICATION_SCHEME_INVALID_SUBJECTID_COUNT, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1), 0, 10));
            }

            for (String subjectId : retrieveTimestampsRequest.getSubjectIDs()) {
                if (subjectId == null || !INSZUtils.isValidINSZ(subjectId)) {
                    errors.add(createError(MEDICATION_SCHEME_INVALID_INSZ, ReferenceType.SUBJECT_SSIN, subjectId));
                }
            }

            if (retrieveTimestampsRequest.getPersonInformation() != null) {
                errors.add(createError(MEDICATION_SCHEME_PERSON_INFORMATION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));

            }
            if (retrieveTimestampsRequest.getOrganizationInformation() != null) {
                errors.add(createError(MEDICATION_SCHEME_ORGANIZATION_INFORMATION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }
        }

        return errors;
    }


    /**
     * Validation method for FetchDataEntriesRequest objects
     *
     * @param getMedicationEntries the FetchDataEntriesRequest
     * @return
     * @throws IntegrationModuleException
     */
    public static List<Error> validate(final FetchDataEntriesRequest getMedicationEntries) {

        final List<Error> errors = new ArrayList<Error>();

        if (getMedicationEntries == null) {
            errors.add(createError(MEDICATION_SCHEME_NO_DATA, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
        } else {
            final String subjectId = getMedicationEntries.getSubjectID();
            if (StringUtils.isBlank(subjectId)) {
                errors.add(createError(MEDICATION_SCHEME_NO_INSZ, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }
            if(StringUtils.isNotBlank(subjectId) && !INSZUtils.isValidINSZ(subjectId)) {
                errors.add(createError(MEDICATION_SCHEME_INVALID_INSZ, ReferenceType.SUBJECT_SSIN, subjectId));
            }
            if (getMedicationEntries.getVersion() != null) {
                errors.add(createError(MEDICATION_SCHEME_VERSION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }
            if (getMedicationEntries.getBreakTheGlass() != null) {
                errors.add(createError(MEDICATION_SCHEME_BREAK_THE_GLASS_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }
            if (!getMedicationEntries.includeBusinessData()) {
                errors.add(createError(MEDICATION_SCHEME_REQUEST_NO_BUSINESSDATA, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }
            if (getMedicationEntries.getSearchCriteria() != null) {
                errors.add(createError(MEDICATION_SCHEME_SEARCHCRITERIA_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }

            if (getMedicationEntries.getPersonInformation() != null) {
                errors.add(createError(MEDICATION_SCHEME_PERSON_INFORMATION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));

            }
            if (getMedicationEntries.getOrganizationInformation() != null) {
                errors.add(createError(MEDICATION_SCHEME_ORGANIZATION_INFORMATION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            }

            if (getMedicationEntries.getPagination() != null) {
                if (getMedicationEntries.getPagination().getPaginationIndex() <= 0) {
                    errors.add(createError(MEDICATION_SCHEME_INVALID_PAGINATION_INDEX, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
                }
                final DataEntryURI uri = new DataEntryURI(getMedicationEntries.getPagination().getUri());
                if (!uri.isReadAllURI()) {
                    errors.add(createError(MEDICATION_SCHEME_INVALID_FETCH_DATA_URI, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
                }
                if(StringUtils.isNotBlank(subjectId) && !subjectId.equals(uri.getSubjectId())){
                    errors.add(createError(MEDICATION_SCHEME_SUBJECT_ID_MISMATCH, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
                }
            }
        }
        return errors;
    }

    /**
     * Validation method for StoreDataEntriesRequest objects
     *
     * @param storeDataEntriesRequest the StoreDataEntriesRequest
     * @return
     * @throws IntegrationModuleException
     */
    public static List<Error> validate(final StoreDataEntriesRequest storeDataEntriesRequest) throws IntegrationModuleException {

        final List<Error> errors = new ArrayList<Error>();

        if (storeDataEntriesRequest == null) {
            errors.add(createError(MEDICATION_SCHEME_NO_DATA, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
            return errors;
        }

        final String subjectId = storeDataEntriesRequest.getSubjectID();
        if (StringUtils.isBlank(subjectId)){
            errors.add(createError(MEDICATION_SCHEME_NO_INSZ, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
        }
        if(StringUtils.isNotBlank(subjectId) && !INSZUtils.isValidINSZ(subjectId)) {
            errors.add(createError(MEDICATION_SCHEME_INVALID_INSZ, ReferenceType.SUBJECT_SSIN, subjectId));
        }

        if (storeDataEntriesRequest.getPersonInformation() != null) {
            errors.add(createError(MEDICATION_SCHEME_PERSON_INFORMATION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));

        }
        if (storeDataEntriesRequest.getOrganizationInformation() != null) {
            errors.add(createError(MEDICATION_SCHEME_ORGANIZATION_INFORMATION_SPECIFIED, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
        }

        if (CollectionUtils.isEmpty(storeDataEntriesRequest.getDataEntries())) {
            errors.add(createError(MEDICATION_SCHEME_NO_DATAENTRIES, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
        } else {
            boolean detectedMissingReference = false;
            Integer nodeVersionOfFirstEntry = storeDataEntriesRequest.getDataEntries().get(0).getNodeVersion();
            boolean differentNodeVersionsDetected = false;

            for (DataEntry dataEntry : storeDataEntriesRequest.getDataEntries()) {
                if (StringUtils.isBlank(dataEntry.getReference())) {
                    if (!detectedMissingReference) {
                        errors.add(createError(MEDICATION_SCHEME_NO_DATA_ENTRY_REFERENCE, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
                        detectedMissingReference = true;
                    }
                    continue;
                }

                if (!differentNodeVersionsDetected
                        && ((nodeVersionOfFirstEntry == null && dataEntry.getNodeVersion() != null) ||
                            (nodeVersionOfFirstEntry != null && !nodeVersionOfFirstEntry.equals(dataEntry.getNodeVersion())))){
                    errors.add(createError(MEDICATION_SCHEME_MULTIPLE_NODE_VERSIONS, ReferenceType.VALIDATION_ERROR, Integer.toString(errors.size() + 1)));
                    differentNodeVersionsDetected = true;
                }

                final Map<String, String> metaData = dataEntry.getMetadata();
                if (!metaData.containsKey(AVAILABILITY_STATUS_KEY)) {
                    errors.add(createError(MEDICATION_SCHEME_NO_AVAILABILITY_STATUS, ReferenceType.CORRELATION_ID, dataEntry.getReference()));
                } else if (!EnumUtils.isValidEnum(AvailabilityStatus.class, metaData.get(AVAILABILITY_STATUS_KEY).toUpperCase())) {
                    errors.add(createError(MEDICATION_SCHEME_INVALID_AVAILABILITY_STATUS, ReferenceType.CORRELATION_ID, dataEntry.getReference(), EnumUtils.getEnumList(AvailabilityStatus.class)));
                }

                Source businessDataSource = null;
                if (dataEntry.getBusinessData() == null || dataEntry.getBusinessData().length == 0) {
                    errors.add(createError(MEDICATION_SCHEME_NO_BUSINESSDATA, ReferenceType.CORRELATION_ID, dataEntry.getReference()));
                } else {
                    businessDataSource = parseDocument(dataEntry.getBusinessData());
                    if (businessDataSource == null) {
                        errors.add(createError(MEDICATION_SCHEME_UNPARSEABLE_BUSINESSDATA, ReferenceType.CORRELATION_ID, dataEntry.getReference()));
                    }
                }

                FormatCode formatCode = null;
                if (!metaData.containsKey(FORMAT_CODE_KEY)) {
                    errors.add(createError(MEDICATION_SCHEME_NO_FORMAT_CODE, ReferenceType.CORRELATION_ID, dataEntry.getReference()));
                } else {
                    try {
                        formatCode = FormatCode.valueOf(metaData.get(FORMAT_CODE_KEY).toUpperCase());
                    } catch (Exception e) {
                        errors.add(createError(MEDICATION_SCHEME_INVALID_FORMAT_CODE, ReferenceType.CORRELATION_ID, dataEntry.getReference(), EnumUtils.getEnumList(FormatCode.class)));
                    }
                }

                if (formatCode != null && businessDataSource != null) {
                    try {
                        final List<String> validationErrors = MedicationSchemeValidator.validateMedicationScheme(businessDataSource, formatCode);
                        for (String validationError : validationErrors) {
                            errors.add(createError(MEDICATION_SCHEME_INVALID_BUSINESSDATA, ReferenceType.CORRELATION_ID, dataEntry.getReference(), validationError));
                        }
                    } catch (ValidatorException ve) {
                        throw new IntegrationModuleException(ve.getMessage(), ve.getCause());
                    }
                }

                final DataEntryURI uri = new DataEntryURI(dataEntry.getDataEntryURI());
                if (!uri.isCreateURI() && !uri.isUpdateURI()) {
                    errors.add(createError(MEDICATION_SCHEME_INVALID_STORE_DATA_URI, ReferenceType.CORRELATION_ID, dataEntry.getReference()));
                } else if (StringUtils.isNotBlank(subjectId) && !subjectId.equals(uri.getSubjectId())) {
                    errors.add(createError(MEDICATION_SCHEME_SUBJECT_ID_MISMATCH, ReferenceType.CORRELATION_ID, dataEntry.getReference()));
                }

            }
        }
        return errors;
    }


    /**
     * Parses the input bytes to a w3c source object
     *
     * @param data
     * @return
     */

    private static Source parseDocument(final byte[] data) throws IntegrationModuleException {
        ByteArrayInputStream bais = null;
        try {
            bais = new ByteArrayInputStream(data);
            return new DOMSource(getSourceParser().parse(bais));
        } catch (IOException e) {
            LOG.error("Failed to read input file", e);
        } catch (SAXException e) {
            LOG.error("Failed to parse input file to w3c structure", e);
        } finally {
            IOUtils.closeQuietly(bais);
        }

        return null;
    }

    /**
     * Utility method to create an error
     *
     * @param code
     * @param referenceType
     * @param reference
     * @return
     */

    private static Error createError(final MSStatusCode code, final ReferenceType referenceType, final String reference, final Object... context) {
        return new Error(code.getCode(), MSStatusResolver.getStatusMessage(code, context), referenceType.name(), reference);
    }


}
