package be.business.connector.gfddpp.utils;

import be.apb.gfddpp.common.medicationscheme.status.MSStatusCode;
import be.apb.gfddpp.common.medicationscheme.status.MSStatusResolver;
import be.apb.standards.smoa.schema.v1.*;
import be.apb.standards.smoa.schema.v1.Organization;
import be.apb.standards.smoa.schema.v1.PaginationInfo;
import be.apb.standards.smoa.schema.v1.Person;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.gfddpp.domain.medicationscheme.*;
import be.business.connector.gfddpp.domain.medicationscheme.Error;
import be.business.connector.gfddpp.domain.medicationscheme.Status;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.FetchDataEntriesResponse;
import be.business.connector.gfddpp.domain.medicationscheme.protocol.RetrieveTimestampsResponse;
import be.business.connector.gfddpp.tipsystem.threading.GeneralDecryptorThread;
import net.sf.ehcache.Cache;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Semaphore;
import static be.apb.gfddpp.common.medicationscheme.MSConstants.*;
/**
 * @author Liesje Demuynck.
 */
public class MappingUtils {

    private final static Logger LOG = Logger.getLogger(MappingUtils.class);

    /**
     * Maps the given RetrieveTimestampsResponse eventType to a RetrieveTimestampsResponse-domain-object
     *
     * @param pcdhResponse the RetrieveTimestampsResponse eventType
     * @return the mapped RetrieveTimestampsResponse object
     */
    public static RetrieveTimestampsResponse mapMedicationSchemeTimestampsResponse(MedicationSchemeTimestampsResponse pcdhResponse) throws IntegrationModuleException{
        final RetrieveTimestampsResponse response = new RetrieveTimestampsResponse();
        response.setCurrentDateTime(pcdhResponse.getCurrentDateTime().toGregorianCalendar());
        response.setClientMessageID(pcdhResponse.getClientMessageId());
        response.setServerMessageID(pcdhResponse.getServerMessageId());
        for (MedicationSchemeTimestampsResponse.SubjectTimestamp st : pcdhResponse.getSubjectTimestamp()) {
            response.getSubjects().add(new SubjectTimestamp(st.getSubjectId(), st.getLastUpdated().toGregorianCalendar(), st.getVersion()));
        }

        final MSStatusCode statusCode = mapStatusCode(pcdhResponse.getStatus().getCode());
        response.setStatus(new Status(statusCode.getCode(), MSStatusResolver.getStatusMessage(statusCode)));
        for (be.apb.standards.smoa.schema.v1.Error error : pcdhResponse.getStatus().getError()) {
            final MSStatusCode errorCode = mapStatusCode(error.getCode());
            response.getStatus().getErrors().add(new Error(errorCode.getCode(), MSStatusResolver.getStatusMessage(errorCode), error.getReferenceType().name(), error.getReference()));
        }
        return response;
    }

    /**
     * Maps the given FetchMedicationEntriesResponse eventType to a FetchDataEntriesResponse-object
     *
     * @param pcdhResponse the MedicationSchemeEntriesResponse eventType
     * @param kgssCache    the kgssCache: the cache is assumed to contain the necessary keys
     * @param threadLimit  the maximal number of allowed concurrent threads
     * @return the mapped FetchDataEntriesResponse object
     */
    public static FetchDataEntriesResponse mapMedicationSchemeEntriesResponse(final MedicationSchemeEntriesResponse pcdhResponse, final Cache kgssCache, final int threadLimit) throws IntegrationModuleException {
        final FetchDataEntriesResponse response = new FetchDataEntriesResponse();
        response.setClientMessageID(pcdhResponse.getClientMessageId());
        response.setServerMessageID(pcdhResponse.getServerMessageId());
        response.setSubjectID(pcdhResponse.getSubjectId());
        response.setStatus(new Status());
        response.getStatus().setMessage(pcdhResponse.getStatus().getMessage());
        final MSStatusCode resultStatusCode = mapStatusCode(pcdhResponse.getStatus().getCode());
        response.getStatus().setCode(resultStatusCode.getCode());
        for (be.apb.standards.smoa.schema.v1.Error error : pcdhResponse.getStatus().getError()) {
            final MSStatusCode errorCode = mapStatusCode(error.getCode());
            response.getStatus().getErrors().add(new Error(errorCode.getCode(), MSStatusResolver.getStatusMessage(errorCode), error.getReferenceType().name(), error.getReference()));
        }

        // error response can be returned right away
        if (!MSStatusCode.MEDICATION_SCHEME_SUCCESS.equals(resultStatusCode)) {
            return response;
        }

        final MedicationSchemeResponse medicationScheme = pcdhResponse.getMedicationScheme();
        if (medicationScheme == null) {
            throw new IntegrationModuleException(I18nHelper.getLabel("error.no.medication.node"));
        }

        response.setLastUpdated(medicationScheme.getLastUpdated().toGregorianCalendar());
        response.setVersion(medicationScheme.getVersion());

        final Node medicationNode = new Node();
        response.getNodes().add(medicationNode);

        medicationNode.setName(MEDICATION_SCHEME_KEY);
        medicationNode.setVersion(medicationScheme.getVersion());
        medicationNode.setUpdatedBy(new UpdatedBy());
        if (medicationScheme.getUpdatedBy().getOrganization() != null) {
            final Organization pcdhOrganization = medicationScheme.getUpdatedBy().getOrganization();
            final be.business.connector.gfddpp.domain.medicationscheme.Organization organization = new be.business.connector.gfddpp.domain.medicationscheme.Organization();
            organization.setName(pcdhOrganization.getName());
            organization.setActorIdValue(pcdhOrganization.getOrganizationId());
            organization.setActorIdSource(Source.fromValue(pcdhOrganization.getOrganizationIdType().name()));
            organization.setRole(pcdhOrganization.getRole().name());
            medicationNode.getUpdatedBy().setOrganization(organization);
        }
        if (medicationScheme.getUpdatedBy().getPerson() != null) {
            final Person pcdhperson = medicationScheme.getUpdatedBy().getPerson();
            final be.business.connector.gfddpp.domain.medicationscheme.Person person = new be.business.connector.gfddpp.domain.medicationscheme.Person();
            person.setFirstName(pcdhperson.getFirstName());
            person.setLastName(pcdhperson.getLastName());
            for (PersonIdentifier personId : pcdhperson.getPersonId()) {
                if (PersonIdType.NIHII.equals(personId.getIdType())) {
                    person.setNihii(personId.getIdValue());
                }
                if (PersonIdType.INSS.equals(personId.getIdType())) {
                    person.setSsin(personId.getIdValue());
                }
            }
            person.setRole(pcdhperson.getRole().name());
            medicationNode.getUpdatedBy().setPerson(person);
        }
        final PaginationInfo pcdhPaginationInfo = medicationScheme.getPaginationInfo();
        medicationNode.setPagination(new be.business.connector.gfddpp.domain.medicationscheme.PaginationInfo());
        medicationNode.getPagination().setUri(DataEntryURI.readAllURI(pcdhResponse.getSubjectId()).toString());
        medicationNode.getPagination().setPaginationSize(pcdhPaginationInfo.getPaginationSize());
        medicationNode.getPagination().setPaginationIndex(pcdhPaginationInfo.getPaginationIndex());
        medicationNode.getPagination().setRecordCount(pcdhPaginationInfo.getRecordCount());

        final List<be.business.connector.gfddpp.domain.medicationscheme.DataEntry> dataEntries = new ArrayList<DataEntry>();
        medicationNode.setDataEntries(dataEntries);

        final Map<DataEntryResponse, GeneralDecryptorThread> decryptorThreads = doDecryptions(medicationScheme.getDataEntry(), kgssCache, threadLimit);
        for (DataEntryResponse pcdhDataEntry : medicationScheme.getDataEntry()) {
            final DataEntry dataEntry = new DataEntry();
            dataEntry.setNodeVersion(medicationNode.getVersion());
            dataEntry.setDataEntryURI(DataEntryURI.readURI(pcdhResponse.getSubjectId(), pcdhDataEntry.getDataEntryId(), pcdhDataEntry.getDataEntryVersion()).toString());
            dataEntry.setBusinessData(decryptorThreads.get(pcdhDataEntry).getDecoded());

            dataEntry.getMetadata().put(SGUID_KEY, medicationScheme.getSguid());
            dataEntry.getMetadata().put(DGUID_KEY, pcdhDataEntry.getDguid());
            dataEntry.getMetadata().put(TGUID_KEY, pcdhDataEntry.getDataEntryId());
            dataEntry.getMetadata().put(AVAILABILITY_STATUS_KEY, pcdhDataEntry.getAvailabilityStatus().name());
            dataEntry.getMetadata().put(FORMAT_CODE_KEY, pcdhDataEntry.getFormatCode().name());

            // author info
            if (pcdhDataEntry.getAuthor() != null) {
                if (pcdhDataEntry.getAuthor().getOrganization() != null) {
                    final Organization organization = pcdhDataEntry.getAuthor().getOrganization();
                    addIfNotBlank(dataEntry.getMetadata(), AUTHOR_NAME_KEY, organization.getName());
                    dataEntry.getMetadata().put(AUTHOR_ORGANIZATION_ID_KEY, organization.getOrganizationId());
                    dataEntry.getMetadata().put(AUTHOR_ORGANIZATION_ID_SOURCE_KEY, organization.getOrganizationIdType().name());
                    dataEntry.getMetadata().put(AUTHOR_ROLE_KEY, organization.getRole().name());
                }
                if (pcdhDataEntry.getAuthor().getPerson() != null) {
                    final Person person = pcdhDataEntry.getAuthor().getPerson();
                    addIfNotBlank(dataEntry.getMetadata(), AUTHOR_FIRST_NAME_KEY, person.getFirstName());
                    addIfNotBlank(dataEntry.getMetadata(), AUTHOR_LAST_NAME_KEY, person.getLastName());
                    for (PersonIdentifier personIdentifier : person.getPersonId()) {
                        dataEntry.getMetadata().put(AUTHOR_PREFIX_KEY + personIdentifier.getIdType(), personIdentifier.getIdValue());
                    }
                    dataEntry.getMetadata().put(AUTHOR_ROLE_KEY, person.getRole().name());

                }
            }
            if (pcdhDataEntry.getMetaDataList() != null) {
                final List<MetaDataType> metaDataList = pcdhDataEntry.getMetaDataList().getMetaData();
                for (int i = 0; i < metaDataList.size(); i++) {
                    dataEntry.getMetadata().put(metaDataList.get(i).getKey(), metaDataList.get(i).getValue());
                }
            }

            dataEntries.add(dataEntry);
        }

        return response;
    }

    private static void addIfNotBlank(final Map<String,String> dataMap, final String key, final String value){
        if(StringUtils.isNotBlank(value)){
            dataMap.put(key, value);
        }
    }

    public static MSStatusCode mapStatusCode(final String statusCode) throws IntegrationModuleException{
        try{
            return MSStatusCode.valueOf(statusCode);
        }catch (IllegalArgumentException iae){
            throw new IntegrationModuleException(I18nHelper.getLabel("error.validation.status.code"));
        }
    }

    /**
     * Decrypts all encrypted content in the provided DataEntries. The results are returned in a map of <DataEntryResponse, GeneralDecryptorThread> pairs.
     *
     * @param dataEntries the dataEntries of which the contents needs decryption
     * @param kgssCache   the kgssCache: the cache is assumed to contain the necessary keys
     * @param threadLimit the maximal number of allowed concurrent threads
     * @return The decrypted contents, bundled in a map of <DataEntryResponse, GeneralDecryptorThread> pairs.
     */
    private static Map<DataEntryResponse, GeneralDecryptorThread> doDecryptions(List<DataEntryResponse> dataEntries, Cache kgssCache, int threadLimit) throws IntegrationModuleException {
        Semaphore availableThreads = new Semaphore(threadLimit, true);

        final Map<DataEntryResponse, GeneralDecryptorThread> decryptorThreads = new HashMap<DataEntryResponse, GeneralDecryptorThread>();
        for (DataEntryResponse dataEntryResponse : dataEntries) {
            if (dataEntryResponse.getEncryptedData() == null) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.no.business.data", new Object[]{dataEntryResponse.getDguid()}));
            }
            final String encryptionKeyId = dataEntryResponse.getEncryptedData().getEncryptionKeyId();
            final byte[] encryptedData = dataEntryResponse.getEncryptedData().getData();

            // businessdata decryptor thread
            final GeneralDecryptorThread businessDataThread = new GeneralDecryptorThread(availableThreads, encryptedData, encryptionKeyId, kgssCache);
            decryptorThreads.put(dataEntryResponse, businessDataThread);
        }

        for (GeneralDecryptorThread decryptorThread : decryptorThreads.values()) {
            availableThreads.acquireUninterruptibly();
            decryptorThread.start();
        }
        try {
            availableThreads.acquireUninterruptibly(threadLimit);
        } catch (IllegalArgumentException e) {
            LOG.debug("Incorrect Thread configuration : " + e);
        }

        return decryptorThreads;
    }


}
