package be.business.connector.gfddpp.utils;

import be.apb.gfddpp.common.constants.SupportedDataTypes;
import be.apb.gfddpp.common.exceptions.GFDDPPException;
import be.apb.gfddpp.common.log.LogsGetDataParameters;
import be.apb.gfddpp.common.log.LogsGetDataParameters.IntermediateMessage;
import be.apb.gfddpp.common.status.StatusResolver;
import be.apb.gfddpp.common.utils.JaxContextCentralizer;
import be.apb.gfddpp.common.utils.Operation;
import be.apb.gfddpp.domain.PersonIdType;
import be.apb.standards.gfddpp.constants.request.DateRangeTypes;
import be.apb.standards.gfddpp.patient.signature.PatientSignature;
import be.apb.standards.gfddpp.request.*;
import be.business.connector.core.exceptions.IntegrationModuleEhealthException;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.exceptions.IntegrationModuleRuntimeException;
import be.business.connector.core.utils.*;
import be.business.connector.gfddpp.domain.ThreeStateBoolean;
import be.ehealth.apb.gfddpp.services.tipsystem.*;
import be.ehealth.technicalconnector.beid.BeIDInfo;
import be.ehealth.technicalconnector.exception.TechnicalConnectorException;
import be.ehealth.technicalconnector.service.sts.security.Credential;
import be.ehealth.technicalconnector.service.sts.security.impl.BeIDCredential;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.Charsets;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import javax.xml.bind.JAXBException;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.text.SimpleDateFormat;
import java.util.GregorianCalendar;
import java.util.List;

public class RequestCreatorUtils {

	private static final Logger LOG = Logger.getLogger(RequestCreatorUtils.class);

	private static JaxContextCentralizer jaxContextCentralizer = JaxContextCentralizer.getInstance();

		
	/**
	 * Creates the routing parameter.
	 * 
	 * @param tipId
	 *            the tip id
	 * @return the routing parameters type
	 */
	public static RoutingParametersType createRoutingParameter(String tipId) {
		RoutingParametersType routingParametersType = new RoutingParametersType();
		routingParametersType.setTIPId(tipId);
		return routingParametersType;
	}

	/**
	 * Creates the authorization parameters.
	 * 
	 * @param patientConsentFlag
	 *            the patient consent flag
	 * @param patientSignature
	 *            the patient signature
	 * @param patientSignatureDate
	 *            the patient signature date
	 * @param therapeuticalRelationShipTypeFlag
	 *            the therapeutical relation ship type flag
	 * @param patientid
	 * @param encryptionUtils
	 * @return the authorization parameters type
	 * @throws IntegrationModuleException
	 */
	public static AuthorizationParametersType createAuthorizationParameters(boolean requestPatientSignature, ThreeStateBoolean patientConsent, ThreeStateBoolean therapeuticalRelationShipTypeFlag, String patientid, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		AuthorizationParametersType authorizationParametersType = new AuthorizationParametersType();
		authorizationParametersType.setPatientConsent(createPatientConsent(requestPatientSignature, patientConsent, patientid, encryptionUtils));
		if (therapeuticalRelationShipTypeFlag != null) {
			authorizationParametersType.setTherapeuticalRelationShip(createTherapeuticalRelationShip(therapeuticalRelationShipTypeFlag));
		}
		return authorizationParametersType;
	}

	public static be.ehealth.apb.gfddpp.services.pcdh.AuthorizationParametersType createPCDHAuthorizationParameters(boolean requestPatientSignature, ThreeStateBoolean therapeuticRelationShip, String dGuid, String patientid, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		be.ehealth.apb.gfddpp.services.pcdh.AuthorizationParametersType authorizationParametersType = new be.ehealth.apb.gfddpp.services.pcdh.AuthorizationParametersType();
		authorizationParametersType.setPatientConsent(createPCDHPatientConsent(requestPatientSignature, dGuid, patientid, encryptionUtils));
		if (therapeuticRelationShip != null && therapeuticRelationShip.getBooleanValue() != null) {
			authorizationParametersType.setTherapeuticalRelationShip(createPCDHTherapeuticalRelationShip(therapeuticRelationShip));
		}
		return authorizationParametersType;
	}

	/**
	 * Creates the patient consent.
	 * 
	 * @param encryptionUtils
	 * 
	 * @param sGuid
	 * 
	 * @param patientConsentFlag
	 *            the patient consent flag
	 * @param patientSignature
	 *            the patient signature
	 * @param patientSignatureDate
	 *            the patient signature date
	 * @return the patient consent type
	 * @throws IntegrationModuleException
	 */
	public static PatientConsentType createPatientConsent(boolean requestPatientSignature, ThreeStateBoolean patientConsent, String patientid, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		PatientConsentType patientConsentType = new PatientConsentType();
		if (patientConsent != null && patientConsent.getBooleanValue() != null) {
			patientConsentType.setPatientConsentFlag(patientConsent.getBooleanValue());
		}
		if (requestPatientSignature) {
			patientConsentType.setPatientSignature(createPatientSignature(patientid, encryptionUtils));
		}
		return patientConsentType;
	}

	public static be.ehealth.apb.gfddpp.services.pcdh.PatientConsentType createPCDHPatientConsent(boolean requestPatientSignature, String dGuid, String patientid, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		be.ehealth.apb.gfddpp.services.pcdh.PatientConsentType patientConsentType = new be.ehealth.apb.gfddpp.services.pcdh.PatientConsentType();
		/**
		 * temporary hardcoded value has in a near futur eHealth will provide
		 * this value
		 **/
		patientConsentType.setPatientConsentFlag(false);
		if (requestPatientSignature) {
			patientConsentType.setPatientSignature(createPCDHPatientSignature(dGuid, patientid, encryptionUtils));
		}
		return patientConsentType;
	}

	/**
	 * Creates the patient signature.
	 * 
	 * @param patientid
	 * 
	 * @param sGuid
	 * 
	 * @param patientSignature
	 *            the patient signature
	 * @param patientSignatureDate
	 *            the patient signature date
	 * @return the patient signature type
	 * @throws IntegrationModuleException
	 */
	public static PatientSignatureType createPatientSignature(String patientid, EncryptionUtils encryptionUtils) throws IntegrationModuleException {
		LOG.debug("********** Create strong patient signature for patient: " + patientid + " *************");
		PatientSignatureType patientSignatureType = new PatientSignatureType();

		byte[] patientSignature = null;

		XMLGregorianCalendar xgcal = createCurrentXmlGregorianCalendar();
		patientSignatureType.setDateTime(xgcal);
		SimpleDateFormat format = new SimpleDateFormat("yyyyMMdd");

		String nISS = null;
		
		PrivateKey prvk = null;
		PublicKey pk = null;
		if (PropertyHandler.getInstance().hasProperty("patientsignature.mock.enable") && PropertyHandler.getInstance().hasProperty("patientsignature.mock.niss") && !StringUtils.isEmpty(PropertyHandler.getInstance().getProperty("patientsignature.mock.enable")) && PropertyHandler.getInstance().getProperty("patientsignature.mock.enable").equals("true") && !StringUtils.isEmpty(PropertyHandler.getInstance().getProperty("patientsignature.mock.niss"))) {
			pk = encryptionUtils.getPublicKey();
			prvk = encryptionUtils.getPrivateKey();
			nISS = PropertyHandler.getInstance().getProperty("patientsignature.mock.niss");
		} else {
			nISS = initNiss();
			try {
				Credential cred = BeIDCredential.getInstance(nISS, BeIDCredential.EID_SIGN_ALIAS);
				prvk = cred.getPrivateKey();
				pk = cred.getPublicKey();
			} catch (TechnicalConnectorException e) {
				Exceptionutils.errorHandler(e);
			}
		}

		if (!nISS.equals(patientid.trim())) {
			throw new IntegrationModuleException(I18nHelper.getLabel("validation.patient.signature.not.equal", new Object[] { patientid }));
		}

		String signatureString = null;

		signatureString = nISS + ";" + format.format(xgcal.toGregorianCalendar().getTime());

		try {
			patientSignature = Base64.encodeBase64(SignatureGenerator.sign(signatureString, prvk, "SHA1withRSA"));
		} catch (Exception e) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.patient.signature.generation"), e);
		}

		PatientSignature patientSig = new PatientSignature();
		patientSig.setPublickey(pk.getEncoded());
		patientSig.setSignature(patientSignature);

		try {
			byte[] ps = jaxContextCentralizer.toXml(PatientSignature.class, patientSig).getBytes(Charsets.UTF_8);
			patientSignatureType.setSignature(ps);
		} catch (GFDDPPException e) {
			LOG.error("Exception in : createPatientSignature", e);
			throw new IntegrationModuleException(StatusResolver.resolveMessage(StatusResolver.getLocalResourceBundle(), e.getStatusCode(), e.getPlaceHolders()), e);
		}

		LOG.debug("********** Strong patient signature created *************");

		return patientSignatureType;

	}

	public static be.ehealth.apb.gfddpp.services.pcdh.PatientSignatureType createPCDHPatientSignature(String dGuid, String patientid, EncryptionUtils encryptionUtils) throws IntegrationModuleException {

		LOG.debug("********** Create strong patient signature for patient: " + patientid + " *************");
		be.ehealth.apb.gfddpp.services.pcdh.PatientSignatureType patientSignatureType = new be.ehealth.apb.gfddpp.services.pcdh.PatientSignatureType();

		byte[] patientSignature = null;

		XMLGregorianCalendar xgcal = createCurrentXmlGregorianCalendar();
		patientSignatureType.setDateTime(xgcal);

		SimpleDateFormat format = new SimpleDateFormat("yyyyMMdd");

		PrivateKey prvk = null;
		PublicKey pk = null;
		String nISS = null;

		if (PropertyHandler.getInstance().hasProperty("patientsignature.mock.enable") && PropertyHandler.getInstance().hasProperty("patientsignature.mock.niss") && !StringUtils.isEmpty(PropertyHandler.getInstance().getProperty("patientsignature.mock.enable")) && PropertyHandler.getInstance().getProperty("patientsignature.mock.enable").equals("true") && !StringUtils.isEmpty(PropertyHandler.getInstance().getProperty("patientsignature.mock.niss"))) {
			pk = encryptionUtils.getPublicKey();
			prvk = encryptionUtils.getPrivateKey();
			nISS = PropertyHandler.getInstance().getProperty("patientsignature.mock.niss");
		} else {
			nISS = initNiss();
			try {
				Credential cred = BeIDCredential.getInstance(nISS, BeIDCredential.EID_SIGN_ALIAS);
				prvk = cred.getPrivateKey();
				pk = cred.getPublicKey();
			} catch (TechnicalConnectorException e) {
				Exceptionutils.errorHandler(e);
			}
		}

		if (!nISS.equals(patientid.trim())) {
			throw new IntegrationModuleException(I18nHelper.getLabel("validation.patient.signature.not.equal", new Object[] { patientid }));
		}

		String signatureString = null;
		if (dGuid == null) {
			signatureString = nISS + ";" + format.format(xgcal.toGregorianCalendar().getTime());
		} else {
			signatureString = dGuid + ";" + format.format(xgcal.toGregorianCalendar().getTime());
		}

		try {
			patientSignature = Base64.encodeBase64(SignatureGenerator.sign(signatureString, prvk, "SHA1withRSA"));
		} catch (Exception e) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.patient.signature.generation"), e);
		}

		PatientSignature patientSig = new PatientSignature();
		patientSig.setPublickey(pk.getEncoded());
		patientSig.setSignature(patientSignature);

		try {
			patientSignatureType.setSignature(jaxContextCentralizer.toXml(PatientSignature.class, patientSig).getBytes(Charsets.UTF_8));
		} catch (GFDDPPException e) {
			e.printStackTrace();
			throw new IntegrationModuleException(StatusResolver.resolveMessage(StatusResolver.getLocalResourceBundle(), e.getStatusCode(), e.getPlaceHolders()), e);
		}

		LOG.debug("********** Strong patient signature created *************");
		return patientSignatureType;
	}

	/**
	 * Creates the therapeutical relationship.
	 * 
	 * @param therapeuticalRelationShipTypeFlag
	 *            the therapeutical relation ship type flag
	 * @return the therapeutical relation ship type
	 */
	public static TherapeuticalRelationShipType createTherapeuticalRelationShip(ThreeStateBoolean therapeuticalRelationShipTypeFlag) {
		TherapeuticalRelationShipType therapeuticalRelationShipType = new TherapeuticalRelationShipType();
		therapeuticalRelationShipType.setTherapeuticalRelationShipFlag(therapeuticalRelationShipTypeFlag.getBooleanValue());
		return therapeuticalRelationShipType;
	}

	public static be.ehealth.apb.gfddpp.services.pcdh.TherapeuticalRelationShipType createPCDHTherapeuticalRelationShip(ThreeStateBoolean therapeuticalRelationShipTypeFlag) {
		be.ehealth.apb.gfddpp.services.pcdh.TherapeuticalRelationShipType therapeuticalRelationShipType = new be.ehealth.apb.gfddpp.services.pcdh.TherapeuticalRelationShipType();
		therapeuticalRelationShipType.setTherapeuticalRelationShipFlag(therapeuticalRelationShipTypeFlag.getBooleanValue());
		return therapeuticalRelationShipType;
	}

	public static String createGetDataParams(PatientType patient, String requestorId, String requestorType, boolean excludeOwnData, String range, String dataType) throws DatatypeConfigurationException, GFDDPPException {

		GetDataRequestParameters getDataRequestParameters = new GetDataRequestParameters();
		getDataRequestParameters.setPatient(patient);
		getDataRequestParameters.setRequestor(createRequestor(requestorId, requestorType));
		if (dataType != null) {
			getDataRequestParameters.setDataType(dataType.toLowerCase());
		}
		// getDataRequestParameters.setPharmacyId(pharmacyIdentifier);
		DataSpecificParametersGetData getDataSpecificParam = new DataSpecificParametersGetData();
		getDataSpecificParam.setExcludeOwnData(excludeOwnData);

		DatatypeFactory dataTypeFactory = DatatypeFactory.newInstance();
		GregorianCalendar gregorianCalendar = new GregorianCalendar();

		if (StringUtils.equals(range, DateRangeTypes.FULL.name())) {
			getDataSpecificParam.setDateRangeType(DateRangeTypes.FULL.name());
		} else if (StringUtils.equals(range, DateRangeTypes.DEFAULT.name())) {
			getDataSpecificParam.setDateRangeType(DateRangeTypes.DEFAULT.name());
		} else {
			getDataSpecificParam.setDateRangeType(DateRangeTypes.CUSTOM.name());
			String[] splitedRange = range.split("-");
			gregorianCalendar.set(Integer.parseInt(splitedRange[0].substring(0, 4)), Integer.parseInt(splitedRange[0].substring(4, 6)) - 1, Integer.parseInt(splitedRange[0].substring(6)));
			getDataSpecificParam.setStartDate(dataTypeFactory.newXMLGregorianCalendar(gregorianCalendar));

			gregorianCalendar.set(Integer.parseInt(splitedRange[1].substring(0, 4)), Integer.parseInt(splitedRange[1].substring(4, 6)) - 1, Integer.parseInt(splitedRange[1].substring(6)));
			getDataSpecificParam.setEndDate(dataTypeFactory.newXMLGregorianCalendar(gregorianCalendar));
		}

		getDataRequestParameters.setDataSpecificParametersGetData(getDataSpecificParam);
		LogsGetDataParameters.logs(getDataRequestParameters, IntermediateMessage.REQUEST_CREATED, Operation.GETDATA);

		return JaxContextCentralizer.getInstance().toXml(GetDataRequestParameters.class, getDataRequestParameters);
	}

	public static String createGetDataTypesParams(PatientType patient, String requestorId, String requestorType) throws JAXBException, DatatypeConfigurationException {
		MarshallerHelper<Object, GetDataRequestParameters> helper = new MarshallerHelper<Object, GetDataRequestParameters>(Object.class, GetDataRequestParameters.class);
		GetDataRequestParameters getDataRequestParameters = new GetDataRequestParameters();
		getDataRequestParameters.setPatient(patient);
		getDataRequestParameters.setRequestor(createRequestor(requestorId, requestorType));
		getDataRequestParameters.setDataType("dataTypes");
		DataSpecificParametersGetDataTypes getDataSpecificParam = new DataSpecificParametersGetDataTypes();

		getDataRequestParameters.setDataSpecificParametersGetDataTypes(getDataSpecificParam);
		LogsGetDataParameters.logs(getDataRequestParameters, IntermediateMessage.REQUEST_CREATED, Operation.GETDATATYPES);

		return helper.marsh(getDataRequestParameters);
	}

	public static String createRequestMedicationSchemeTimestamps(String requestorId, String requestorType, List<String> subjectIds, String clientMessageId) throws JAXBException {
		MarshallerHelper<Object, GetDataRequestParameters> helper = new MarshallerHelper<Object, GetDataRequestParameters>(Object.class, GetDataRequestParameters.class);
		GetDataRequestParameters getDataRequestParameters = new GetDataRequestParameters();
		getDataRequestParameters.setRequestor(createRequestor(requestorId, requestorType));
		getDataRequestParameters.setDataType(SupportedDataTypes.MEDICATION_SCHEME_TIMESTAMPS.getName());
		DataSpecificParametersMedicationSchemeTimestamps specificParameters = new DataSpecificParametersMedicationSchemeTimestamps();
		if(StringUtils.isNotBlank(clientMessageId)) {
			specificParameters.setClientMessageId(clientMessageId);
		}
		specificParameters.getPatientINSS().addAll(subjectIds);

		getDataRequestParameters.setDataSpecificParametersMedicationSchemeTimestamps(specificParameters);
		LogsGetDataParameters.logs(getDataRequestParameters, IntermediateMessage.REQUEST_CREATED, Operation.MEDICATIONSCHEME_GET_TIMESTAMPS);

		return helper.marsh(getDataRequestParameters);
	}

	public static String createRequestMedicationSchemeEntries(String requestorId, String requestorType, String patientId, int paginationIndex, String clientMessageId) throws JAXBException{
		MarshallerHelper<Object, GetDataRequestParameters> helper = new MarshallerHelper<Object, GetDataRequestParameters>(Object.class, GetDataRequestParameters.class);
		GetDataRequestParameters getDataRequestParameters = new GetDataRequestParameters();
		getDataRequestParameters.setRequestor(createRequestor(requestorId, requestorType));
		getDataRequestParameters.setPatient(createPatientType(patientId, PersonIdType.INSS.name()));
		getDataRequestParameters.setDataType(SupportedDataTypes.MEDICATION_SCHEME_ENTRIES.getName());
		DataSpecificParametersMedicationSchemeEntries specificParameters = new DataSpecificParametersMedicationSchemeEntries();
		if(StringUtils.isNotBlank(clientMessageId)) {
			specificParameters.setClientMessageId(clientMessageId);
		}
		specificParameters.setPaginationIndex(paginationIndex);

		getDataRequestParameters.setDataSpecificParametersMedicationSchemeEntries(specificParameters);
		LogsGetDataParameters.logs(getDataRequestParameters, IntermediateMessage.REQUEST_CREATED, Operation.MEDICATIONSCHEME_GET_ENTRIES);

		return helper.marsh(getDataRequestParameters);
	}


	public static AuthorizationParametersType createAuthorizationRevokeParameters() {
		be.ehealth.apb.gfddpp.services.tipsystem.AuthorizationParametersType authorizationParametersType = new be.ehealth.apb.gfddpp.services.tipsystem.AuthorizationParametersType();
		PatientConsentType patientConsentType = new PatientConsentType();
		patientConsentType.setPatientConsentFlag(false);
		authorizationParametersType.setPatientConsent(patientConsentType);

		return authorizationParametersType;
	}

	public static PatientType createPatientType(String patientId, String patientIdType) {

		be.apb.standards.gfddpp.request.ObjectFactory o = new be.apb.standards.gfddpp.request.ObjectFactory();
		PatientType patient = o.createPatientType();
		patient.setPatientId(patientId);
		patient.setPatientIdType(patientIdType);
		return patient;
	}

	public static String createRequestPharmacyDetails(String requestorId, String requestorType, PatientType patient, String dGuid, String motivationType, String motivationText) throws JAXBException {

		MarshallerHelper<Object, GetDataRequestParameters> helper = new MarshallerHelper<Object, GetDataRequestParameters>(Object.class, GetDataRequestParameters.class);
		GetDataRequestParameters getDataRequestParameters = new GetDataRequestParameters();
		getDataRequestParameters.setPatient(patient);
		getDataRequestParameters.setRequestor(createRequestor(requestorId, requestorType));
		getDataRequestParameters.setDataType("pharmacydetails");
		DataSpecificParametersPharmacyDetails dataSpecificParametersPharmacyDetails = new DataSpecificParametersPharmacyDetails();
		dataSpecificParametersPharmacyDetails.setDGuid(dGuid);
		Motivation mvt = new Motivation();
		mvt.setMotivationtext(motivationText);
		mvt.setMotivationtype(motivationType);
		dataSpecificParametersPharmacyDetails.setMotivation(mvt);
		getDataRequestParameters.setDataSpecificParametersPharmacyDetails(dataSpecificParametersPharmacyDetails);
		LogsGetDataParameters.logs(getDataRequestParameters, IntermediateMessage.REQUEST_CREATED, Operation.GETPHARMACYDETAILS);
		return helper.marsh(getDataRequestParameters);
	}

	private static RequestorType createRequestor(String requestorId, String requestorType) {
		RequestorType type = new RequestorType();
		type.setRequestorId(requestorId);
		type.setRequestorType(requestorType);
		return type;
	}
	
	public static String createRequestGetStatus(String requestorId, String requestorType, String sGuid, String dGuid) throws JAXBException {
		MarshallerHelper<Object, GetDataRequestParameters> helper = new MarshallerHelper<Object, GetDataRequestParameters>(Object.class, GetDataRequestParameters.class);
		GetDataRequestParameters getDataRequestParameters = new GetDataRequestParameters();
		getDataRequestParameters.setRequestor(createRequestor(requestorId, requestorType));
		getDataRequestParameters.setDataType("StatusMessage");
		DataSpecificParametersGetStatus dataSpecificParametersGetStatus = new DataSpecificParametersGetStatus();

		if (!StringUtils.isEmpty(sGuid)) {
			dataSpecificParametersGetStatus.setSGUID(sGuid);
		}
		if (!StringUtils.isEmpty(dGuid)) {
			dataSpecificParametersGetStatus.setDGUID(dGuid);
		}
		getDataRequestParameters.setVersion("1.0");
		getDataRequestParameters.setDataSpecificParametersGetStatus(dataSpecificParametersGetStatus);
		LogsGetDataParameters.logs(getDataRequestParameters, IntermediateMessage.REQUEST_CREATED, Operation.GETSTATUS);

		return helper.marsh(getDataRequestParameters);
	}

	private static XMLGregorianCalendar createCurrentXmlGregorianCalendar() throws IntegrationModuleException {
		GregorianCalendar gcal = (GregorianCalendar) GregorianCalendar.getInstance();
		XMLGregorianCalendar xgcal = null;
		try {
			xgcal = DatatypeFactory.newInstance().newXMLGregorianCalendar(gcal);
		} catch (DatatypeConfigurationException e1) {
			e1.printStackTrace();
			throw new IntegrationModuleException(I18nHelper.getLabel("error.technical"));
		}
		return xgcal;
	}

	private static String initNiss() throws IntegrationModuleEhealthException, IntegrationModuleRuntimeException, IntegrationModuleException {
		BeIDInfo beIDInfo = null;
		try {
			beIDInfo = BeIDInfo.getInstance("test");
		} catch (TechnicalConnectorException e) {
			if (StringUtils.contains(e.getMessage().toLowerCase(), "EID is not present".toLowerCase())) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.eid.read"), e);
			}
			Exceptionutils.errorHandler(e);
		}
		return beIDInfo.getIdentity().getNationalNumber();
	}

}
