package be.business.connector.recipe.executor;

import static be.business.connector.recipe.utils.RidValidator.validateRid;

import java.util.List;

import org.apache.log4j.Logger;
import org.perf4j.aop.Profiled;

import com.sun.xml.ws.client.ClientTransportException;

import be.business.connector.common.ApplicationConfig;
import be.business.connector.common.StandaloneRequestorProvider;
import be.business.connector.core.domain.IdentifierTypes;
import be.business.connector.core.domain.KgssIdentifierType;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.handlers.InsurabilityHandler;
import be.business.connector.core.utils.Exceptionutils;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.core.utils.MarshallerHelper;
import be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult;
import be.business.connector.recipe.executor.services.RecipeExecutorServiceImpl;
import be.business.connector.recipe.utils.RidValidator;
import be.ehealth.technicalconnector.service.kgss.domain.KeyResult;
import be.fgov.ehealth.etee.crypto.encrypt.EncryptionToken;
import be.fgov.ehealth.recipe.core.v2.ExecutorServiceAdministrativeInformationType;
import be.fgov.ehealth.recipe.core.v2.SecuredContentType;
import be.fgov.ehealth.recipe.protocol.v2.CreateFeedbackRequest;
import be.fgov.ehealth.recipe.protocol.v2.GetPrescriptionForExecutorRequest;
import be.fgov.ehealth.recipe.protocol.v2.GetPrescriptionForExecutorResponse;
import be.fgov.ehealth.recipe.protocol.v2.ListNotificationsRequest;
import be.fgov.ehealth.recipe.protocol.v2.ListNotificationsResponse;
import be.fgov.ehealth.recipe.protocol.v2.MarkAsArchivedRequest;
import be.fgov.ehealth.recipe.protocol.v2.MarkAsDeliveredRequest;
import be.fgov.ehealth.recipe.protocol.v2.MarkAsUnDeliveredRequest;
import be.fgov.ehealth.recipe.protocol.v2.RevokePrescriptionForExecutorRequest;
import be.recipe.services.executor.GetOpenPrescriptionForExecutor;
import be.recipe.services.executor.GetPrescriptionForExecutorResultSealed;
import be.recipe.services.executor.ListNotificationsItem;

/**
 * The Class ExecutorIntegrationModuleImpl.
 */
public class ExecutorIntegrationModuleImpl extends AbstractExecutorIntegrationModule implements ExecutorIntegrationModule {

	/** The Constant LOG. */
	private static final Logger LOG = Logger.getLogger(ExecutorIntegrationModuleImpl.class);
	

	/**
	 * Instantiates a new executor integration module impl.
	 *
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
    public ExecutorIntegrationModuleImpl() throws IntegrationModuleException {
        super();
        LOG.info("*************** Executor System module init correctly *******************");
    }

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "ExecutorIntegrationModule#getPrescription")
	@Override
	public GetPrescriptionForExecutorResult getPrescription(final String rid) throws IntegrationModuleException {
		RidValidator.validateRid(rid);
        ApplicationConfig.getInstance().assertValidPharmacySession();
        InsurabilityHandler.setInsurability(null);
        InsurabilityHandler.setMessageId(null);

        try {
            // create get prescription request
            final byte[] sealedGetPrescriptionForExecutorParam = getSealedGetPrescriptionForExecutorParam(rid);

			final GetPrescriptionForExecutorRequest request = new GetPrescriptionForExecutorRequest();
            request.setDisablePatientInsurabilityCheckParam(Boolean.parseBoolean(getPropertyHandler().getProperty("patient.insurability.disable"))); // will be false if not defined in property file
            request.setSecuredGetPrescriptionForExecutorRequest(createSecuredContentType(sealedGetPrescriptionForExecutorParam));
            GetPrescriptionForExecutorResponse response = null;

            try {
                response = RecipeExecutorServiceImpl.getInstance().getPrescriptionForExecutor(request);
			} catch (final ClientTransportException cte) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
            }
            checkStatus(response);

            final byte[] securedContent = response.getSecuredGetPrescriptionForExecutorResponse().getSecuredContent();
			final GetPrescriptionForExecutorResult finalResult = createGetPrescriptionForExecutorResult(securedContent);
            getPrescriptionCache().put(rid, finalResult);
            return finalResult;
		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
        return null;
	}

	/**
	 * Creates the get prescription for executor result.
	 *
	 * @param securedContent
	 *            the secured content
	 * @return the be.business.connector.recipe.executor.domain. get prescription for executor result
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult createGetPrescriptionForExecutorResult(
			final byte[] securedContent) throws IntegrationModuleException {
		final MarshallerHelper<GetPrescriptionForExecutorResultSealed, Object> marshaller = new MarshallerHelper<>(
				GetPrescriptionForExecutorResultSealed.class, Object.class);
		final String requestorIdInformation = StandaloneRequestorProvider.getRequestorIdInformation();
		final String requestorTypeInformation = StandaloneRequestorProvider.getRequestorTypeInformation();
		final GetPrescriptionForExecutorResultSealed sealedResult = marshaller.unsealWithSymmKey(securedContent, getSymmKey());
		
							
		
		final be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult finalResult = new be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult(
				sealedResult);
		finalResult.setStatus(sealedResult.getStatus());
		//Don't check status in INT, because it isx done elsewhere
		//	checkStatus(finalResult);
		finalResult.setSealedContent(sealedResult.getPrescription());
		
		final KeyResult key = getKeyFromKgss(sealedResult.getEncryptionKeyId(),
				getEtkHelper().getEtks(KgssIdentifierType.NIHII_PHARMACY, requestorIdInformation).get(0).getEncoded());
		final byte[] unsealedPrescription = unsealWithSymKey(sealedResult, key, requestorIdInformation, requestorTypeInformation);
		
		finalResult.setPrescription(unsealedPrescription);
		finalResult.setEncryptionKey(key.getSecretKey().getEncoded());
		finalResult.setInsurabilityResponse(InsurabilityHandler.getInsurability());
		finalResult.setMessageId(InsurabilityHandler.getMessageId());
		return finalResult;
	}

	protected GetOpenPrescriptionForExecutor decryptGetOpenPrescriptionForExecutor(
			final GetOpenPrescriptionForExecutor getOpenPrescriptionForExecutor) throws IntegrationModuleException {
		final String requestorIdInformation = StandaloneRequestorProvider.getRequestorIdInformation();
		final KeyResult key = getKeyFromKgss(getOpenPrescriptionForExecutor.getEncryptionKeyId(),
				getEtkHelper().getEtks(KgssIdentifierType.NIHII_PHARMACY, requestorIdInformation).get(0).getEncoded());
		final byte[] unsealedPrescription = unsealPrescriptionForUnknown(key, getOpenPrescriptionForExecutor.getPrescription());
		final GetOpenPrescriptionForExecutor finalResult = new GetOpenPrescriptionForExecutor();
		finalResult.setPrescription(unsealedPrescription);
		return finalResult;
	}

	/**
	 * {@inheritDoc}
	 */
    @Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModule#markAsArchived")
    @Override
	public void markAsArchived(final String rid) throws IntegrationModuleException {
		//RidValidator.validateRid(rid);
        ApplicationConfig.getInstance().assertValidPharmacySession();
        try {
            // create param
			final byte[] sealedMarkAsArchivedParam = getSealedMarkAsArchivedParam(rid);

            // create request
			final be.fgov.ehealth.recipe.protocol.v2.MarkAsArchivedRequest request = new MarkAsArchivedRequest();
            request.setSecuredMarkAsArchivedRequest(createSecuredContentType(sealedMarkAsArchivedParam));
            request.setAdministrativeInformation(getAdministrativeInfo(rid));

            // call WS
            try {
                checkStatus(RecipeExecutorServiceImpl.getInstance().markAsArchived(request));
			} catch (final ClientTransportException cte) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
            }
		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
    }

	/**
	 * {@inheritDoc}
	 */
    @Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModule#markAsDelivered")
    @Override
	public void markAsDelivered(final String rid) throws IntegrationModuleException {
		//RidValidator.validateRid(rid);
        ApplicationConfig.getInstance().assertValidPharmacySession();
        try {
			final byte[] sealedMarkAsDeliveredParam = getSealedMarkAsDeliveredParam(rid);

			final MarkAsDeliveredRequest request = new MarkAsDeliveredRequest();
            request.setSecuredMarkAsDeliveredRequest(createSecuredContentType(sealedMarkAsDeliveredParam));
            request.setAdministrativeInformation(getAdministrativeInfo(rid));

            try {
                checkStatus(RecipeExecutorServiceImpl.getInstance().markAsDelivered(request));
			} catch (final ClientTransportException cte) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
            }
		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
    }

	/**
	 * {@inheritDoc}
	 */
    @Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModule#markAsUndelivered")
    @Override
	public void markAsUndelivered(final String rid) throws IntegrationModuleException {
		//RidValidator.validateRid(rid);
        ApplicationConfig.getInstance().assertValidPharmacySession();
        try {
			final byte[] sealedMarkAsUndeliveredParam = getSealedMarkAsUnDeliveredParam(rid);

			final MarkAsUnDeliveredRequest request = new MarkAsUnDeliveredRequest();
            request.setSecuredMarkAsUnDeliveredRequest(createSecuredContentType(sealedMarkAsUndeliveredParam));
            request.setAdministrativeInformation(getAdministrativeInfo(rid));

            try {
                checkStatus(RecipeExecutorServiceImpl.getInstance().markAsUnDelivered(request));
			} catch (final ClientTransportException cte) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
            }
		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
    }

	/**
	 * {@inheritDoc}
	 */
    @Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModule#revokePrescription")
    @Override
	public void revokePrescription(final String rid, final String reason) throws IntegrationModuleException {
//		//RidValidator.validateRid(rid);
        ApplicationConfig.getInstance().assertValidPharmacySession();
        try {
            final byte[] sealedRevokePrescriptionParam = getSealedRevokePrescriptionParam(rid, reason);

			final RevokePrescriptionForExecutorRequest request = new RevokePrescriptionForExecutorRequest();
            request.setAdministrativeInformation(getAdministrativeInfo(rid));
            request.setSecuredRevokePrescriptionRequest(createSecuredContentType(sealedRevokePrescriptionParam));

            try {
                checkStatus(RecipeExecutorServiceImpl.getInstance().revokePrescriptionForExecutor(request));
			} catch (final ClientTransportException cte) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
            }

		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
    }

	/**
	 * {@inheritDoc}
	 */
    @Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModule#listNotifications")
    @Override
	public List<ListNotificationsItem> listNotifications(final boolean readFlag) throws IntegrationModuleException {
        ApplicationConfig.getInstance().assertValidPharmacySession();
        try {
            final byte[] sealedListNotificationsParam = getSealedListNotificationsParam(readFlag);

			final ListNotificationsRequest request = new ListNotificationsRequest();
            request.setSecuredListNotificationsRequest(createSecuredContentType(sealedListNotificationsParam));

            ListNotificationsResponse response = null;
            try {
                response = RecipeExecutorServiceImpl.getInstance().listNotifications(request);
			} catch (final ClientTransportException cte) {
                throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
            }
            checkStatus(response);

            final byte[] securedContent = response.getSecuredListNotificationsResponse().getSecuredContent();
            return createListNotificationItems(securedContent);
		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
        return null;
    }

	/**
	 * {@inheritDoc}
	 */
    @Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModule#createFeedback")
    @Override
	public void createFeedback(final String prescriberId, final String rid, final byte[] feedbackText) throws IntegrationModuleException {
		validateRid(rid);
        ApplicationConfig.getInstance().assertValidPharmacySession();
        try {
            getKmehrHelper().assertValidFeedback(feedbackText);
			final List<EncryptionToken> etkRecipients = getEtkHelper().getEtks(KgssIdentifierType.NIHII, prescriberId);

            for (int i = 0; i < etkRecipients.size(); i++) {
				final EncryptionToken etkRecipient = etkRecipients.get(i);
                final byte[] sealedCreateFeedbackParam = getSealedCreateFeedbackParam(feedbackText, etkRecipient, rid, prescriberId);

				final CreateFeedbackRequest request = new CreateFeedbackRequest();
                request.setSecuredCreateFeedbackRequest(createSecuredContentType(sealedCreateFeedbackParam));
				final ExecutorServiceAdministrativeInformationType info = getAdministrativeInfo(rid);
                info.setPrescriberIdentifier(createIdentifierType(prescriberId, IdentifierTypes.SSIN.name()));
                request.setAdministrativeInformation(info);

                try {
                    checkStatus(RecipeExecutorServiceImpl.getInstance().createFeedback(request));
				} catch (final ClientTransportException cte) {
                    throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
                }
            }
		} catch (final Throwable t) {
            Exceptionutils.errorHandler(t);
        }
    }

	/**
	 * Creates the secured content type.
	 *
	 * @param content
	 *            the content
	 * @return the secured content type
	 */
	private SecuredContentType createSecuredContentType(final byte[] content) {
		final SecuredContentType secured = new SecuredContentType();
        secured.setSecuredContent(content);
        return secured;
    }

	/**
	 * Gets the administrative info.
	 *
	 * @param rid
	 *            the rid
	 * @return the administrative info
	 */
	private ExecutorServiceAdministrativeInformationType getAdministrativeInfo(final String rid) {
		final GetPrescriptionForExecutorResult prescription = getPrescriptionCache().get(rid);
		final ExecutorServiceAdministrativeInformationType info = new ExecutorServiceAdministrativeInformationType();
        if (prescription != null) {
            info.setPatientIdentifier(createIdentifierType(prescription.getPatientId(), IdentifierTypes.SSIN.name()));
            info.setPrescriberIdentifier(createIdentifierType(prescription.getPrescriberId(), IdentifierTypes.NIHII11.name()));
        } else {
            info.setPatientIdentifier(createIdentifierType("72081061175", IdentifierTypes.SSIN.name()));
            info.setPrescriberIdentifier(createIdentifierType("10998018001", IdentifierTypes.NIHII11.name()));
        }
        return info;
    }
}
