package be.business.connector.recipe.executor;

import static be.business.connector.recipe.utils.RidValidator.validateRid;

import java.io.IOException;
import java.math.BigInteger;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.collections.keyvalue.MultiKey;
import org.joda.time.DateTime;
import org.perf4j.aop.Profiled;

import com.sun.xml.ws.client.ClientTransportException;

import be.business.connector.common.ApplicationConfig;
import be.business.connector.common.StandaloneRequestorProvider;
import be.business.connector.core.domain.KgssIdentifierType;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.handlers.InsurabilityHandler;
import be.business.connector.core.utils.Exceptionutils;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.core.utils.IOUtils;
import be.business.connector.core.utils.MarshallerHelper;
import be.business.connector.core.utils.PropertyHandler;
import be.business.connector.projects.common.utils.ValidationUtils;
import be.business.connector.recipe.executor.services.RecipeExecutorServiceV4Impl;
import be.business.connector.recipe.utils.RidValidator;
import be.ehealth.technicalconnector.service.kgss.domain.KeyResult;
import be.fgov.ehealth.etee.crypto.encrypt.EncryptionToken;
import be.fgov.ehealth.recipe.core.v4.SecuredContentType;
import be.fgov.ehealth.recipe.protocol.v4.CreateFeedbackRequest;
import be.fgov.ehealth.recipe.protocol.v4.CreateFeedbackResponse;
import be.fgov.ehealth.recipe.protocol.v4.GetPrescriptionForExecutorRequest;
import be.fgov.ehealth.recipe.protocol.v4.GetPrescriptionForExecutorResponse;
import be.fgov.ehealth.recipe.protocol.v4.GetPrescriptionStatusRequest;
import be.fgov.ehealth.recipe.protocol.v4.GetPrescriptionStatusResponse;
import be.fgov.ehealth.recipe.protocol.v4.ListNotificationsRequest;
import be.fgov.ehealth.recipe.protocol.v4.ListNotificationsResponse;
import be.fgov.ehealth.recipe.protocol.v4.ListOpenPrescriptionsRequest;
import be.fgov.ehealth.recipe.protocol.v4.ListOpenPrescriptionsResponse;
import be.fgov.ehealth.recipe.protocol.v4.ListRelationsRequest;
import be.fgov.ehealth.recipe.protocol.v4.ListRelationsResponse;
import be.fgov.ehealth.recipe.protocol.v4.ListReservationsRequest;
import be.fgov.ehealth.recipe.protocol.v4.ListReservationsResponse;
import be.fgov.ehealth.recipe.protocol.v4.ListRidsHistoryRequest;
import be.fgov.ehealth.recipe.protocol.v4.ListRidsHistoryResponse;
import be.fgov.ehealth.recipe.protocol.v4.ListRidsInProcessRequest;
import be.fgov.ehealth.recipe.protocol.v4.ListRidsInProcessResponse;
import be.fgov.ehealth.recipe.protocol.v4.MarkAsArchivedRequest;
import be.fgov.ehealth.recipe.protocol.v4.MarkAsArchivedResponse;
import be.fgov.ehealth.recipe.protocol.v4.MarkAsDeliveredRequest;
import be.fgov.ehealth.recipe.protocol.v4.MarkAsDeliveredResponse;
import be.fgov.ehealth.recipe.protocol.v4.MarkAsUnDeliveredRequest;
import be.fgov.ehealth.recipe.protocol.v4.MarkAsUnDeliveredResponse;
import be.fgov.ehealth.recipe.protocol.v4.PutRidsInProcessRequest;
import be.fgov.ehealth.recipe.protocol.v4.PutRidsInProcessResponse;
import be.fgov.ehealth.recipe.protocol.v4.RevokePrescriptionRequest;
import be.fgov.ehealth.recipe.protocol.v4.RevokePrescriptionResponse;
import be.recipe.services.core.Page;
import be.recipe.services.executor.CreateFeedbackResult;
import be.recipe.services.executor.GetOpenPrescriptionForExecutor;
import be.recipe.services.executor.GetPrescriptionForExecutorResultSealed;
import be.recipe.services.executor.GetPrescriptionStatusParam;
import be.recipe.services.executor.GetPrescriptionStatusResult;
import be.recipe.services.executor.ListNotificationsItem;
import be.recipe.services.executor.ListNotificationsResult;
import be.recipe.services.executor.ListOpenPrescriptionsParam;
import be.recipe.services.executor.ListOpenPrescriptionsResult;
import be.recipe.services.executor.ListRelationsParam;
import be.recipe.services.executor.ListRelationsResult;
import be.recipe.services.executor.ListReservationsParam;
import be.recipe.services.executor.ListReservationsResult;
import be.recipe.services.executor.ListRidsHistoryParam;
import be.recipe.services.executor.ListRidsHistoryResult;
import be.recipe.services.executor.ListRidsInProcessParam;
import be.recipe.services.executor.ListRidsInProcessResult;
import be.recipe.services.executor.MarkAsArchivedResult;
import be.recipe.services.executor.MarkAsDeliveredResult;
import be.recipe.services.executor.MarkAsUndeliveredResult;
import be.recipe.services.executor.PutRidsInProcessParam;
import be.recipe.services.executor.PutRidsInProcessResult;
import be.recipe.services.executor.RevokePrescriptionResult;

/**
 * The Class HospitalExecutorIntegrationModuleV4Impl.
 */
public class HospitalExecutorIntegrationModuleV4Impl extends ExecutorIntegrationModuleImpl implements ExecutorIntegrationModuleV4 {

	
	private static final Map<MultiKey, byte[]> sessionMap = new HashMap<MultiKey, byte[]>();
	
	/**
	 * Instantiates a new executor integration module V 4 impl.
	 *
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public HospitalExecutorIntegrationModuleV4Impl() throws IntegrationModuleException {
		super();
	}

	@Override
	public ListOpenPrescriptionsResult getData(final ListOpenPrescriptionsParam param) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		ValidationUtils.validatePatientId(param.getPatientId());
		ValidationUtils.validateMandateHolderId(param.getMandateHolderId(), true);
		final ListOpenPrescriptionsRequest listOpenPrescriptionsRequest = getOpenPrescriptionList(param);
		try {
			try {
				final ListOpenPrescriptionsResponse response = RecipeExecutorServiceV4Impl.getInstance()
						.listOpenPrescriptions(listOpenPrescriptionsRequest);
				final ListOpenPrescriptionsResult result = unsealGetDataResponse(response);
				checkStatus(result);
				sessionMap.put(new MultiKey(StandaloneRequestorProvider.getRequestorIdInformation(), param.getPatientId()), result.getSession());
				return result;
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, listOpenPrescriptionsRequest.getId());
		}
		return null;
	}

	/**
	 * Unseal get data response.
	 *
	 * @param response
	 *            the get data response
	 * @return the gets the open prescription list response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListOpenPrescriptionsResult unsealGetDataResponse(final ListOpenPrescriptionsResponse response)
			throws IntegrationModuleException {
		final MarshallerHelper<ListOpenPrescriptionsResult, Object> marshaller = new MarshallerHelper<>(ListOpenPrescriptionsResult.class,
				Object.class);
		return marshaller.unsealWithSymmKey(response.getSecuredListOpenPrescriptionsResponse().getSecuredContent(), getSymmKey());
	}

	/**
	 * Gets the gets the open prescription list request.
	 *
	 * @param listOpenPrescriptionsParam
	 *            the data
	 * @return the gets the open prescription list request
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListOpenPrescriptionsRequest getOpenPrescriptionList(final ListOpenPrescriptionsParam listOpenPrescriptionsParam)
			throws IntegrationModuleException {
		listOpenPrescriptionsParam.setSession(sessionMap.get(new MultiKey( StandaloneRequestorProvider.getRequestorIdInformation(), listOpenPrescriptionsParam.getPatientId())));
		final ListOpenPrescriptionsRequest request = new ListOpenPrescriptionsRequest();
		request.setSecuredListOpenPrescriptionsRequest(createSecuredContentType(getSealedData(listOpenPrescriptionsParam)));
		request.setIssueInstant(new DateTime());
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setId(getId());
		return request;
	}

	private GetPrescriptionStatusRequest getPrescriptionStatus(final GetPrescriptionStatusParam getPrescriptionStatusParam)
			throws IntegrationModuleException {

		final GetPrescriptionStatusRequest request = new GetPrescriptionStatusRequest();
		request.setSecuredGetPrescriptionStatusRequest(createSecuredContentType(getSealedData(getPrescriptionStatusParam)));
		request.setIssueInstant(new DateTime());
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setId(getId());
		return request;
	}

	/**
	 * Gets the sealed data.
	 *
	 * @param listOpenPrescriptionsParam
	 *            the request
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private byte[] getSealedData(final ListOpenPrescriptionsParam listOpenPrescriptionsParam) throws IntegrationModuleException {
		listOpenPrescriptionsParam.setSymmKey(getSymmKey().getEncoded());
		return sealForRecipe(listOpenPrescriptionsParam, ListOpenPrescriptionsParam.class);
	}

	/**
	 * Creates the secured content type.
	 *
	 * @param content
	 *            the content
	 * @return the secured content type
	 */
	private SecuredContentType createSecuredContentType(final byte[] content) {
		final SecuredContentType secured = new SecuredContentType();
		secured.setSecuredContent(content);
		return secured;
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#getData(GetPrescriptionStatusParam)")
	@Override
	public GetPrescriptionStatusResult getData(final GetPrescriptionStatusParam param) throws IntegrationModuleException {// 222
		RidValidator.validateRid(param.getRid());
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final GetPrescriptionStatusRequest getPrescriptionStatus = getPrescriptionStatus(param);
		try {

			try {
				final GetPrescriptionStatusResponse response = RecipeExecutorServiceV4Impl.getInstance()
						.getPrescriptionStatus(getPrescriptionStatus);
				final GetPrescriptionStatusResult result = unsealGetPrescriptionStatusResponse(response);
				checkStatus(result);
				return result;
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, getPrescriptionStatus.getId());
		}
		return null;
	}

	/**
	 * Unseal get prescription status response.
	 *
	 * @param response
	 *            the get data response
	 * @return the gets the prescription status response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private GetPrescriptionStatusResult unsealGetPrescriptionStatusResponse(final GetPrescriptionStatusResponse response)
			throws IntegrationModuleException {
		final MarshallerHelper<GetPrescriptionStatusResult, Object> marshaller = new MarshallerHelper<>(GetPrescriptionStatusResult.class,
				Object.class);
		return marshaller.unsealWithSymmKey(response.getSecuredGetPrescriptionStatusResponse().getSecuredContent(), getSymmKey());
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#getData(ListRidsHistoryParam)")
	@Override
	public ListRidsHistoryResult getData(final ListRidsHistoryParam param) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		ValidationUtils.validatePatientId(param.getPatientId());
		final ListRidsHistoryRequest listPrescriptionHistory = getListRidsHistory(param);
		try {
			try {
				final ListRidsHistoryResponse response = RecipeExecutorServiceV4Impl.getInstance().listRidsHistory(listPrescriptionHistory);
				final ListRidsHistoryResult result = unsealListRidsHistoryResponse(response);
				checkStatus(result);
				sessionMap.put(new MultiKey(param.getExecutorId(), param.getPatientId()), result.getSession());
				return result;
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, listPrescriptionHistory.getId());
		}
		return null;
	}

	private ListRidsHistoryRequest getListRidsHistory(final ListRidsHistoryParam listRidsHistoryParam) throws IntegrationModuleException {
		listRidsHistoryParam.setSession(sessionMap.get(new MultiKey( listRidsHistoryParam.getExecutorId(), listRidsHistoryParam.getPatientId())));
		final ListRidsHistoryRequest request = new ListRidsHistoryRequest();
		request.setSecuredListRidsHistoryRequest(createSecuredContentType(getSealedData(listRidsHistoryParam)));
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setIssueInstant(new DateTime());
		request.setId(getId());
		return request;
	}

	private byte[] getSealedData(final ListRidsHistoryParam listRidsHistoryParam) throws IntegrationModuleException {
		listRidsHistoryParam.setSymmKey(getSymmKey().getEncoded());
		return sealForRecipe(listRidsHistoryParam, ListRidsHistoryParam.class);
	}

	/**
	 * Unseal list prescription history response.
	 *
	 * @param response
	 *            the get data response
	 * @return the list prescription history response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListRidsHistoryResult unsealListRidsHistoryResponse(final ListRidsHistoryResponse response) {
		final MarshallerHelper<ListRidsHistoryResult, Object> marshaller = new MarshallerHelper<>(ListRidsHistoryResult.class,
				Object.class);
		return marshaller.unsealWithSymmKey(response.getSecuredListRidsHistoryResponse().getSecuredContent(), getSymmKey());
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "ExecutorIntegrationModule#getPrescription")
	@Override
	public be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult getPrescription(final String rid)
			throws IntegrationModuleException {
		RidValidator.validateRid(rid);
		ApplicationConfig.getInstance().assertValidHospitalPharmacySession();
		InsurabilityHandler.setInsurability(null);
		InsurabilityHandler.setMessageId(null);
		final String guid = getId();
		try {
			// create get prescription request
			final byte[] sealedGetPrescriptionForExecutorParam = getSealedGetPrescriptionForExecutorParam(rid);

			final GetPrescriptionForExecutorRequest request = new GetPrescriptionForExecutorRequest();
			request.setDisablePatientInsurabilityCheckParam(
					Boolean.parseBoolean(getPropertyHandler().getProperty("patient.insurability.disable")));
			request.setSecuredGetPrescriptionForExecutorRequest(createSecuredContentType(sealedGetPrescriptionForExecutorParam));
			request.setIssueInstant(new DateTime());
			request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
			request.setId(guid);
			GetPrescriptionForExecutorResponse response = null;

			try {
				response = RecipeExecutorServiceV4Impl.getInstance().getPrescriptionForExecutor(request);
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
			final byte[] securedContent = response.getSecuredGetPrescriptionForExecutorResponse().getSecuredContent();
			final be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult finalResult = createGetPrescriptionForExecutorResult(
					securedContent);
			checkStatus(finalResult);
			getPrescriptionCache().put(rid, finalResult);
			return finalResult;
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, guid);
		}
		return null;
	}
	
	/**
	 * Creates the get prescription for executor result.
	 *
	 * @param securedContent
	 *            the secured content
	 * @return the be.business.connector.recipe.executor.domain. get prescription for executor result
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult createGetPrescriptionForExecutorResult(
			final byte[] securedContent) throws IntegrationModuleException {
		final MarshallerHelper<GetPrescriptionForExecutorResultSealed, Object> marshaller = new MarshallerHelper<>(
				GetPrescriptionForExecutorResultSealed.class, Object.class);
		final String requestorIdInformation = StandaloneRequestorProvider.getRequestorIdInformation();
		final String requestorTypeInformation = StandaloneRequestorProvider.getRequestorTypeInformation();
		final GetPrescriptionForExecutorResultSealed sealedResult = marshaller.unsealWithSymmKey(securedContent, getSymmKey());
		
							
		
		final be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult finalResult = new be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult(
				sealedResult);
		finalResult.setStatus(sealedResult.getStatus());
		checkStatus(finalResult);
		finalResult.setSealedContent(sealedResult.getPrescription());
		
		final KeyResult key = getKeyFromKgss(sealedResult.getEncryptionKeyId(),
				getEtkHelper().getEtks(KgssIdentifierType.NIHII_HOSPITAL, requestorIdInformation).get(0).getEncoded());
		final byte[] unsealedPrescription = unsealWithSymKey(sealedResult, key, requestorIdInformation, requestorTypeInformation);
		
		finalResult.setPrescription(unsealedPrescription);
		finalResult.setEncryptionKey(key.getSecretKey().getEncoded());
		finalResult.setInsurabilityResponse(InsurabilityHandler.getInsurability());
		finalResult.setMessageId(InsurabilityHandler.getMessageId());
		return finalResult;
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#decryptGetOpenPrescriptionForExecutor")
	@Override
	public GetOpenPrescriptionForExecutor decryptGetOpenPrescriptionForExecutor(final GetOpenPrescriptionForExecutor gopfe)
			throws IntegrationModuleException {
		try {
			final String requestorIdInformation = StandaloneRequestorProvider.getRequestorIdInformation();
			final KeyResult key = getKeyFromKgss(gopfe.getEncryptionKeyId(),
					getEtkHelper().getEtks(KgssIdentifierType.NIHII_PHARMACY, requestorIdInformation).get(0).getEncoded());
			final byte[] unsealedPrescription = unsealPrescriptionForUnknown(key, gopfe.getPrescription());
			final GetOpenPrescriptionForExecutor finalResult = new GetOpenPrescriptionForExecutor();
			finalResult.setCreationDate(gopfe.getCreationDate());
			finalResult.setEncryptionKeyId(gopfe.getEncryptionKeyId());
			finalResult.setFeedbackAllowed(gopfe.isFeedbackAllowed());
			finalResult.setId(gopfe.getId());
			finalResult.setPatientId(gopfe.getPatientId());
			finalResult.setPrescriberId(gopfe.getPrescriberId());
			finalResult.setPrescription(IOUtils.decompress(unsealedPrescription));
			finalResult.setPrescriptionType(gopfe.getPrescriptionType());
			finalResult.setRid(gopfe.getRid());
			finalResult.setStatus(gopfe.getStatus());
			finalResult.setExpirationDate(gopfe.getExpirationDate());
			return finalResult;
		} catch (final IOException ex) {
			Exceptionutils.errorHandler(ex, "error.data.uncompression");
		}
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#markAsArchived")
	@Override
	public void markAsArchived(final String rid) throws IntegrationModuleException {
		RidValidator.validateRid(rid);
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final MarkAsArchivedRequest request = new MarkAsArchivedRequest();
		try {

			final byte[] sealedMarkAsArchivedParam = getSealedMarkAsArchivedParam(rid);
			request.setSecuredMarkAsArchivedRequest(createSecuredContentType(sealedMarkAsArchivedParam));
			request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
			request.setIssueInstant(new DateTime());
			request.setId(getId());

			// call WS
			MarkAsArchivedResponse response;
			try {
				response = RecipeExecutorServiceV4Impl.getInstance().markAsArchived(request);
				final MarshallerHelper<MarkAsArchivedResult, Object> marshaller = new MarshallerHelper<>(MarkAsArchivedResult.class,
						Object.class);
				final MarkAsArchivedResult result = (MarkAsArchivedResult) marshaller
						.unsealWithSymmKey(response.getSecuredMarkAsArchivedResponse().getSecuredContent(), getSymmKey());
				checkStatus(result);
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#markAsDelivered")
	@Override
	public void markAsDelivered(final String rid) throws IntegrationModuleException {
		RidValidator.validateRid(rid);
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final MarkAsDeliveredRequest request = new MarkAsDeliveredRequest();
		try {
			final byte[] sealedMarkAsDeliveredParam = getSealedMarkAsDeliveredParam(rid);

			request.setSecuredMarkAsDeliveredRequest(createSecuredContentType(sealedMarkAsDeliveredParam));
			request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
			request.setIssueInstant(new DateTime());
			request.setId(getId());

			MarkAsDeliveredResponse response = null;
			try {
				response = RecipeExecutorServiceV4Impl.getInstance().markAsDelivered(request);
				final MarshallerHelper<MarkAsDeliveredResult, Object> marshaller = new MarshallerHelper<>(MarkAsDeliveredResult.class,
						Object.class);
				final MarkAsDeliveredResult result = (MarkAsDeliveredResult) marshaller
						.unsealWithSymmKey(response.getSecuredMarkAsDeliveredResponse().getSecuredContent(), getSymmKey());
				checkStatus(result);
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#markAsUnDelivered")
	@Override
	public void markAsUndelivered(final String rid) throws IntegrationModuleException {
		RidValidator.validateRid(rid);
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final MarkAsUnDeliveredRequest request = new MarkAsUnDeliveredRequest();
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setId(getId());
		try {
			final byte[] sealedMarkAsUnDeliveredParam = getSealedMarkAsUnDeliveredParam(rid);
			request.setSecuredMarkAsUnDeliveredRequest(createSecuredContentType(sealedMarkAsUnDeliveredParam));
			request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
			request.setIssueInstant(new DateTime());
			request.setId(getId());

			MarkAsUnDeliveredResponse response = null;
			try {
				response = RecipeExecutorServiceV4Impl.getInstance().markAsUnDelivered(request);
				final MarshallerHelper<MarkAsUndeliveredResult, Object> marshaller = new MarshallerHelper<>(MarkAsUndeliveredResult.class,
						Object.class);
				final MarkAsUndeliveredResult result = (MarkAsUndeliveredResult) marshaller
						.unsealWithSymmKey(response.getSecuredMarkAsUnDeliveredResponse().getSecuredContent(), getSymmKey());
				checkStatus(result);
				if (getPrescriptionCache().containsKey(rid)) {
					getPrescriptionCache().remove(rid);
				} 
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#revokePrescription")
	@Override
	public void revokePrescription(final String rid, final String reason) throws IntegrationModuleException {
		RidValidator.validateRid(rid);
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final RevokePrescriptionRequest request = new RevokePrescriptionRequest();
		try {
			final byte[] sealedRevokePrescriptionParam = getSealedRevokePrescriptionParam(rid, reason);

			request.setSecuredRevokePrescriptionRequest(createSecuredContentType(sealedRevokePrescriptionParam));
			request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
			request.setIssueInstant(new DateTime());
			request.setId(getId());

			RevokePrescriptionResponse response = null;
			try {
				response = RecipeExecutorServiceV4Impl.getInstance().revokePrescriptionForExecutor(request);
				final MarshallerHelper<RevokePrescriptionResult, Object> marshaller = new MarshallerHelper<>(RevokePrescriptionResult.class,
						Object.class);
				final RevokePrescriptionResult result = (RevokePrescriptionResult) marshaller
						.unsealWithSymmKey(response.getSecuredRevokePrescriptionResponse().getSecuredContent(), getSymmKey());
				checkStatus(result);
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}

		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#listNotifications")
	@Override
	public List<ListNotificationsItem> listNotifications(final boolean readFlag) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final ListNotificationsRequest request = new ListNotificationsRequest();

		try {
			final byte[] sealedListNotificationsParam = getSealedListNotificationsParam(readFlag);

			request.setSecuredListNotificationsRequest(createSecuredContentType(sealedListNotificationsParam));
			request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
			request.setIssueInstant(new DateTime());
			request.setId(getId());

			ListNotificationsResponse response = null;
			try {
				response = RecipeExecutorServiceV4Impl.getInstance().listNotifications(request);
				final MarshallerHelper<ListNotificationsResult, Object> marshaller = new MarshallerHelper<>(ListNotificationsResult.class,
						Object.class);
				final ListNotificationsResult result = (ListNotificationsResult) marshaller
						.unsealWithSymmKey(response.getSecuredListNotificationsResponse().getSecuredContent(), getSymmKey());
				checkStatus(result);
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}

			final byte[] securedContent = response.getSecuredListNotificationsResponse().getSecuredContent();
			return createListNotificationItems(securedContent);
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#createFeedback")
	@Override
	public void createFeedback(final String prescriberId, final String rid, final byte[] feedbackText) throws IntegrationModuleException {
		validateRid(rid);
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final CreateFeedbackRequest request = new CreateFeedbackRequest();
		try {
			getKmehrHelper().assertValidFeedback(feedbackText);
			final List<EncryptionToken> etkRecipients = getEtkHelper().getEtks(KgssIdentifierType.NIHII_HOSPITAL, prescriberId);

			for (int i = 0; i < etkRecipients.size(); i++) {
				final EncryptionToken etkRecipient = etkRecipients.get(i);
				final byte[] sealedCreateFeedbackParam = getSealedCreateFeedbackParam(feedbackText, etkRecipient, rid, prescriberId);

				request.setSecuredCreateFeedbackRequest(createSecuredContentType(sealedCreateFeedbackParam));
				request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
				request.setIssueInstant(new DateTime());
				request.setId(getId());

				try {
					final CreateFeedbackResponse response = RecipeExecutorServiceV4Impl.getInstance().createFeedback(request);
					final CreateFeedbackResult result = unsealCreateFeedbackResponse(response);
					checkStatus(result);
				} catch (final ClientTransportException cte) {
					throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
				}

			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#getData(ListReservationsParam)")
	@Override
	public ListReservationsResult getData(final ListReservationsParam param) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final Calendar lastSyncDate = Calendar.getInstance();
		try {
			try {
				final ListReservationsResult completeResult = new ListReservationsResult();
				ListReservationsResult result;
				do {
					BigInteger currentPage;
					if (param.getPage() == null) {
						param.setPage(new Page());
					}
					if (param.getPage() != null && param.getPage().getPageNumber() == null) {
						currentPage = BigInteger.valueOf(0);
					}else {
						currentPage = param.getPage().getPageNumber();
					}
					param.getPage().setPageNumber(currentPage);
					final ListReservationsRequest request = getListReservations(param);
					final ListReservationsResponse response = RecipeExecutorServiceV4Impl.getInstance().listReservations(request);
					result = unsealListReservationsResponse(response);
					checkStatus(result);
					writeReservationsOnDisk(param, result, lastSyncDate);
					completeResult.getItems().addAll(result.getItems());
					param.getPage().setPageNumber(BigInteger.valueOf(currentPage.intValue() + 1));
				} while (result != null && result.isHasMoreResults() != null && result.isHasMoreResults().booleanValue());

				return completeResult;

			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}

	/**
	 * Gets the all reservations request.
	 *
	 * @param param
	 *            the data
	 * @return the all reservations request
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected ListReservationsRequest getListReservations(final ListReservationsParam param) throws IntegrationModuleException {
		param.setSymmKey(getSymmKey().getEncoded());
		param.setStartDate(param.getStartDate() == null ?readLastDateToDisk(): param.getStartDate());
		final ListReservationsRequest request = new ListReservationsRequest();
		request.setSecuredListReservationsRequest(createSecuredContentType(getSealedData(param)));
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setIssueInstant(new DateTime());
		request.setId(getId());
		return request;
	}

	private byte[] getSealedData(final ListReservationsParam listReservationsParam) throws IntegrationModuleException {
		listReservationsParam.setSymmKey(getSymmKey().getEncoded());
		return sealForRecipe(listReservationsParam, ListReservationsParam.class);
	}

	/**
	 * Unseal get all reservations response.
	 * 
	 * @param response
	 *            the get data response
	 * @return the gets the all reservations response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListReservationsResult unsealListReservationsResponse(final ListReservationsResponse response) {
		final MarshallerHelper<ListReservationsResult, Object> marshaller = new MarshallerHelper<>(ListReservationsResult.class,
				Object.class);
		final ListReservationsResult result = marshaller
				.unsealWithSymmKey(response.getSecuredListReservationsResponse().getSecuredContent(), getSymmKey());
		return result;
	}
	
	private CreateFeedbackResult unsealCreateFeedbackResponse(final CreateFeedbackResponse response) {
		final MarshallerHelper<CreateFeedbackResult, Object> marshaller = new MarshallerHelper<>(CreateFeedbackResult.class,
				Object.class);
		final CreateFeedbackResult result = marshaller
				.unsealWithSymmKey(response.getSecuredCreateFeedbackResponse().getSecuredContent(), getSymmKey());
		return result;
	}
	

	//
	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#getData(ListRidsInProcessParam)")
	@Override
	public ListRidsInProcessResult getData(final ListRidsInProcessParam param) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		try {
			final ListRidsInProcessRequest request = getListRidsInProcess(param);
			try {
				final ListRidsInProcessResponse response = RecipeExecutorServiceV4Impl.getInstance().listRidsInProcess(request);
				final ListRidsInProcessResult result = unsealListRidsInProcessResponse(response);
				checkStatus(result);
				return result;
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}

	/**
	 * Gets the all ridsi N process request.
	 *
	 * @param listRidsInProcessParam
	 *            the data
	 * @return the all ridsi N process request
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListRidsInProcessRequest getListRidsInProcess(final ListRidsInProcessParam listRidsInProcessParam)
			throws IntegrationModuleException {
		listRidsInProcessParam.setSymmKey(getSymmKey().getEncoded());
		final ListRidsInProcessRequest request = new ListRidsInProcessRequest();
		request.setSecuredListRidsInProcessRequest(createSecuredContentType(getSealedData(listRidsInProcessParam)));
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setIssueInstant(new DateTime());
		request.setId(getId());
		return request;
	}

	/**
	 * Gets the sealed data.
	 *
	 * @param listRidsInProcessParam
	 *            the request
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private byte[] getSealedData(final ListRidsInProcessParam listRidsInProcessParam) throws IntegrationModuleException {
		return sealForRecipe(listRidsInProcessParam, ListRidsInProcessParam.class);
	}

	//
	/**
	 * {@inheritDoc}
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#putData(PutAllRidInProcessParam)")
	@Override
	public PutRidsInProcessResult putData(final PutRidsInProcessParam param) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		final PutRidsInProcessRequest request = getPutAllRidInProcess(param);
		try {

			try {
				final PutRidsInProcessResponse response = RecipeExecutorServiceV4Impl.getInstance().putRidsInProcess(request);
				final PutRidsInProcessResult result = unsealPutRidsInProcessResponse(response);
				checkStatus(result);
				return result;
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t, request.getId());
		}
		return null;
	}

	/**
	 * Put all ridsi N process request.
	 *
	 * @param putRidsInProcessParam
	 *            the data
	 * @return the put data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected PutRidsInProcessRequest getPutAllRidInProcess(final PutRidsInProcessParam putRidsInProcessParam)
			throws IntegrationModuleException {
		// Update param with symmKey
		putRidsInProcessParam.setSymmKey(getSymmKey().getEncoded());

		final PutRidsInProcessRequest request = new PutRidsInProcessRequest();
		request.setSecuredPutRidsInProcessRequest(createSecuredContentType(getSealedData(putRidsInProcessParam)));
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setId(getId());
		request.setIssueInstant(new DateTime());
		return request;
	}

	private byte[] getSealedData(final PutRidsInProcessParam putRidsInProcessParam) throws IntegrationModuleException {
		putRidsInProcessParam.setSymmKey(getSymmKey().getEncoded());
		return sealForRecipe(putRidsInProcessParam, PutRidsInProcessParam.class);
	}

	/**
	 * Unseal put all rid in process response.
	 *
	 * @param response
	 *            the get data response
	 * @return the put all rid in process response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private PutRidsInProcessResult unsealPutRidsInProcessResponse(final PutRidsInProcessResponse response) {
		final MarshallerHelper<PutRidsInProcessResult, Object> marshaller = new MarshallerHelper<>(PutRidsInProcessResult.class,
				Object.class);
		final PutRidsInProcessResult result = marshaller
				.unsealWithSymmKey(response.getSecuredPutRidsInProcessResponse().getSecuredContent(), getSymmKey());
		return result;
	}

	/**
	 * Unseal get all rid in process response.
	 *
	 * @param getDataResponse
	 *            the get data response
	 * @return the gets the all rid in process result
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListRidsInProcessResult unsealListRidsInProcessResponse(final ListRidsInProcessResponse response) {
		final MarshallerHelper<ListRidsInProcessResult, Object> marshaller = new MarshallerHelper<>(ListRidsInProcessResult.class,
				Object.class);
		final ListRidsInProcessResult result = marshaller
				.unsealWithSymmKey(response.getSecuredListRidsInProcessResponse().getSecuredContent(), getSymmKey());
		return result;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws IntegrationModuleException
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.ExecutorIntegrationModuleV4#getAndMarkAsDelivered")
	@Override
	public be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult getAndMarkAsDelivered(final String rid)
			throws IntegrationModuleException {
		try {
			final be.business.connector.recipe.executor.domain.GetPrescriptionForExecutorResult getPrescriptionForExecutorResult;
			if (getPrescriptionCache().containsKey(rid)) {
				getPrescriptionForExecutorResult = getPrescriptionCache().get(rid);
			} else {
				getPrescriptionForExecutorResult = getPrescription(rid);
			}
			markAsDelivered(rid);
			markAsArchived(rid);
			return getPrescriptionForExecutorResult;
		} catch (final Exception t) {
			Exceptionutils.errorHandler(t);
		} return null;
	}
	
	/**
	 * Mark as archived async.
	 *
	 * @param rid
	 *            the rid
	 */
	private void getPrescriptionAsync(final String rid) {
		new Thread(new Runnable() {
			public void run() {
				try {
					getPrescription(rid);
				} catch (final IntegrationModuleException e) {
					throw new RuntimeException(e);
				}
			}
		}).start();
	}

	/**
	 * Mark as archived async.
	 *
	 * @param rid
	 *            the rid
	 */
	private void markAsDeliveredAsync(final String rid) {
		new Thread(new Runnable() {
			public void run() {
				try {
					markAsDelivered(rid);
				} catch (final IntegrationModuleException e) {
					throw new RuntimeException(e);
				}
			}
		}).start();
	}

	/**
	 * Mark as archived async.
	 *
	 * @param rid
	 *            the rid
	 */
	private void markAsArchivedAsync(final String rid) {
		new Thread(new Runnable() {
			public void run() {
				try {
					markAsArchived(rid);
				} catch (final IntegrationModuleException e) {
					throw new RuntimeException(e);
				}
			}
		}).start();
	}

	@Override
	public ListRelationsResult getData(final ListRelationsParam patientRelationParam) {
		ApplicationConfig.getInstance().assertValidPharmacySession();
		ValidationUtils.validatePatientId(patientRelationParam.getMandateHolderId());
		try {
			final ListRelationsRequest request = getListPatientRelation(patientRelationParam);
			try {
				final ListRelationsResponse dataResponse = RecipeExecutorServiceV4Impl.getInstance().listRelations(request);
				final ListRelationsResult unsealedResponse = unsealListRelationsResponse(dataResponse);
				checkStatus(unsealedResponse);
				sessionMap.put(new MultiKey(patientRelationParam.getExecutorId(), patientRelationParam.getPatientId()), unsealedResponse.getSession());
				return unsealedResponse;
			} catch (final ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}

	/**
	 * Creates a {@link ListRelationsParam} request.
	 *
	 * @param data
	 *            the data
	 * @return the patient relation
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private ListRelationsRequest getListPatientRelation(final ListRelationsParam data) throws IntegrationModuleException {
		data.setSymmKey(getSymmKey().getEncoded());
		data.setSession(sessionMap.get(new MultiKey( data.getExecutorId(), data.getPatientId())));
		
		final ListRelationsRequest request = new ListRelationsRequest();
		request.setSecuredListRelationsRequest(createSecuredContentType(getSealedData(data)));
		request.setIssueInstant(new DateTime());
		request.setProgramId(PropertyHandler.getInstance().getProperty("programIdentification"));
		request.setId(getId());
		return request;
	}

	/**
	 * Gets the sealed data.
	 *
	 * @param data
	 *            the data
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private byte[] getSealedData(final ListRelationsParam data) throws IntegrationModuleException {
		data.setSymmKey(getSymmKey().getEncoded());
		return sealForRecipe(data, ListRelationsParam.class);
	}

	/**
	 * Unseal list relations response.
	 *
	 * @param response
	 *            the response
	 * @return the list relations result
	 */
	private ListRelationsResult unsealListRelationsResponse(final ListRelationsResponse response) {
		final MarshallerHelper<ListRelationsResult, Object> marshaller = new MarshallerHelper<>(ListRelationsResult.class, Object.class);
		final ListRelationsResult result = marshaller.unsealWithSymmKey(response.getSecuredListRelationsResponse().getSecuredContent(),
				getSymmKey());
		return result;
	}
}