package be.business.connector.recipe.patient;

import static be.business.connector.recipe.utils.RidValidator.validateRid;

import java.util.List;

import com.sun.xml.ws.client.ClientTransportException;

import be.business.connector.common.ApplicationConfig;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.handlers.InsurabilityHandler;
import be.business.connector.core.utils.Exceptionutils;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.core.utils.MarshallerHelper;
import be.business.connector.recipe.patient.services.RecipePatientServiceImpl;
import be.business.connector.recipe.utils.RidValidator;
import be.fgov.ehealth.etee.crypto.encrypt.EncryptionToken;
import be.fgov.ehealth.recipe.core.v1.SecuredContentType;
import be.fgov.ehealth.recipe.protocol.v1.GetPrescriptionForPatientRequest;
import be.fgov.ehealth.recipe.protocol.v1.GetPrescriptionForPatientResponse;
import be.fgov.ehealth.recipe.protocol.v1.ListPatientPrescriptionsRequest;
import be.fgov.ehealth.recipe.protocol.v1.ListPatientPrescriptionsResponse;
import be.fgov.ehealth.recipe.protocol.v1.RevokePatientPrescriptionRequest;
import be.fgov.ehealth.recipe.protocol.v1.UpdatePatientFeedbackFlagRequest;
import be.recipe.services.patient.GetPrescriptionForPatientResult;
import be.recipe.services.patient.ListPatientPrescriptionsParam;
import be.recipe.services.patient.ListPatientPrescriptionsResult;
import be.recipe.services.patient.UpdateFeedbackFlagParam;

/**
 * The Class PatientIntegrationModuleImpl.
 */
public class PatientIntegrationModuleImpl extends AbstractPatientIntegrationModule implements PatientIntegrationModule {

	/**
	 * Instantiates a new patient integration module impl.
	 *
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public PatientIntegrationModuleImpl() throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertInitialized();
		init();
	}

	@Override
	public ListPatientPrescriptionsResult listOpenPrescriptions() throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidSession();
		try {
			// init helper
			MarshallerHelper<ListPatientPrescriptionsResult, ListPatientPrescriptionsParam> helper = new MarshallerHelper<ListPatientPrescriptionsResult, ListPatientPrescriptionsParam>(
					ListPatientPrescriptionsResult.class, ListPatientPrescriptionsParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create param
			ListPatientPrescriptionsParam param = new ListPatientPrescriptionsParam();
			param.setSymmKey(getSymmKey().getEncoded());

			// create request
			ListPatientPrescriptionsRequest request = new ListPatientPrescriptionsRequest();
			request.setSecuredListPatientPrescriptionsRequest(
					createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param))));

			// call sealed WS
			ListPatientPrescriptionsResponse response = null;
			try {
				response = RecipePatientServiceImpl.getInstance().listOpenPrescriptions(request);
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}

			checkStatus(response);

			// unseal WS response
			ListPatientPrescriptionsResult result = helper
					.unsealWithSymmKey(response.getSecuredListPatientPrescriptionsResponse().getSecuredContent(), getSymmKey());

			return result;

		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}

		return null;
	}

	@Override
	public void updateFeedbackFlag(String rid, boolean feedbackAllowed) throws IntegrationModuleException {
		validateRid(rid);
		ApplicationConfig.getInstance().assertValidSession();
		try {

			// init helper
			MarshallerHelper<Object, UpdateFeedbackFlagParam> helper = new MarshallerHelper<>(Object.class, UpdateFeedbackFlagParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create param
			final UpdateFeedbackFlagParam param = new UpdateFeedbackFlagParam();
			param.setAllowFeedback(feedbackAllowed);
			param.setRid(rid);

			final UpdatePatientFeedbackFlagRequest request = new UpdatePatientFeedbackFlagRequest();
			request.setSecuredUpdatePatientFeedbackFlagRequest(
					createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param))));

			// call sealed WS
			try {
				checkStatus(RecipePatientServiceImpl.getInstance().updateFeedbackFlag(request));
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}
		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}
	}

	@Override
	public void revokePrescription(String rid, String reason) throws IntegrationModuleException {
		RidValidator.validateRid(rid);
		ApplicationConfig.getInstance().assertValidSession();
		try {
			final byte[] sealedRevokePrescriptionParam = getSealedRevokePrescriptionParam(rid, reason);

			RevokePatientPrescriptionRequest request = new RevokePatientPrescriptionRequest();
			request.setSecuredRevokePatientPrescriptionRequest(createSecuredContentType(sealedRevokePrescriptionParam));

			try {
				checkStatus(RecipePatientServiceImpl.getInstance().revokePrescription(request));
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}

		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}

	}

	@Override
	public GetPrescriptionForPatientResult getPrescription(String rid) throws IntegrationModuleException {
		validateRid(rid);
		ApplicationConfig.getInstance().assertValidSession();
		InsurabilityHandler.setInsurability(null);
		InsurabilityHandler.setMessageId(null);

		try {
			final byte[] sealedContent = getSealedGetPrescriptionForPatientParam(rid);

			GetPrescriptionForPatientRequest request = new GetPrescriptionForPatientRequest();
			SecuredContentType securedContent = new SecuredContentType();
			securedContent.setSecuredContent(sealedContent);
			request.setSecuredGetPrescriptionForPatientRequest(securedContent);

			GetPrescriptionForPatientResponse response = null;
			try {
				response = RecipePatientServiceImpl.getInstance().getPrescriptionForPatient(request);
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.executor"), cte);
			}
			checkStatus(response);

			final byte[] securedContentResponse = response.getSecuredGetPrescriptionForPatientResponse().getSecuredContent();
			final GetPrescriptionForPatientResult finalResult = unsealPrescription(securedContentResponse);

			return finalResult;
		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}
}