package be.business.connector.recipe.patient.services;

import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.services.GenericWebserviceCaller;
import be.business.connector.core.services.GenericWebserviceRequest;
import be.business.connector.projects.common.utils.EndpointResolver;
import be.fgov.ehealth.recipe.protocol.v1.GetPrescriptionForPatientRequest;
import be.fgov.ehealth.recipe.protocol.v1.GetPrescriptionForPatientResponse;
import be.fgov.ehealth.recipe.protocol.v1.ListPatientPrescriptionsRequest;
import be.fgov.ehealth.recipe.protocol.v1.ListPatientPrescriptionsResponse;
import be.fgov.ehealth.recipe.protocol.v1.RevokePatientPrescriptionRequest;
import be.fgov.ehealth.recipe.protocol.v1.RevokePatientPrescriptionResponse;
import be.fgov.ehealth.recipe.protocol.v1.UpdatePatientFeedbackFlagRequest;
import be.fgov.ehealth.recipe.protocol.v1.UpdatePatientFeedbackFlagResponse;

/**
 * The Class RecipePatientServiceImpl.
 */
public class RecipePatientServiceImpl implements RecipePatientService {

	/** The Constant ENDPOINT_NAME. */
	private static final String ENDPOINT_NAME = "endpoint.patient";

	/** The Constant SERVICE_NAME. */
	private static final String SERVICE_NAME = RecipePatientServiceImpl.class.getName();

	/** The recipe patient service. */
	private static RecipePatientService recipePatientService;

	/**
	 * Instantiates a new recipe patient service impl.
	 */
	private RecipePatientServiceImpl() {
	}

	/**
	 * Gets the singleton instance of RecipePrescriberServiceImpl.
	 *
	 * @return singleton instance of RecipePrescriberServiceImpl
	 */
	public static RecipePatientService getInstance() {
		if (recipePatientService == null) {
			recipePatientService = new RecipePatientServiceImpl();
		}
		return recipePatientService;
	}

	/**
	 * Override.
	 *
	 * @param getPrescriptionForPrescriberRequest
	 *            the get prescription for prescriber request
	 * @return the prescription for patient
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Override
	public GetPrescriptionForPatientResponse getPrescriptionForPatient(
			GetPrescriptionForPatientRequest getPrescriptionForPrescriberRequest) throws IntegrationModuleException {
		return GenericWebserviceCaller.callGenericWebservice(
				createDefaultGenericWebserviceRequest(getPrescriptionForPrescriberRequest),
				GetPrescriptionForPatientResponse.class);
	}

	/**
	 * Override.
	 *
	 * @param revokePrescriptionRequest
	 *            the revoke prescription request
	 * @return the revoke patient prescription response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Override
	public RevokePatientPrescriptionResponse revokePrescription(
			RevokePatientPrescriptionRequest revokePrescriptionRequest) throws IntegrationModuleException {
		return GenericWebserviceCaller.callGenericWebservice(
				createDefaultGenericWebserviceRequest(revokePrescriptionRequest),
				RevokePatientPrescriptionResponse.class);
	}

	/**
	 * Override.
	 *
	 * @param listOpenPrescriptionsRequest
	 *            the list open prescriptions request
	 * @return the list patient prescriptions response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Override
	public ListPatientPrescriptionsResponse listOpenPrescriptions(
			ListPatientPrescriptionsRequest listOpenPrescriptionsRequest) throws IntegrationModuleException {
		return GenericWebserviceCaller.callGenericWebservice(
				createDefaultGenericWebserviceRequest(listOpenPrescriptionsRequest),
				ListPatientPrescriptionsResponse.class);
	}

	/**
	 * Override.
	 *
	 * @param updateFeedbackFlagRequest
	 *            the update feedback flag request
	 * @return the update patient feedback flag response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Override
	public UpdatePatientFeedbackFlagResponse updateFeedbackFlag(
			UpdatePatientFeedbackFlagRequest updateFeedbackFlagRequest)
			throws IntegrationModuleException {
		return GenericWebserviceCaller.callGenericWebservice(
				createDefaultGenericWebserviceRequest(updateFeedbackFlagRequest),
				UpdatePatientFeedbackFlagResponse.class);
	}

	/**
	 * Creates the default generic webservice request.
	 *
	 * @param requestObject
	 *            the request object
	 * @return the generic webservice request
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private GenericWebserviceRequest createDefaultGenericWebserviceRequest(final Object requestObject)
			throws IntegrationModuleException {
		final GenericWebserviceRequest request = new GenericWebserviceRequest();
		request.setRequest(requestObject);
		request.setEndpoint(EndpointResolver.getEndpointUrlString(ENDPOINT_NAME));
		request.setServiceName(SERVICE_NAME);
		request.setAddLoggingHandler(true);
		request.setAddSoapFaultHandler(true);
		request.setAddMustUnderstandHandler(true);
		request.setAddInsurabilityHandler(false);
		return request;
	}
}