package be.business.connector.recipe.prescriber;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.perf4j.aop.Profiled;

import be.business.connector.common.StandaloneRequestorProvider;
import be.business.connector.common.module.AbstractIntegrationModule;
import be.business.connector.core.domain.KgssIdentifierType;
import be.business.connector.core.ehealth.services.KgssService;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.technical.connector.utils.Crypto;
import be.business.connector.core.utils.EncryptionUtils;
import be.business.connector.core.utils.Exceptionutils;
import be.business.connector.core.utils.MarshallerHelper;
import be.business.connector.core.utils.STSHelper;
import be.business.connector.core.utils.SessionValidator;
import be.business.connector.recipe.utils.KmehrHelper;
import be.ehealth.technicalconnector.service.kgss.domain.KeyResult;
import be.ehealth.technicalconnector.session.Session;
import be.ehealth.technicalconnector.session.SessionItem;
import be.fgov.ehealth.commons.core.v1.IdentifierType;
import be.fgov.ehealth.commons.core.v1.LocalisedString;
import be.fgov.ehealth.commons.core.v1.StatusType;
import be.fgov.ehealth.commons.protocol.v1.ResponseType;
import be.fgov.ehealth.etee.crypto.encrypt.EncryptionToken;
import be.recipe.services.prescriber.GetPrescriptionStatusParam;
import be.recipe.services.prescriber.ListOpenRidsParam;
import be.recipe.services.prescriber.ListRidsHistoryParam;
import be.recipe.services.prescriber.PutVisionParam;
import be.recipe.services.prescriber.ValidationPropertiesParam;

/**
 * The Class AbstractPrescriberIntegrationModule.
 */
public abstract class AbstractPrescriberIntegrationModule extends AbstractIntegrationModule {

	/** The Constant LOG. */
	private static final Logger LOG = Logger.getLogger(AbstractPrescriberIntegrationModule.class);

	/** The prescription cache. */
	private final Map<String, String> prescriptionCache = new HashMap<>();

	/** The kmehr helper. */
	private final KmehrHelper kmehrHelper = new KmehrHelper();

	/** The key cache. */
	protected Map<String, KeyResult> keyCache = new HashMap<>();

	/** The kgss service. */
	protected KgssService kgssService = be.business.connector.core.ehealth.services.KgssServiceImpl.getInstance();

	/**
	 * Instantiates a new abstract prescriber integration module.
	 *
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public AbstractPrescriberIntegrationModule() throws IntegrationModuleException {
		super();
		LOG.info("*************** Prescriber System module init correctly *******************");
	}

	/**
	 * Check status.
	 *
	 * @param response
	 *            the response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public void checkStatus(final ResponseType response) throws IntegrationModuleException {
		if (!EHEALTH_SUCCESS_CODE_100.equals(response.getStatus().getCode())
				&& !EHEALTH_SUCCESS_CODE_200.equals(response.getStatus().getCode())) {
			LOG.error("Error Status received : " + response.getStatus().getCode());
			throw new IntegrationModuleException(getLocalisedMsg(response.getStatus()));
		}
	}

	/**
	 * Check status.
	 *
	 * @param response
	 *            the response
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public void checkStatus(final be.recipe.services.core.ResponseType response) throws IntegrationModuleException {
		if (!EHEALTH_SUCCESS_CODE_100.equals(response.getStatus().getCode())
				&& !EHEALTH_SUCCESS_CODE_200.equals(response.getStatus().getCode())) {
			LOG.error("Error Status received : " + response.getStatus().getCode());
			throw new IntegrationModuleException(getLocalisedMsg(response.getStatus()), response);
		}
	}

	/**
	 * Gets the localised msg.
	 *
	 * @param status
	 *            the status
	 * @return the localised msg
	 */
	private String getLocalisedMsg(final StatusType status) {
		final String locale = IntegrationModuleException.getUserLocale();
		for (final LocalisedString msg : status.getMessages()) {
			if (msg.getLang() != null && locale.equalsIgnoreCase(msg.getLang().value())) {
				return msg.getValue();
			}
		}
		if (!status.getMessages().isEmpty()) {
			return status.getMessages().get(0).getValue();
		}
		return status.getCode();
	}

	/**
	 * Gets the localised msg.
	 *
	 * @param status
	 *            the status
	 * @return the localised msg
	 */
	private String getLocalisedMsg(final be.recipe.services.core.StatusType status) {
		final String locale = IntegrationModuleException.getUserLocale();
		for (final be.recipe.services.core.LocalisedString msg : status.getMessages()) {
			if (msg.getLang() != null && locale.equalsIgnoreCase(msg.getLang().value())) {
				return msg.getValue();
			}
		}
		if (!status.getMessages().isEmpty()) {
			return status.getMessages().get(0).getValue();
		}
		return status.getCode();
	}

	/**
	 * Creates the identifier type.
	 *
	 * @param id
	 *            the id
	 * @param type
	 *            the type
	 * @return the identifier type
	 */
	protected IdentifierType createIdentifierType(final String id, final String type) {
		final IdentifierType ident = new IdentifierType();
		ident.setId(id);
		ident.setType(type);
		return ident;
	}

	/**
	 * Gets the patient id.
	 *
	 * @param rid
	 *            the rid
	 * @return the patient id
	 */
	protected String getPatientId(final String rid) {
		if (prescriptionCache.containsKey(rid)) {
			return prescriptionCache.get(rid);
		}
		return "72081061175";
	}

	/**
	 * Sets the patient id.
	 *
	 * @param rid
	 *            the rid
	 * @param patientId
	 *            the patient id
	 */
	protected void setPatientId(final String rid, final String patientId) {
		prescriptionCache.put(rid, patientId);
	}

	/**
	 * Sets the personal password.
	 *
	 * @param personalPassword
	 *            the new personal password
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public void setPersonalPassword(final String personalPassword) throws IntegrationModuleException {
		final SessionItem sessionItem = Session.getInstance().getSession();
		SessionValidator.assertValidSession(sessionItem);

		try {
			final String niss = STSHelper.getNiss(sessionItem.getSAMLToken().getAssertion());
			final String nihii = STSHelper.getNihii(sessionItem.getSAMLToken().getAssertion());
			
			final EncryptionUtils encryptionUtils = EncryptionUtils.getInstance();
			encryptionUtils.unlockPersonalKey(StringUtils.isNotBlank(niss)? niss:nihii, personalPassword);
			dataUnsealer = encryptionUtils.initUnsealing();
			final List<EncryptionToken> tokens = getEtkHelper().getEtks(getIdentifierType(),
					StandaloneRequestorProvider.getRequestorIdInformation());
			encryptionUtils.verifyDecryption(tokens.get(0));
		} catch (final Exception e) {
			throw new IntegrationModuleException(e);
		}

	}

	private KgssIdentifierType getIdentifierType() {
		final String type = STSHelper.getType(Session.getInstance().getSession().getSAMLToken().getAssertion());
		return type.equals("HOSPITAL")? KgssIdentifierType.NIHII_HOSPITAL: KgssIdentifierType.NIHII;
	}

	/**
	 * Gets the kmehr helper.
	 *
	 * @return the kmehr helper
	 */
	protected KmehrHelper getKmehrHelper() {
		return kmehrHelper;
	}

	/**
	 * Gets the sealed data.
	 *
	 * @param validationPropertiesParam
	 *            the validation properties param
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected byte[] getSealedData(final ValidationPropertiesParam validationPropertiesParam) throws IntegrationModuleException {
		return sealForRecipe(validationPropertiesParam, ValidationPropertiesParam.class);
	}

	/**
	 * Seal for recipe.
	 *
	 * @param <T>
	 *            the generic type
	 * @param data
	 *            the data
	 * @param type
	 *            the type
	 * @return the byte[]
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private <T> byte[] sealForRecipe(final T data, final Class<T> type) throws IntegrationModuleException {
		final MarshallerHelper<Object, T> helper = new MarshallerHelper<>(Object.class, type);
		final EncryptionToken etkRecipe = getEtkHelper().getRecipe_ETK().get(0);
		return sealRequest(etkRecipe, helper.toXMLByteArray(data));
	}

	/**
	 * Gets the new key.
	 *
	 * @param patientId
	 *            the patient id
	 * @param prescriptionType
	 *            the prescription type
	 * @return the new key
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	public KeyResult getNewKey(final String patientId, final String prescriptionType) throws IntegrationModuleException {
		KeyResult key = null;

		final String cacheId = "(" + patientId + "#" + prescriptionType + ")";
		if (keyCache.containsKey(cacheId)) {
			key = keyCache.get(cacheId);
			keyCache.remove(cacheId);
		} else {
			key = getNewKeyFromKgss(prescriptionType, StandaloneRequestorProvider.getRequestorIdInformation(), null, patientId,
					getEtkHelper().getSystemETK().get(0).getEncoded());
		}
		return key;
	}

	/**
	 * Gets the new key from kgss.
	 *
	 * @param prescriptionType
	 *            the prescription type
	 * @param prescriberId
	 *            the prescriber id
	 * @param executorId
	 *            the executor id
	 * @param patientId
	 *            the patient id
	 * @param myEtk
	 *            the my etk
	 * @return the new key from kgss
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#getNewKeyFromKgss")
	protected KeyResult getNewKeyFromKgss(final String prescriptionType, final String prescriberId, final String executorId,
			final String patientId, final byte[] myEtk) throws IntegrationModuleException {
		// For test, when a sim key is specified in the config
		if (getPropertyHandler().hasProperty("test_kgss_key")) {
			return getKeyFromKgss(null, null);
		}

		final EncryptionToken etkKgss = getEtkHelper().getKGSS_ETK().get(0);
		final List<String> credentialTypes = getPropertyHandler().getMatchingProperties("kgss.createPrescription.ACL." + prescriptionType);

		try {
			return kgssService.retrieveNewKey(etkKgss.getEncoded(), credentialTypes, prescriberId, executorId, patientId, myEtk);
		} catch (final Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}

	/**
	 * Seal prescription for unknown.
	 *
	 * @param key
	 *            the key
	 * @param messageToProtect
	 *            the message to protect
	 * @return the byte[]
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#sealPrescriptionForUnknown")
	public byte[] sealPrescriptionForUnknown(final KeyResult key, final byte[] messageToProtect) throws IntegrationModuleException {
		return Crypto.seal(messageToProtect, key.getSecretKey(), key.getKeyId());
	}

	/**
	 * Unseal feedback.
	 *
	 * @param message
	 *            the message
	 * @return the byte[]
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#unsealFeedback")
	protected byte[] unsealFeedback(final byte[] message) throws IntegrationModuleException {
		return unsealNotiffeed(message);
	}

	/**
	 * Seal notification.
	 *
	 * @param paramEncryptionToken
	 *            the param encryption token
	 * @param paramArrayOfByte
	 *            the param array of byte
	 * @return the byte[]
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#sealNotification")
	protected byte[] sealNotification(final EncryptionToken paramEncryptionToken, final byte[] paramArrayOfByte)
			throws IntegrationModuleException {
		return Crypto.seal(paramEncryptionToken, paramArrayOfByte);
	}

//	/**
//	 * Gets the validation properties request.
//	 *
//	 * @param param
//	 *            the data
//	 * @return the validation properties request
//	 * @throws IntegrationModuleException
//	 *             the integration module exception
//	 */
//	protected GetValidationPropertiesRequest getValidationProperties(final ValidationPropertiesParam param) throws IntegrationModuleException {
//		param.setSymmKey(getSymmKey().getEncoded());
//		param.setClientPropertiesVersion(getPropertyHandler().getProperty("sdk.version"));
//		final GetValidationPropertiesRequest getValidationProperties = new GetValidationPropertiesRequest();
//		getValidationProperties.setSecuredGetValidationPropertiesRequest(createSecuredContentType((getSealedData(param)));
//		getValidationProperties.setProgramIdentification(PropertyHandler.getInstance().getProperty("programIdentification"));
//		getValidationProperties.setMguid(UUID.randomUUID().toString());
//		return getValidationProperties;
//	}

	/**
	 * Marshall.
	 *
	 * @param <T>
	 *            the generic type
	 * @param data
	 *            the data
	 * @param type
	 *            the type
	 * @return the byte[]
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	private <T> byte[] marshall(final T data, final Class<T> type) {
		final MarshallerHelper<Object, T> helper = new MarshallerHelper<>(Object.class, type);
		return helper.toXMLByteArray(data);
	}

	/**
	 * Gets the sealed data.
	 *
	 * @param listRidHistoryParam
	 *            the request
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected byte[] getSealedData(final ListRidsHistoryParam listRidHistoryParam) throws IntegrationModuleException {
		return sealForRecipe(listRidHistoryParam, ListRidsHistoryParam.class);
	}
	
	/**
	 * Gets the sealed data.
	 *
	 * @param request
	 *            the request
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected byte[] getSealedData(final GetPrescriptionStatusParam request) throws IntegrationModuleException {
		return sealForRecipe(request, GetPrescriptionStatusParam.class);
	}

	/**
	 * Gets the sealed data.
	 *
	 * @param putVisionParam
	 *            the request
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected byte[] getSealedData(final PutVisionParam putVisionParam) throws IntegrationModuleException {
		return sealForRecipe(putVisionParam, PutVisionParam.class);
	}

	

	/**
	 * Gets the sealed data.
	 *
	 * @param listOpenRidsParam
	 *            the request
	 * @return the sealed data
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	protected byte[] getSealedData(final ListOpenRidsParam listOpenRidsParam) throws IntegrationModuleException {
		return sealForRecipe(listOpenRidsParam, ListOpenRidsParam.class);
	}
}