package be.business.connector.recipe.prescriber;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.bouncycastle.util.encoders.Base64;
import org.perf4j.aop.Profiled;

import com.sun.xml.ws.client.ClientTransportException;

import be.business.connector.common.ApplicationConfig;
import be.business.connector.common.StandaloneRequestorProvider;
import be.business.connector.core.domain.IdentifierTypes;
import be.business.connector.core.domain.KgssIdentifierType;
import be.business.connector.core.exceptions.IntegrationModuleException;
import be.business.connector.core.utils.Exceptionutils;
import be.business.connector.core.utils.I18nHelper;
import be.business.connector.core.utils.IOUtils;
import be.business.connector.core.utils.MarshallerHelper;
import be.business.connector.projects.common.utils.ValidationUtils;
import be.business.connector.recipe.prescriber.services.RecipePrescriberServiceImpl;
import be.ehealth.technicalconnector.service.kgss.domain.KeyResult;
import be.fgov.ehealth.etee.crypto.encrypt.EncryptionToken;
import be.fgov.ehealth.recipe.core.v1.CreatePrescriptionAdministrativeInformationType;
import be.fgov.ehealth.recipe.core.v1.PrescriberServiceAdministrativeInformationType;
import be.fgov.ehealth.recipe.core.v1.SecuredContentType;
import be.fgov.ehealth.recipe.core.v1.SendNotificationAdministrativeInformationType;
import be.fgov.ehealth.recipe.protocol.v1.AliveCheckRequest;
import be.fgov.ehealth.recipe.protocol.v1.AliveCheckResponse;
import be.fgov.ehealth.recipe.protocol.v1.CreatePrescriptionRequest;
import be.fgov.ehealth.recipe.protocol.v1.CreatePrescriptionResponse;
import be.fgov.ehealth.recipe.protocol.v1.GetPrescriptionForPrescriberRequest;
import be.fgov.ehealth.recipe.protocol.v1.GetPrescriptionForPrescriberResponse;
import be.fgov.ehealth.recipe.protocol.v1.ListFeedbacksRequest;
import be.fgov.ehealth.recipe.protocol.v1.ListFeedbacksResponse;
import be.fgov.ehealth.recipe.protocol.v1.ListOpenPrescriptionsRequest;
import be.fgov.ehealth.recipe.protocol.v1.ListOpenPrescriptionsResponse;
import be.fgov.ehealth.recipe.protocol.v1.RevokePrescriptionRequest;
import be.fgov.ehealth.recipe.protocol.v1.SendNotificationRequest;
import be.fgov.ehealth.recipe.protocol.v1.UpdateFeedbackFlagRequest;
import be.recipe.services.prescriber.CreatePrescriptionParam;
import be.recipe.services.prescriber.CreatePrescriptionResult;
import be.recipe.services.prescriber.GetListOpenPrescriptionParam;
import be.recipe.services.prescriber.GetListOpenPrescriptionResult;
import be.recipe.services.prescriber.GetPrescriptionForPrescriberParam;
import be.recipe.services.prescriber.GetPrescriptionForPrescriberResult;
import be.recipe.services.prescriber.ListFeedbackItem;
import be.recipe.services.prescriber.ListFeedbacksParam;
import be.recipe.services.prescriber.ListFeedbacksResult;
import be.recipe.services.prescriber.RevokePrescriptionParam;
import be.recipe.services.prescriber.SendNotificationParam;
import be.recipe.services.prescriber.UpdateFeedbackFlagParam;

public class PrescriberIntegrationModuleImpl extends AbstractPrescriberIntegrationModule
		implements PrescriberIntegrationModule {

	private static final  Logger LOG = Logger.getLogger(PrescriberIntegrationModuleImpl.class);

    public PrescriberIntegrationModuleImpl() throws IntegrationModuleException {
        super();
    }

	/**
	 * Prepare create prescription.
	 *
	 * @param patientId
	 *            the patient id
	 * @throws IntegrationModuleException
	 */
	@Profiled(logFailuresSeparately = true, tag = "PrescriberIntegrationModule#prepareCreatePrescription")
	public void prepareCreatePrescription(String patientId, String prescriptionType) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidSession();

		String cacheId = "(" + patientId + "#" + prescriptionType + ")";
		keyCache.put(cacheId, getNewKeyFromKgss(prescriptionType, StandaloneRequestorProvider.getRequestorIdInformation(), null, patientId,
				getEtkHelper().getSystemETK().get(0).getEncoded()));
	}
	
	/**
	 * Creates the prescription.
	 *
	 * @param feedbackRequested
	 *            the feedback requested
	 * @param patientId
	 *            the patient id
	 * @param prescription
	 *            the prescription
	 * @param prescriptionType
	 *            the prescription type
	 * @return the string
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#createPrescription")
	@Override
	public String createPrescription(boolean feedbackRequested, String patientId, byte[] prescription, String prescriptionType)
			throws IntegrationModuleException {

		ApplicationConfig.getInstance().assertValidSession();
		ValidationUtils.validatePatientId(patientId);
		try {

			getKmehrHelper().assertValidKmehrPrescription(prescription, prescriptionType);

			// init helper
			MarshallerHelper<CreatePrescriptionResult, CreatePrescriptionParam> helper = new MarshallerHelper<>(
					CreatePrescriptionResult.class, CreatePrescriptionParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create sealed prescription
			byte[] message = IOUtils.compress(prescription);

			KeyResult key = getNewKey(patientId, prescriptionType);
			message = sealPrescriptionForUnknown(key, message);

			final String requestorIdInformation = StandaloneRequestorProvider.getRequestorIdInformation();

			// create sealed content
			CreatePrescriptionParam params = new CreatePrescriptionParam();
			params.setPrescription(message);// ok
			params.setPrescriptionType(prescriptionType);
			params.setFeedbackRequested(feedbackRequested);
			params.setKeyId(key.getKeyId());
			params.setSymmKey(getSymmKey().getEncoded());
			params.setPatientId(patientId);
			params.setPrescriberId(requestorIdInformation);

			// create request
			CreatePrescriptionRequest request = new CreatePrescriptionRequest();
			request.setSecuredCreatePrescriptionRequest(
					createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(params))));

			// create administrative info
			CreatePrescriptionAdministrativeInformationType info = new CreatePrescriptionAdministrativeInformationType();
			request.setAdministrativeInformation(info);
			info.setKeyIdentifier(Base64.decode(key.getKeyId()));
			info.setPrescriptionType(prescriptionType);
			info.setPatientIdentifier(createIdentifierType(patientId, IdentifierTypes.SSIN.name()));
			info.setPrescriberIdentifier(createIdentifierType(requestorIdInformation, IdentifierTypes.NIHII11.name()));

			// WS call
			CreatePrescriptionResponse response = RecipePrescriberServiceImpl.getInstance().createPrescription(request);
			checkStatus(response);

			// unseal response
			CreatePrescriptionResult result = helper.unsealWithSymmKey(response.getSecuredCreatePrescriptionResponse().getSecuredContent(),
					getSymmKey());

			// update the local cache
			setPatientId(result.getRid(), patientId);

			return result.getRid();
		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}
	
	/**
	 * Creates the secured content type.
	 *
	 * @param content
	 *            the content
	 * @return the secured content type
	 */
	private SecuredContentType createSecuredContentType(final byte[] content) {
		final SecuredContentType secured = new SecuredContentType();
		secured.setSecuredContent(content);
		return secured;
	}

	/**
	 * Gets the prescription.
	 *
	 * @param rid
	 *            the rid
	 * @return the prescription
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#getPrescription")
	@Override
	public GetPrescriptionForPrescriberResult getPrescription(String rid) throws IntegrationModuleException {
		//RidValidator.validatePrescriberRid(rid);
		ApplicationConfig.getInstance().assertValidSession();
		try {

			// init helper
			MarshallerHelper<GetPrescriptionForPrescriberResult, GetPrescriptionForPrescriberParam> helper = new MarshallerHelper<GetPrescriptionForPrescriberResult, GetPrescriptionForPrescriberParam>(
					GetPrescriptionForPrescriberResult.class, GetPrescriptionForPrescriberParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create sealed request
			GetPrescriptionForPrescriberParam param = new GetPrescriptionForPrescriberParam();
			param.setRid(rid);
			param.setSymmKey(getSymmKey().getEncoded());
			param.setPrescriberId(StandaloneRequestorProvider.getRequestorIdInformation());

			// build request
			GetPrescriptionForPrescriberRequest request = new GetPrescriptionForPrescriberRequest();
			request.setSecuredGetPrescriptionForPrescriberRequest(
					createSecuredContentType((sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param)))));

			PrescriberServiceAdministrativeInformationType info = new PrescriberServiceAdministrativeInformationType();

			// id added
			info.setPatientIdentifier(createIdentifierType(getPatientId(rid), IdentifierTypes.SSIN.name()));
			request.setAdministrativeInformation(info);

			// call sealed WS
			GetPrescriptionForPrescriberResponse response = null;

			try {
				response = RecipePrescriberServiceImpl.getInstance().getPrescriptionForPrescriber(request);
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}

			checkStatus(response);

			// unseal WS response
			GetPrescriptionForPrescriberResult result = helper
					.unsealWithSymmKey(response.getSecuredGetPrescriptionForPrescriberResponse().getSecuredContent(), getSymmKey());

			KeyResult key = getKeyFromKgss(result.getEncryptionKeyId(), getEtkHelper().getSystemETK().get(0).getEncoded());
			byte[] unsealedPrescription = IOUtils.decompress(unsealPrescriptionForUnknown(key, result.getPrescription()));
			result.setPrescription(unsealedPrescription);

			// update the local cache
			setPatientId(result.getRid(), result.getPatientId());

			return result;
		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}

	@Profiled(logFailuresSeparately = true, tag = "PrescriberIntegrationModule#ping")
	@Override
	public void ping() throws IntegrationModuleException {

		AliveCheckResponse response = null;
		try {
			response = RecipePrescriberServiceImpl.getInstance().aliveCheck(new AliveCheckRequest());
		} catch (ClientTransportException cte) {
			throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
		}
		LOG.info("Ping response : " + response.getAliveCheckResult());
		checkStatus(response);

	}

	/**
	 * Cancel prescription.
	 *
	 * @param rid
	 *            the rid
	 * @param reason
	 *            the reason
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#revokePrescription")
	@Override
	public void revokePrescription(String rid, String reason) throws IntegrationModuleException {
		//RidValidator.validatePrescriberRid(rid);
		ApplicationConfig.getInstance().assertValidSession();

		try {
			// init helper
			MarshallerHelper<Object, RevokePrescriptionParam> helper = new MarshallerHelper<Object, RevokePrescriptionParam>(Object.class,
					RevokePrescriptionParam.class);

			// get Recipe ETK
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create params
			RevokePrescriptionParam params = new RevokePrescriptionParam();
			params.setRid(rid);
			params.setReason(reason);
			params.setPrescriberId(StandaloneRequestorProvider.getRequestorIdInformation());

			// create request
			RevokePrescriptionRequest request = new RevokePrescriptionRequest();
			request.setSecuredRevokePrescriptionRequest(
					createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(params))));

			// Admin Info for eHealth
			PrescriberServiceAdministrativeInformationType info = new PrescriberServiceAdministrativeInformationType();
			info.setPatientIdentifier(createIdentifierType(getPatientId(rid), IdentifierTypes.SSIN.name()));

			request.setAdministrativeInformation(info);

			// call WS
			try {
				checkStatus(RecipePrescriberServiceImpl.getInstance().revokePrescription(request));
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}

		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}

	}

	/**
	 * List open prescription.
	 *
	 * @param patientId
	 * @return the list for a geiven patient.
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#listOpenPrescription")
	@Override
	public List<String> listOpenPrescription(String patientId) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidSession();
		try {
			// init helper
			MarshallerHelper<GetListOpenPrescriptionResult, GetListOpenPrescriptionParam> helper = new MarshallerHelper<GetListOpenPrescriptionResult, GetListOpenPrescriptionParam>(
					GetListOpenPrescriptionResult.class, GetListOpenPrescriptionParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create param
			GetListOpenPrescriptionParam param = new GetListOpenPrescriptionParam();
			param.setSymmKey(getSymmKey().getEncoded());
			param.setPrescriberId(StandaloneRequestorProvider.getRequestorIdInformation());
			param.setPatientId(patientId);

			// create request
			ListOpenPrescriptionsRequest request = new ListOpenPrescriptionsRequest();
			request.setSecuredListOpenPrescriptionsRequest(
					createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param))));

			// call sealed WS
			ListOpenPrescriptionsResponse response = null;
			try {
				response = RecipePrescriberServiceImpl.getInstance().listOpenPrescriptions(request);
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}

			checkStatus(response);

			// unseal WS response
			GetListOpenPrescriptionResult result = helper
					.unsealWithSymmKey(response.getSecuredListOpenPrescriptionsResponse().getSecuredContent(), getSymmKey());

			return result.getPrescriptions() == null ? new ArrayList<String>() : result.getPrescriptions();

		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}

		return null;
	}

	/**
	 * List open prescription.
	 *
	 * @return the list
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#listOpenPrescription")
	@Override
	public List<String> listOpenPrescription() throws IntegrationModuleException {
		return listOpenPrescription(null);
	}

	/**
	 * Address prescription.
	 *
	 * @param notificationText
	 *            the notification text
	 * @param patientId
	 *            the patient id
	 * @param executorId
	 *            the executor id
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "PrescriberIntegrationModule#sendNotification")
	@Override
	public void sendNotification(byte[] notificationText, String patientId, String executorId) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidSession();
		ValidationUtils.validatePatientId(patientId);
		try {
			getKmehrHelper().assertValidNotification(notificationText);
			ValidationUtils.validatePatientId(patientId);

			// init helper
			MarshallerHelper<Object, SendNotificationParam> helper = new MarshallerHelper<>(Object.class, SendNotificationParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// get recipient etk
			List<EncryptionToken> etkRecipients = getEtkHelper().getEtks(KgssIdentifierType.NIHII_PHARMACY, executorId);

			byte[] notificationZip = IOUtils.compress(notificationText);

			for (int i = 0; i < etkRecipients.size(); i++) {
				EncryptionToken etkRecipient = etkRecipients.get(0);

				byte[] notificationSealed = sealNotification(etkRecipient, notificationZip);

				// create param
				SendNotificationParam param = new SendNotificationParam();
				param.setContent(notificationSealed);
				param.setExecutorId(executorId);
				param.setPrescriberId(StandaloneRequestorProvider.getRequestorIdInformation());
				param.setPatientId(patientId);

				// create request
				SendNotificationRequest request = new SendNotificationRequest();
				request.setSecuredSendNotificationRequest(
						createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param))));
				SendNotificationAdministrativeInformationType info = new SendNotificationAdministrativeInformationType();
				info.setExecutorIdentifier(createIdentifierType(executorId, IdentifierTypes.SSIN.name()));
				info.setPatientIdentifier(createIdentifierType(patientId, IdentifierTypes.SSIN.name()));
				request.setAdministrativeInformation(info);

				// call sealed WS
				try {
					checkStatus(RecipePrescriberServiceImpl.getInstance().sendNotification(request));
				} catch (ClientTransportException cte) {
					throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
				}
			}
		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}
	}

	/**
	 * Update feedback flag.
	 *
	 * @param rid
	 *            the rid
	 * @param feedbackAllowed
	 *            the feedback allowed
	 * @throws IntegrationModuleException
	 *             the integration module exception
	 */
	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#updateFeedbackFlag")
	@Override
	public void updateFeedbackFlag(String rid, boolean feedbackAllowed) throws IntegrationModuleException {
//		//RidValidator.validatePrescriberRid(rid);
		ApplicationConfig.getInstance().assertValidSession();
		try {

			// init helper
			MarshallerHelper<Object, UpdateFeedbackFlagParam> helper = new MarshallerHelper<>(Object.class, UpdateFeedbackFlagParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create param
			UpdateFeedbackFlagParam param = new UpdateFeedbackFlagParam();
			param.setAllowFeedback(feedbackAllowed);
			param.setRid(rid);
			param.setPrescriberId(StandaloneRequestorProvider.getRequestorIdInformation());

			UpdateFeedbackFlagRequest request = new UpdateFeedbackFlagRequest();
			request.setSecuredUpdateFeedbackFlagRequest(
					createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param))));
			PrescriberServiceAdministrativeInformationType info = new PrescriberServiceAdministrativeInformationType();

			// Admin Info for eHealth
			info.setPatientIdentifier(createIdentifierType(getPatientId(rid), IdentifierTypes.SSIN.name()));
			request.setAdministrativeInformation(info);

			// call sealed WS
			try {
				checkStatus(RecipePrescriberServiceImpl.getInstance().updateFeedbackFlag(request));
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}
		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}

	}

	@Profiled(logFailuresSeparately = true, tag = "0.PrescriberIntegrationModule#listFeedback")
	@Override
	public List<ListFeedbackItem> listFeedback(boolean readFlag) throws IntegrationModuleException {
		ApplicationConfig.getInstance().assertValidSession();

		final String requestorIdInformation = StandaloneRequestorProvider.getRequestorIdInformation();
		try {
			// init helper
			MarshallerHelper<ListFeedbacksResult, ListFeedbacksParam> helper = new MarshallerHelper<ListFeedbacksResult, ListFeedbacksParam>(
					ListFeedbacksResult.class, ListFeedbacksParam.class);

			// get recipe etk
			final List<EncryptionToken> etkRecipes = getEtkHelper().getRecipe_ETK();

			// create param
			ListFeedbacksParam param = new ListFeedbacksParam();
			param.setReadFlag(readFlag);
			param.setSymmKey(getSymmKey().getEncoded());
			param.setPrescriberId(requestorIdInformation);//TODO: remove once SAML-integration is done.

			// create request
			ListFeedbacksRequest request = new ListFeedbacksRequest();
			request.setSecuredListFeedbacksRequest(createSecuredContentType(sealRequest(etkRecipes.get(0), helper.toXMLByteArray(param))));

			// call sealed WS
			ListFeedbacksResponse response = null;

			try {
				response = RecipePrescriberServiceImpl.getInstance().listFeedbacks(request);
			} catch (ClientTransportException cte) {
				throw new IntegrationModuleException(I18nHelper.getLabel("error.connection.prescriber"), cte);
			}
			checkStatus(response);

			// unseal WS response
			List<ListFeedbackItem> feedbacks = helper
					.unsealWithSymmKey(response.getSecuredListFeedbacksResponse().getSecuredContent(), getSymmKey()).getFeedbacks();

			for (int i = 0; i < feedbacks.size(); i++) {
				be.business.connector.recipe.prescriber.domain.ListFeedbackItem item = new be.business.connector.recipe.prescriber.domain.ListFeedbackItem(
						feedbacks.get(i));
				byte[] content = item.getContent();
				try {
					content = unsealFeedback(content);
					content = content == null ? content : IOUtils.decompress(content);
					item.setContent(content);
				} catch (Throwable t) {
					item.setLinkedException(t);
				}
				feedbacks.set(i, item);
			}
			return feedbacks;

		} catch (Throwable t) {
			Exceptionutils.errorHandler(t);
		}
		return null;
	}
}