package be.business.connector.recipe.utils;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.io.LineIterator;
import org.apache.commons.lang3.StringUtils;

import be.business.connector.core.utils.PropertyHandler;

/**
 * The Class MedFileUtils.
 */
public class MedFileUtils {

	/** The med files map. */
	private static Map<String, MedFile> medFilesMap = new HashMap<String, MedFile>();

	/** The property handler. */
	private PropertyHandler propertyHandler;

	/**
	 * Instantiates a new med file utils.
	 *
	 * @param propertyHandler
	 *            the property handler
	 */
	public MedFileUtils(PropertyHandler propertyHandler) {
		try {
			this.propertyHandler = propertyHandler;
			final String filePath = propertyHandler.getProperty("medFilesDirectory");
			final Path dir = Paths.get(filePath);
			if (Files.exists(dir)) {
				try (DirectoryStream<Path> directoryStream = Files.newDirectoryStream(dir)) {
					parseDirectoryStream(directoryStream);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Gets the CNK code.
	 *
	 * @param cnk
	 *            the cnk
	 * @return the CNK code
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 * @throws ParseException
	 *             the parse exception
	 */
	public MedFile getCNKCode(final String cnk) throws IOException, ParseException {
		return medFilesMap.get(cnk);
	}

	/**
	 * Parses the directory stream.
	 *
	 * @param directoryStream
	 *            the directory stream
	 * @throws IOException
	 *             Signals that an I/O exception has occurred.
	 * @throws ParseException
	 *             the parse exception
	 */
	private static void parseDirectoryStream(final DirectoryStream<Path> directoryStream) throws IOException, ParseException {
		for (final Path path : directoryStream) {
			if (path.getFileName().toString().startsWith("MED")) {
				final LineIterator it = org.apache.commons.io.FileUtils.lineIterator(path.toFile(), "UTF-8");
				try {
					while (it.hasNext()) {
						final MedFile medfile = parseLine(it.nextLine());
						medFilesMap.put(medfile.getCNK(), medfile);
					}
				} finally {
					LineIterator.closeQuietly(it);
				}
			}
		}
	}

	/**
	 * Parses the line.
	 *
	 * @param line
	 *            the line
	 * @return the med file
	 * @throws ParseException
	 *             the parse exception
	 */
	private static MedFile parseLine(final String line) throws ParseException {
		final MedFile medFile = new MedFile();
		if (StringUtils.isEmpty(line)) {
			return medFile;
		}
		switch (MedFileTypeEnum.valueOf(line.substring(0, 5))) {
		case MED01:
			// if (cnk.equalsIgnoreCase(line.substring(13, 20))) {
				medFile.setMedFileType(MedFileTypeEnum.MED01);
				medFile.setDateSuplement(parseDate(line.substring(5, 13)));
				medFile.setCNK(line.substring(13, 20));
				medFile.setCommercialStatus(CommercialStatusEnum.valueOf(line.substring(20, 21)));
				medFile.setCategorie(CategorieEnum.valueOf(line.substring(21, 22)));
				medFile.setFrBenaming(line.substring(22, 72).trim());
				medFile.setNlBenaming(line.substring(72, 122).trim());
				medFile.setPubliekprijsIncBTW(parseLong(line.substring(122, 129)));
				medFile.setAtcCode(line.substring(129, 136).trim());
			// }
			break;
		case MED02:
			// medFile.setMedFileType(MedFileTypeEnum.MED02);
			// medFile.setDateSuplement(parseDate(line.substring(5, 13)));
			// medFile.setTerugbetaling(CategorieTerugbetaling.valueOf(line.substring(13, 14)));
			// medFile.setRemGeldActief(parseLong(line.substring(14, 19)));
			// medFile.setRemGeldWIGW(parseLong(line.substring(19, 24)));
			break;
		case MED03:
			// medFile.setMedFileType(MedFileTypeEnum.MED03);
			// medFile.setAtcCode(line.substring(5, 12).trim());
			// medFile.setOmschrijvingFR(line.substring(12, 82).trim());
			// medFile.setOmschrijvingNL(line.substring(82, 152).trim());
			break;
		default:
			throw new RuntimeException("Invalid line found in file.");
		}
		return medFile;
	}

	/**
	 * Parses the date.
	 *
	 * @param substring
	 *            the substring
	 * @return the date
	 * @throws ParseException
	 *             the parse exception
	 */
	private static Date parseDate(final String substring) throws ParseException {
		final DateFormat formatter = new SimpleDateFormat("yyyyMMdd");
		return (Date) formatter.parse(substring);
	}

	/**
	 * Parses the long.
	 *
	 * @param number
	 *            the number
	 * @return the float
	 */
	private static float parseLong(String number) {
		number = number.substring(0, number.length() - 2) + "." + number.substring(number.length() - 2);
		return Float.parseFloat(number);
	}

	/**
	 * Gets the property handler.
	 *
	 * @return the propertyHandler
	 */
	public PropertyHandler getPropertyHandler() {
		if (propertyHandler == null) {
			propertyHandler = PropertyHandler.getInstance();
		}
		return propertyHandler;
	}

	/**
	 * Sets the property handler.
	 *
	 * @param propertyHandler
	 *            the propertyHandler to set
	 */
	public void setPropertyHandler(PropertyHandler propertyHandler) {
		this.propertyHandler = propertyHandler;
	}
}